/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  araprod.c
*
* Part of: PROD's program araprod
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     The araprod program computes a parallel composition of given
*     labelled transition systems (LTS's) and hides specified actions.
*     As a side effect, a high level net corresponding to the
*     composition is generated as well as the (PROD style)
*     reachability graph of the net. The LTS files have the format
*     designed for the ARA tool by Technical Research Centre of Finland.
*     On-the-fly verification is possible with the aid of a tester.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "error.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "gphutil.h"
#include "graph.h"
#include "mark.h"
#include "encode.h"
#include "ptnet.h"
#include "ptsymm.h"
#include "genesig.h"
#include "cmdnm.h"
#include "araprod.h"

BOOL lacB = FALSE; 
             /* LTS transitions associated with an action are kept in a tree
	      * if lacB is TRUE, and otherwise in a list. */
BOOL transB = FALSE;
             /* Target nodes of an LTS transition are kept in
	      * a tree if transB is TRUE, and otherwise in a list. */
BALTREE *actTree = NULL;     /* A tree of actions ordered by name. */
BALTREE *actNumTree = NULL;  /* A tree of actions ordered by number. */
TUPLE *arpDouble = NULL;     /* A binary tuple. */
TUPLE *arpSingle = NULL;     /* A unary tuple. */
PLC *arpPlc;          /* A place in a low level net. */
PLC *arp2Plc;         /* A place in a low level net. */
TRN *arpTrn;          /* A transition in a low level net. */
LIST *domList = NULL;        /* A list of elements of DOMAIN. */
LIST *transList = NULL;      /* A list of LTS transitions. */
ACTION **actTbl = NULL;      /* For each transition in a high level net,
			      * this table tells the corresponding action. */
DATATYPE *arpPointer; /* A pointer to an encoded instance of
			      * a transition in a high level net. */
BALTREE *flowTree = NULL;    /* A tree of flows. */
ACTION *arpAction;    /* An action. */
unsigned long arpCounter; /* A counter used, among other purposes,
				  * in the generation of the description
				  * of a high level net. */
BOOL goEnd;           /* A flag used in the generation of
			      * the description of a high level net. */
LIST *goList = NULL;         /* A list of numbers used in the generation of
				the description of a high level net. */
DOMAIN arpMin;        /* The least node number (testers excluded).*/
DOMAIN arpMax;        /* The greatest node number (testers excluded). */
DOMAIN tstrSys;       /* The number of a tester LTS. */
DOMAIN tstrMin;       /* The least node number in the tester. */
DOMAIN tstrMax;       /* The greatest node number in the tester. */
LACTION *arpLaction;  /* A reference to the part of an LTS associated
				with a certain action .*/
LIST *sysotList = NULL;      /* A flow, i.e. a list of flow components. */
LIST *sysot2List = NULL;     /* A flow. */
SYSOT *arpSysot;      /* A flow component, consisting of an LTS
			      * number, an LTS transition and a target node
			      * of the transition. */
unsigned arpHiTr;     /* The number of a transition in a high level
			      * net. */
unsigned arpIndex;    /* An index used for referring to variables
			      * of a transition in a high level net. */
TRANS *arpTrans;      /* An LTS transition. */
DOMAIN *arpTarget;    /* A pointer to a target node of an LTS
			      * transition. */
DOMAIN *arpVars = NULL;      /* An array of values of variables of
			      * a transition in a high level net. */
FILE *netFile = NULL;        /* A file for the description of a high level
			      * net. */

/***************************************************************************
*
* Function:  TransCons
* Purpose:   This routine constructs an LTS transition. Such transition
*            is associated with a certain action and a certain LTS
*            and consists of a so called source node and a set of target
*            nodes, i.e. nodes that are reachable from the source node by
*            a single execution of the action. Target nodes are kept
*            in a tree if transB is TRUE, and otherwise in a list. The set
*            of target nodes is initialized to be the empty set.
* Arguments: source - the source node
* Returns:   a pointer to the constructed transition
* See also:  TransDiscard, TransDump, TransKeyCmp
*
****************************************************************************
*/
FE1(TRANS *, TransCons,
    DOMAIN, source)
{
    TRANS *trans;

    trans = TransAlloc();
    TransSSource(trans, source);
    if (transB)
	TransSBalTargets(trans);
    else
	TransSLisTargets(trans, LstCreate(&domTools));
    return(trans);
}

/***************************************************************************
*
* Function:  TransDracsid
* Purpose:   This routine removes an LTS transition from the memory.
* Arguments: trans - the transition
* Returns:
* See also:  TransCons, TransDump, TransKeyCmp
*
****************************************************************************
*/
FE1(void, TransDracsid,
    TRANS *, trans)
{
    if (trans)
	{
	    if (transB)
		TransDBalTargets(trans);
	    else
		TransDLisTargets(trans);
	    ClDeallocate(trans);
	}
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  TransDump
* Purpose:   This routine displays an LTS transition.
* Arguments: trans - the transition
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  TransCons, TransDiscard, TransKeyCmp
*
****************************************************************************
*/
FE2(void, TransDump,
    TRANS *, trans,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fputs("TRANS source ", _err);
    (void) fprintf(_err, DOMFORMAT, TransGSource(trans));
    (void) fputs("\n", _err);
    DumpIndent(depth);
    (void) fputs("TRANS targets\n", _err);
    if (transB)
	BalTrDump(TransGBalTargets(trans), depth + 1);
    else
	LstDump(TransGLisTargets(trans), depth + 1);
    DumpIndent(depth);
    (void) fputs("End of TRANS\n", _err);
    (void) fflush(_err);
}
#endif

/***************************************************************************
*
* Function:  TransKeyCmp
* Purpose:   This routine compares the source node of an LTS transition
*            to a key.
* Arguments: trans - the transition
*            key - the key
* Returns:   -1 if the source node is less than the key,
*             1 if the source node is greater than the key, and
*             0 otherwise
* See also:  TransCons, TransDiscard, TransDump
*
****************************************************************************
*/
FE2(int, TransKeyCmp,
    TRANS *, trans,
    DOMAIN *, key)
{
    return((TransGSource(trans) < *key) ? -1 :
	   ((TransGSource(trans) > *key) ? 1 : 0));
}

/* A collection of routines for the TRANS type to be used by
 * list routines of dynam.c and table routines of ptrbl.c. */
TOOLS transTools =
{ (CFUNC) NULL, TransDracsid
#ifdef DEBUG
, TransDump
#endif
 };

/* A collection of routines for the TRANS type to be used by
 * tree routines of baltree.c. */
BALTOOLS transBalTools =
{ (CFUNC) NULL, TransDracsid
#ifdef DEBUG
, TransDump
#endif
, TransKeyCmp, (FOOL) NULL };

/***************************************************************************
*
* Function:  DomTrLst
* Purpose:   This routine is called for each target node of an LTS
*            transition in order to collect the target nodes into a list.
* Arguments: n - a target node of an LTS transition
* Returns:
* See also:  TransTrLst, ActTrLst
*
****************************************************************************
*/
FE1(BOOL, DomTrLst,
    DOMAIN *, n)
{
    (void) LstInsLast(domList, (char *) n);
    return(FALSE);
}

/***************************************************************************
*
* Function:  TransTrLst
* Purpose:   This routine is called for each LTS transition of an action
*            in order to collect the transitions into a list. In addition,
*            the presentation of the set of target nodes of an LTS
*            transition is changed from a tree to a list. 
* Arguments: trans - an LTS transition
* Returns:
* See also:  DomTrLst, ActTrLst
*
****************************************************************************
*/
FE1(BOOL, TransTrLst,
    TRANS *, trans)
{
    domList = LstCreate(&domTools);
    (void) BalTrApply(TransGBalTargets(trans), DomTrLst);
    BalTrDestroy(TransGBalTargets(trans), FALSE);
    TransSLisTargets(trans, domList);
    (void) LstInsLast(transList, (char *) trans);
    return(FALSE);
}

/***************************************************************************
*
* Function:  LacCons
* Purpose:   This routine constructs an element of LACTION, consisting
*            of an LTS number and a set of LTS transitions associated
*            with a certain action and the LTS. An element of LACTION
*            thus corresponds to that part of the LTS which is associated
*            with the action. LTS transitions are kept in a tree if lacB is
*            TRUE, and otherwise in a list.
* Arguments: lacSystem - the LTS number
* Returns:   a pointer to the constructed element
* See also:  LacDiscard, LacDump
*
****************************************************************************
*/
FE1(LACTION *, LacCons,
    DOMAIN, lacSystem)
{
    LACTION *laction;

    laction = LacAlloc();
    LacSSystem(laction, lacSystem);
    if (lacB)
	LacSBalTransitions(laction);
    else
	LacSLisTransitions(laction, LstCreate(&transTools));
    return(laction);
}

/***************************************************************************
*
* Function:  LacDracsid
* Purpose:   This routine removes an element of LACTION from the memory.
* Arguments: laction - an element of LACTION
* Returns:
* See also:  LacCons, LacDump
*
****************************************************************************
*/
FE1(void, LacDracsid,
    LACTION *, laction)
{
    if (laction)
	{
	    if (lacB)
		LacDBalTransitions(laction);
	    else
		LacDLisTransitions(laction);
	    ClDeallocate(laction);
	}
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  LacDump
* Purpose:   This routine displays an element of LACTION
* Arguments: laction - an element of LACTION
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  LacCons, LacDiscard
*
****************************************************************************
*/
FE2(void, LacDump,
    LACTION *, laction,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fputs("LACTION syst ", _err);
    (void) fprintf(_err, DOMFORMAT, LacGSystem(laction));
    (void) fputs("\n", _err);
    DumpIndent(depth);
    (void) fputs("LACTION transitions\n", _err);
    if (lacB)
	BalTrDump(LacGBalTransitions(laction), depth + 1);
    else
	LstDump(LacGLisTransitions(laction), depth + 1);
    DumpIndent(depth);
    (void) fputs("End of LACTION\n", _err);
    (void) fflush(_err);
}
#endif

/* A collection of routines for the LACTION type to be used by
 * list routines of dynam.c and table routines of ptrbl.c. */
TOOLS lacTools =
{ (CFUNC) NULL, LacDracsid
#ifdef DEBUG
, LacDump
#endif
 };

/***************************************************************************
*
* Function:  ActCons
* Purpose:   This routine constructs an action, consisting of a number,
*            a name, a hiding flag and a list of references to parts of
*            LTS's associated with the action.
* Arguments: number - the number of the action
*            name - the name of the action
*            hidden - the action is made invisible if hidden is TRUE,
*                     and visible otherwise
* Returns:   a pointer to the action
* See also:  ActDiscard, ActDump, ActKeyCmp, ActNumKeyCmp
*
****************************************************************************
*/
FE3(ACTION *, ActCons,
    unsigned, number,
    char *, name,
    BOOL, hidden)
{
    ACTION *action;

    action = ActAlloc();
    ActSNumber(action, number);
    ActSName(action, name);
    if (hidden)
	ActHide(action);
    else
	ActDisplay(action);
    ActSUsed(action);
    return(action);
}

/***************************************************************************
*
* Function:  ActDracsid
* Purpose:   This routine removes an action from the memory.
* Arguments: action - the action ro be removed
* Returns:
* See also:  ActCons, ActDump, ActKeyCmp, ActNumKeyCmp
*
****************************************************************************
*/
FE1(void, ActDracsid,
    ACTION *, action)
{
    if (action)
	{
	    ClDeallocate(ActGName(action));
	    ActDUsed(action);
	    ClDeallocate(action);
	}
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  ActDump
* Purpose:   This routine displays an action.
* Arguments: action - the action to be displayed
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  ActCons, ActDiscard, ActKeyCmp, ActNumKeyCmp
*
****************************************************************************
*/
FE2(void, ActDump,
    ACTION *, action,
    unsigned, depth)
{
    DumpIndent(depth);
    if (ActHidden(action))
	(void) fprintf(_err, "ACTION number %u name %s hidden\n",
		       ActGNumber(action), ActGName(action));
    else
	(void) fprintf(_err, "ACTION number %u name %s\n",
		       ActGNumber(action), ActGName(action));
    DumpIndent(depth);
    (void) fputs("ACTION used\n", _err);
    LstDump(ActGUsed(action), depth + 1);
    DumpIndent(depth);
    (void) fputs("End of ACTION\n", _err);
    (void) fflush(_err);
}
#endif

/***************************************************************************
*
* Function:  ActKeyCmp
* Purpose:   This routine compares the name of an action lexicographically
*            to a key.
* Arguments: action - the action
*            key - the key
* Returns:   some negative number if the name is less than the key,
*            some positive number if the name is greater than the key, and
*            0 otherwise
* See also:  ActCons, ActDiscard, ActDump, ActNumKeyCmp
*
****************************************************************************
*/
FE2(int, ActKeyCmp,
    ACTION *, action,
    char *, key)
{
    return(strcmp(ActGName(action), key));
}

/***************************************************************************
*
* Function:  ActNumKeyCmp
* Purpose:   This routine compares the number of an action to a key.
* Arguments: action - the action
*            key - the key
* Returns:   -1 if the number of the action is less than the key,
*             1 if the number of the action is greater than the key, and
*             0 otherwise
* See also:  ActCons, ActDiscard, ActDump, ActKeyCmp
*
****************************************************************************
*/
FE2(int, ActNumKeyCmp,
    ACTION *, action,
    unsigned *, key)
{
    return((ActGNumber(action) < *key) ? -1 :
	   ((ActGNumber(action) > *key) ? 1 : 0));
}

/* A collection of routines for the ACTION type to be used by
 * list routines of dynam.c and table routines of ptrbl.c. */
TOOLS actTools =
{ (CFUNC) NULL, ActDracsid
#ifdef DEBUG
, ActDump
#endif
 };

/* A collection of routines for the ACTION type to be used by
 * tree routines of baltree.c. */
BALTOOLS actBalTools =
{ (CFUNC) NULL, ActDracsid
#ifdef DEBUG
, ActDump
#endif
, ActKeyCmp, (FOOL) NULL };

/* Another collection of routines for the ACTION type to be used by
 * the tree routines of baltree.c. */
BALTOOLS actNumBalTools =
{ (CFUNC) NULL, ActDracsid
#ifdef DEBUG
, ActDump
#endif
, ActNumKeyCmp, (FOOL) NULL };

/***************************************************************************
*
* Function:  ActTrLst
* Purpose:   This routine is called for each action in order to collect
*            the actions into a list. In addition, the presentation of
*            the set of LTS transitions of an action is changed from
*            a tree to a list. 
* Arguments: action - an action
* Returns:
* See also:  DomTrLst, TransTrLst
*
****************************************************************************
*/
FE1(BOOL, ActTrLst,
    ACTION *, action)
{
    LACTION *laction;
    VOID *cursor;
    
    LSTFOR (ActGUsed(action), laction, LACTION)
	{
	    transList = LstCreate(&transTools);
	    (void) BalTrApply(LacGBalTransitions(laction), TransTrLst);
	    BalTrDestroy(LacGBalTransitions(laction), FALSE);
	    LacSLisTransitions(laction, transList);
	}
    cursor = NULL;
    LSTFOR (ActGUsed(action), laction, LACTION)
	{
	    if (LstEmpty(LacGLisTransitions(laction)))
		{
		    if (action != IAction())
			{
			    if (ActHidden(action))
				{
				    ActDiscard(action);
				    return(FALSE);
				}
			    ActDUsed(action);
			    ActSUsed(action);
			    (void) LstInsLast(ActList(), (char *) action);
			    return(FALSE);
			}
		    LstSet(ActGUsed(action), cursor);
		    (void) LstDelNext(ActGUsed(action), TRUE);
		}
	    cursor = LstGet(ActGUsed(action));
	}
    if (LstEmpty(ActGUsed(action)) &&
	((action == IAction()) || ActHidden(action)))
	{
	    if (action == IAction())
		SetIAction((ACTION *) NULL);
	    ActDiscard(action);
	    return(FALSE);
	}
    (void) LstInsLast(ActList(), (char *) action);
    return(FALSE);
}

/***************************************************************************
*
* Function:  SysotCons
* Purpose:   This routine constructs a flow component, consisting of
*            an LTS number, an LTS transition and a target node of the
*            transition.
* Arguments: sysoSystem - the LTS number
*            trans - the LTS transition
*            target - the target node
* Returns:   a pointer to the constructed flow component
* See also:  SysotCopy, SysotDump, SysotCompare
*
****************************************************************************
*/
FE3(SYSOT *, SysotCons,
    DOMAIN, sysoSystem,
    TRANS *, trans,
    DOMAIN, target)
{
    SYSOT *sysot;

    sysot = SysotAlloc();
    SysotSSystem(sysot, sysoSystem);
    SysotSTrans(sysot, trans);
    SysotSTarget(sysot, target);
    return(sysot);
}

/***************************************************************************
*
* Function:  SysotCopy
* Purpose:   This routine makes a copy of a flow component.
* Arguments: sysot - the original flow component
* Returns:   a pointer to the copy
* See also:  SysotCons, SysotDump, SysotCompare
*
****************************************************************************
*/
FE1(SYSOT *, SysotCopy,
    SYSOT *, sysot)
{
    SYSOT *copy;

    copy = SysotAlloc();
    SysotSSystem(copy, SysotGSystem(sysot));
    SysotSTrans(copy, SysotGTrans(sysot));
    SysotSTarget(copy, SysotGTarget(sysot));
    return(copy);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  SysotDump
* Purpose:   This routine displays a flow component.
* Arguments: sysot - the flow component
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  SysotCons, SysotCopy, SysotCompare
*
****************************************************************************
*/
FE2(void, SysotDump,
    SYSOT *, sysot,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fputs("SYSOT syst ", _err);
    (void) fprintf(_err, DOMFORMAT, SysotGSystem(sysot));
    (void) fputs(" target ", _err);
    (void) fprintf(_err, DOMFORMAT, SysotGTarget(sysot));
    (void) fputs("\n", _err);
    DumpIndent(depth);
    (void) fputs("SYSOT trans\n", _err);
    if (SysotGTrans(sysot))
	TransDump(SysotGTrans(sysot), depth + 1);
    else
	(void) fflush(_err);
}
#endif

/***************************************************************************
*
* Function:  SysotCompare
* Purpose:   This routine compares two flow components to each other.
*            The components are first compared with respect to their
*            LTS numbers, then, if necessary, with respect to the source
*            nodes of their LTS transitions, and finally, if necessary,
*            with respect to their target nodes.
* Arguments: sysot1 - a flow component
*            sysot2 - a flow component
* Returns:   -1 if sysot1 is less than sysot2,
*             1 if sysot1 is greater than sysot2, and
*             0 otherwise
* See also:  SysotCons, SysotCopy, SysotDump
*
****************************************************************************
*/
FE2(int, SysotCompare,
    SYSOT *, sysot1,
    SYSOT *, sysot2)
{
    if (SysotGSystem(sysot1) < SysotGSystem(sysot2))
	return(-1);
    if (SysotGSystem(sysot1) > SysotGSystem(sysot2))
	return(1);
    if ((!SysotGTrans(sysot1)) && SysotGTrans(sysot2))
	return(-1);
    if (SysotGTrans(sysot1) && !SysotGTrans(sysot2))
	return(1);
    if (SysotGTrans(sysot1) && SysotGTrans(sysot2))
	{
	    if (TransGSource(SysotGTrans(sysot1)) <
		TransGSource(SysotGTrans(sysot2)))
		return(-1);
	    if (TransGSource(SysotGTrans(sysot1)) >
		TransGSource(SysotGTrans(sysot2)))
		return(1);
	}
    if (SysotGTarget(sysot1) < SysotGTarget(sysot2))
	return(-1);
    if (SysotGTarget(sysot1) > SysotGTarget(sysot2))
	return(1);
    return(0);
}

/* A collection of routines for the SYSOT type to be used by
 * list routines of dynam.c and table routines of ptrbl.c. */
TOOLS sysotTools =
{ (CFUNC) SysotCopy, Deallocate
#ifdef DEBUG
, SysotDump
#endif
 };

/***************************************************************************
*
* Function:  FlowCompare
* Purpose:   This routine compares two flows, i.e. lists of flow
*            components, lexicographically to each other.
* Arguments: flow1 - a flow
*            flow2 - a flow
* Returns:   -1 if flow1 is less than flow2,
*             1 if flow1 is greater than flow2, and
*             0 otherwise
* See also:  SysotCompare
*
****************************************************************************
*/
FE2(int, FlowCompare,
    LIST *, flow1,
    LIST *, flow2)
{
    SYSOT *sysot1;
    SYSOT *sysot2;
    int cmpResult;

    if (flow1 == flow2)
	return(0);
    sysot1 = (SYSOT *) LstBegin(flow1);
    sysot2 = (SYSOT *) LstBegin(flow2);
    while (sysot1 && sysot2)
	{
	    if ((cmpResult = SysotCompare(sysot1, sysot2)))
		return(cmpResult);
	    sysot1 = (SYSOT *) LstStep(flow1);
	    sysot2 = (SYSOT *) LstStep(flow2);
	}
    return(sysot1 ? 1 : (sysot2 ? -1 : 0));
}

/* A collection of routines for flows to be used by
 * tree routines of baltree.c. */
BALTOOLS flowBalTools =
{ (CFUNC) LstCp, LstDc
#ifdef DEBUG
, LstDump
#endif
, FlowCompare, (FOOL) NULL };

/***************************************************************************
*
* Function:  HiTrNmOutput
* Purpose:   This routine writes the first line of a high level transition
*            declaration to a net description file.
* Arguments: name - the name of some action, forms the basic part of the
*                   name of the high level transition
*            ind - if the action is invisible, this index may have to
*                  be added to the name of the high level transition
* Returns:
* See also:  HiTrNmWrite, HiVarWrite
*
****************************************************************************
*/
FE2(void, HiTrNmOutput,
    char *, name,
    unsigned, ind)
{
    char *s;

    if (ind != 0) 
	{
	    NetOutput(netFile, "#trans %s_%u\n", name, ind);
	    return;
	}
    s = name + strlen(name);
    if (!isdigit((UCHAR) *--s))
	{
	    NetOutput(netFile, "#trans %s\n", name);
	    return;
	}
    while (isdigit((UCHAR) *--s))
	;
    if (*s == '_')
	{ NetOutput(netFile, "#trans %s_0\n", name); }
    else
	{ NetOutput(netFile, "#trans %s\n", name); }
}

/***************************************************************************
*
* Function:  HiTrNmWrite
* Purpose:   This routine writes the name of a high level transition
*            to a binary data file.
* Arguments: name - the name of some action, forms the basic part of the
*                   name of the high level transition
*            ind - if the action is invisible, this index may have to
*                  be added to the name of the high level transition
*            bFile - the file
* Returns:
* See also:  HiTrNmOutput, HiVarWrite, DatOutput
*
****************************************************************************
*/
FE3(void, HiTrNmWrite,
    char *, name,
    unsigned, ind,
    BINARYFILE, bFile)
{
    char *s;
    unsigned sLen, i;

    sLen = (unsigned) strlen(name);
    if (ind != 0) 
	{
	    sLen++;
	    i = ind;
	    while (i != 0)
		{
		    sLen++;
		    i = i / 10;
		}
	    WriteUns(&sLen, 1, bFile);
	    s = (char *) Allocate(sLen + 1);
	    (void) sprintf(s, "%s_%u", name, ind);
	    WriteChr(s, sLen, bFile);
	    ClDeallocate(s);
	    return;
	}
    s = name + sLen;
    if (!isdigit((UCHAR) *--s))
	{
	    WriteUns(&sLen, 1, bFile);
	    WriteChr(name, sLen, bFile);
	    return;
	}
    while (isdigit((UCHAR) *--s))
	;
    if (*s != '_')
	{
	    WriteUns(&sLen, 1, bFile);
	    WriteChr(name, sLen, bFile);
	    return;
	}
    sLen += 2;
    WriteUns(&sLen, 1, bFile);
    s = (char *) Allocate(sLen + 1);
    (void) sprintf(s, "%s_0", name);
    WriteChr(s, sLen, bFile);
    ClDeallocate(s);
}

/***************************************************************************
*
* Function:  HiVarWrite
* Purpose:   This routine writes the name of a variable of a high level
*            transition to a binary data file.
* Arguments: name - the basic part of the name of the variable
*            ind - this index is the LTS number corresponding to the
*                  variable and may have to be added to the name of
*                  the variable
*            bFile - the file
* Returns:
* See also:  HiTrNmOutput, HiTrNmWrite, DatOutput
*
****************************************************************************
*/
FE3(void, HiVarWrite,
    char *, name,
    unsigned long, ind,
    BINARYFILE, bFile)
{
    char *s;
    unsigned sLen;
    unsigned long a;

    sLen = (unsigned) strlen(name);
    if (ind == 0L)
	sLen++;
    else
	{
	    a = ind;
	    while (a != 0L)
		{
		    sLen++;
		    a = a / (unsigned long) 10;
		}
	}
    WriteUns(&sLen, 1, bFile);
    s = (char *) Allocate(sLen + 1);
    (void) sprintf(s, "%s%lu", name, ind);
    WriteChr(s, sLen, bFile);
    ClDeallocate(s);
}

static unsigned long monPos;  /* a position in a binary data file */
static DATATYPE *monBuffer;   /* a buffer for encoded information */
static unsigned monPl;        /* the number of a place in a high level net */
static unsigned monLen;       /* when used, tells how many bytes are to
			       * be written in a single write operation */
static BINARYFILE datFile = NonBinaryFile();    /* a binary data file */

/***************************************************************************
*
* Function:  PlcMon
* Purpose:   If the argument place corresponds to some monitor state of
*            a tester, this routine writes information about the monitor
*            state to a binary data file.
* Arguments: plc - a place in a low level net
* Returns:
* See also:  DatOutput, pplst_dat (pplst.c), MonOut (pplst.c)
*
****************************************************************************
*/
FE1(BOOL, PlcMon,
    PLC *, plc)
{
    register unsigned i, len;
    register UDOMAIN value;
    register DATATYPE *pointer;

    if (PlcMonitor(plc))
	{
	    arpDouble = TupAlloc();
	    TupDecode(&monPl, PlcGKey(plc), arpDouble);
	    len = 0;
	    value = TupGField(arpDouble, (unsigned) 0);
	    value = Encode((DOMAIN) value);
	    IncLen(value, len);
	    i = PlcMonitor(plc) >> 2;
	    IncLen(i, len);
	    monLen = len;
	    WriteUns(&monLen, 1, datFile);
	    monPos += (unsigned long) sizeof(unsigned);
	    monBuffer = (DATATYPE *) AllocArray(monLen, sizeof(DATATYPE));
	    pointer = monBuffer;
	    value = TupGField(arpDouble, (unsigned) 0);
	    TupDiscard(arpDouble);
	    value = Encode((DOMAIN) value);
	    PutValue(value, pointer);
	    i = PlcMonitor(plc) >> 2;
	    PutValue(i, pointer);
	    WriteData(monBuffer, monLen, datFile);
	    monPos += (unsigned long) (monLen * sizeof(DATATYPE));
	    ClDeallocate(monBuffer);
	}
    return(FALSE);
}

/***************************************************************************
*
* Function:  DatOutput
* Purpose:   This routine writes information about a high level net
*            to a binary data file.
* Arguments: bFile - the file
* Returns:
* See also:  HiTrNmWrite, HiVarWrite, PlcMon, pplst_dat (pplst.c)
*
****************************************************************************
*/
FE1(void, DatOutput,
    BINARYFILE, bFile)
{
    unsigned p, sLen;
    unsigned long pos;
    DOMAIN systNumber, *root;
    DATATYPE *buf;
    register DATATYPE *pointer;
    register unsigned i, len;
    register UDOMAIN value;

    pos = 0L;
    WriteLng(&pos, 1, bFile);
    pos += (unsigned long) sizeof(unsigned long);
    if (OnTheFly())
	p = MonPlInd();
    else
	p = (unsigned) -1;
    WriteUns(&p, 1, bFile);
    pos += (unsigned long) sizeof(unsigned);
    p = 1;
    WriteUns(&p, 1, bFile);
    pos += (unsigned long) sizeof(unsigned);
    len = 0;
    i = MagGNofPlaces();
    IncLen(i, len);
    for (p = 0; p < MagGNofPlaces(); p++)
	{
	    i = MagGProfile(p);
	    IncLen(i, len);
	}
    sLen = len;
    WriteUns(&sLen, 1, bFile);
    pos += (unsigned long) sizeof(unsigned);
    buf = (DATATYPE *) AllocArray(sLen, sizeof(DATATYPE));
    pointer = buf;
    i = MagGNofPlaces();
    PutValue(i, pointer);
    for (p = 0; p < MagGNofPlaces(); p++)
	{
	    i = MagGProfile(p);
	    PutValue(i, pointer);
	}
    WriteData(buf, sLen, bFile);
    pos += (unsigned long) (sLen * sizeof(DATATYPE));
    ClDeallocate(buf);
    len = 0;
    i = IngGNofTransitions();
    IncLen(i, len);
    for (p = 0; p < IngGNofTransitions(); p++)
	{
	    i = IngGProfile(p);
	    IncLen(i, len);
	}
    sLen = len;
    WriteUns(&sLen, 1, bFile);
    pos += (unsigned long) sizeof(unsigned);
    buf = (DATATYPE *) AllocArray(sLen, sizeof(DATATYPE));
    pointer = buf;
    i = IngGNofTransitions();
    PutValue(i, pointer);
    for (p = 0; p < IngGNofTransitions(); p++)
	{
	    i = IngGProfile(p);
	    PutValue(i, pointer);
	}
    WriteData(buf, sLen, bFile);
    pos += (unsigned long) (sLen * sizeof(DATATYPE));
    ClDeallocate(buf);
    len = 0;
    for (p = 0; p < IngGNofTransitions(); p++)
	{
	    i = 0;
	    IncLen(i, len);
	}
    sLen = len;
    WriteUns(&sLen, 1, bFile);
    pos += (unsigned long) sizeof(unsigned);
    buf = (DATATYPE *) AllocArray(sLen, sizeof(DATATYPE));
    pointer = buf;
    for (p = 0; p < IngGNofTransitions(); p++)
	{
	    i = 0;
	    PutValue(i, pointer);
	}
    WriteData(buf, sLen, bFile);
    pos += (unsigned long) (sLen * sizeof(DATATYPE));
    ClDeallocate(buf);
    sLen = 0;
    for (p = 0; p < IngGNofTransitions(); p++)
	WriteUns(&sLen, 1, bFile);
    pos += (unsigned long) (IngGNofTransitions() * sizeof(unsigned));
    if (RootList())
	{
	    if ((!OnTheFly()) || (tstrSys != (DOMAIN) 0L))
		{
		    systNumber = (DOMAIN) 0L;
		    LSTFOR (RootList(), root, DOMAIN)
			{
			    if (OnTheFly() && !LstNext(RootList()))
				break;
			    len = 0;
			    i = 0;
			    IncLen(i, len);
			    value = Encode(systNumber);
			    IncLen(value, len);
			    value = Encode(*root);
			    IncLen(value, len);
			    sLen = len;
			    WriteUns(&sLen, 1, bFile);
			    pos += (unsigned long) sizeof(unsigned);
			    buf = (DATATYPE *) AllocArray(sLen,
							  sizeof(DATATYPE));
			    pointer = buf;
			    i = 0;
			    PutValue(i, pointer);
			    value = Encode(systNumber);
			    PutValue(value, pointer);
			    value = Encode(*root);
			    PutValue(value, pointer);
			    WriteData(buf, sLen, bFile);
			    pos += (unsigned long) (sLen * sizeof(DATATYPE));
			    ClDeallocate(buf);
			    systNumber++;
			}
		}
	    if (OnTheFly())
		{
		    root = (DOMAIN *) LstLast(RootList());
		    len = 0;
		    i = MonPlInd();
		    IncLen(i, len);
		    value = Encode(*root);
		    IncLen(value, len);
		    sLen = len;
		    WriteUns(&sLen, 1, bFile);
		    pos += (unsigned long) sizeof(unsigned);
		    buf = (DATATYPE *) AllocArray(sLen,
						  sizeof(DATATYPE));
		    pointer = buf;
		    i = MonPlInd();
		    PutValue(i, pointer);
		    value = Encode(*root);
		    PutValue(value, pointer);
		    WriteData(buf, sLen, bFile);
		    pos += (unsigned long) (sLen * sizeof(DATATYPE));
		    ClDeallocate(buf);
		}
	}
    sLen =  (unsigned) -1;
    WriteUns(&sLen, 1, bFile);
    pos += (unsigned long) sizeof(unsigned);
    if (OnTheFly())
	{
	    monPos = pos;
	    datFile = bFile;
	    (void) BalTrApply(PlcBalTree(), PlcMon);
	    pos = monPos;
	    sLen = (unsigned) -1;
	    WriteUns(&sLen, 1, bFile);
	    pos += (unsigned long) sizeof(unsigned);
	    sLen = 0;
	    WriteUns(&sLen, 1, bFile);
	    pos += (unsigned long) sizeof(unsigned);
	}
    if (RootList())
	{
	    arpHiTr = 0;
	    if (IAction())
		{
		    if ((!OnTheFly()) ||
			(LacGSystem((LACTION *) LstFirst(ActGUsed(IAction())))
			 != tstrSys))
			{
			    HiTrNmWrite(IName(), 0, bFile);
			    sLen = 1;
			    WriteUns(&sLen, 1, bFile);
			    WriteChr("x", 1, bFile);
			    WriteUns(&sLen, 1, bFile);
			    WriteChr("y", 1, bFile);
			    WriteUns(&sLen, 1, bFile);
			    WriteChr("z", 1, bFile);
			    arpHiTr = 1;
			}
		    if (OnTheFly() &&
			(LacGSystem((LACTION *) LstLast(ActGUsed(IAction())))
			 == tstrSys))
			{
			    HiTrNmWrite(IName(), (unsigned)
					((LstFirst(ActGUsed(IAction())) !=
					  LstLast(ActGUsed(IAction()))) ? 1 : 0),
					bFile);
			    sLen = 1;
			    WriteUns(&sLen, 1, bFile);
			    WriteChr("y", 1, bFile);
			    WriteUns(&sLen, 1, bFile);
			    WriteChr("z", 1, bFile);
			    arpHiTr++;
			}
		}
	    LSTFOR (ActList(), arpAction, ACTION)
		{
		    if ((arpAction == IAction()) || !ActHidden(arpAction))
			continue;
		    HiTrNmWrite(IName(), arpHiTr, bFile);
		    LSTFOR (ActGUsed(arpAction), arpLaction, LACTION)
			HiVarWrite("y", (unsigned long) LacGSystem(arpLaction),
				   bFile);
		    LSTFOR (ActGUsed(arpAction), arpLaction, LACTION)
			HiVarWrite("z", (unsigned long) LacGSystem(arpLaction),
				   bFile);
		    arpHiTr++;
		}
	    LSTFOR (ActList(), arpAction, ACTION)
		{
		    if ((arpAction == IAction()) || ActHidden(arpAction))
			continue;
		    HiTrNmWrite(ActGName(arpAction), (unsigned) 0, bFile);
		    LSTFOR (ActGUsed(arpAction), arpLaction, LACTION)
			HiVarWrite("y", (unsigned long) LacGSystem(arpLaction),
				   bFile);
		    LSTFOR (ActGUsed(arpAction), arpLaction, LACTION)
			HiVarWrite("z", (unsigned long) LacGSystem(arpLaction),
				   bFile);
		}
	}
    p = 0;
    WriteUns(&p, 1, bFile);
    if (RootList())
	{
	    if ((!OnTheFly()) || (tstrSys != (DOMAIN) 0L))
		{
		    sLen = 1;
		    WriteUns(&sLen, 1, bFile);
		    WriteChr("p", 1, bFile);
		}
	    if (OnTheFly())
		{
		    sLen = 1;
		    WriteUns(&sLen, 1, bFile);
		    WriteChr("q", 1, bFile);
		}
	}
    SeekAbs(bFile, (unsigned long) 0L);
    WriteLng(&pos, 1, bFile);
}

/***************************************************************************
*
* Function:  TauInstances
* Purpose:   This routine constructs low and high level transitions
*            corresponding to an invisible action.
* Arguments:
* Returns:
* See also:  TstrTauInstances, Instances
*
****************************************************************************
*/
FE0(void, TauInstances)
{
    arpCounter = 0L;
    goEnd = FALSE;
    HiTrNmOutput(IName(), (unsigned) 0);
    NetOutput(netFile,
	      "    in  { p: <.x, y.>; }\n    out { p: <.x, z.>; }\n    gate ");
    LSTFOR (ActGUsed(IAction()), arpLaction, LACTION)
	{
	    if (OnTheFly() && (LacGSystem(arpLaction) == tstrSys))
		break;
	    if (arpLaction != (LACTION *) LstFirst(ActGUsed(IAction())))
		{
		    NetOutput(netFile, " || ");
		}
	    NetOutput(netFile, "(x == ");
	    NetOutput(netFile, DOMFORMAT, LacGSystem(arpLaction));
	    NetOutput(netFile, ")");
	}
    NetOutput(netFile, ";\n    comp {");
    if (!OnTheFly())
	{
	    NetOutput(netFile, " Invisible();");
	}
    NetOutput(netFile, "\n");
    LSTFOR (ActGUsed(IAction()), arpLaction, LACTION)
	{
	    if (OnTheFly() && (LacGSystem(arpLaction) == tstrSys))
		break;
	    if (LstNext(ActGUsed(IAction())))
		{
		    NetOutput(netFile, "        if (x != ");
		    NetOutput(netFile, DOMFORMAT, LacGSystem(arpLaction));
		    NetOutput(netFile, ") goto g%lu;\n", arpCounter);
		}
	    TupSField(arpDouble, (unsigned) 0, LacGSystem(arpLaction));
	    (void) LstInsLast(sysotList, (char *)
			       SysotCons(LacGSystem(arpLaction),
					 (TRANS *) NULL, (DOMAIN) MINDOMAIN));
	    arpVars[0] = LacGSystem(arpLaction);
	    LSTFOR (LacGLisTransitions(arpLaction), arpTrans, TRANS)
		{
		    NetOutput(netFile, "        if (y == ");
		    NetOutput(netFile, DOMFORMAT, TransGSource(arpTrans));
		    NetOutput(netFile, ") {");
		    TupSField(arpDouble, (unsigned) 1, TransGSource(arpTrans));
		    (void) TupEncode((unsigned) 0, arpDouble,
				     EngABuffer(), EngABufSize());
		    arpPlc = GetPlc(EngBuffer());
		    SysotSTrans((SYSOT *) LstLast(sysotList), arpTrans);
		    arpVars[1] = TransGSource(arpTrans);
		    LSTFOR (TransGLisTargets(arpTrans),
			    arpTarget, DOMAIN)
			{
			    SysotSTarget((SYSOT *) LstLast(sysotList),
					 *arpTarget);
			    sysot2List = LstCopy(sysotList, TRUE);
			    if (BalTrInsert(flowTree,
					    (char *) sysot2List,
					    (char *) sysot2List, TRUE))
				Troba2("%s, internal error: unexpected duplicates",
				       myName);
			    NetOutput(netFile, " z = ");
			    NetOutput(netFile, DOMFORMAT, *arpTarget);
			    NetOutput(netFile, "; Accept();");
			    arpVars[2] = *arpTarget;
			    (void) InsEncode((unsigned) 0, (unsigned) 0,
					     arpVars, EngABuffer(),
					     EngABufSize()
#ifdef WEIRD
					     , (unsigned) 0
#endif
					     );
			    arpTrn = GetTrn((unsigned) 0,
					    EngBuffer(), &theNovelty);
			    if (TransGSource(arpTrans) == *arpTarget)
				{
				    if (!TrnGNeutral(arpTrn))
					TrnSNeutral(arpTrn);
				    (void) LstInsLast(TrnGNeutral(arpTrn),
						      (char *)
						      TrArcCons(arpPlc,
								(unsigned long)
								1L));
				    if (!PlcGNeutral(arpPlc))
					PlcSNeutral(arpPlc);
				    (void) LstInsLast(PlcGNeutral(arpPlc),
						      (char *)
						      PlArcCons(arpTrn,
								(unsigned long)
								1L));
				}
			    else
				{
				    if (!TrnGPurIn(arpTrn))
					TrnSPurIn(arpTrn);
				    (void) LstInsLast(TrnGPurIn(arpTrn),
						      (char *)
						      TrArcCons(arpPlc,
								(unsigned long)
								1L));
				    if (!PlcGPurOut(arpPlc))
					PlcSPurOut(arpPlc);
				    (void) LstInsLast(PlcGPurOut(arpPlc),
						      (char *)
						      PlArcCons(arpTrn,
								(unsigned long)
								1L));
				    TupSField(arpDouble, (unsigned) 1,
					      *arpTarget);
				    (void) TupEncode((unsigned) 0, arpDouble,
						     EngABuffer(),
						     EngABufSize());
				    arp2Plc = GetPlc(EngBuffer());
				    if (!TrnGPurOut(arpTrn))
					TrnSPurOut(arpTrn);
				    (void) LstInsLast(TrnGPurOut(arpTrn),
						      (char *)
						      TrArcCons(arp2Plc,
								(unsigned long)
								1L));
				    if (!PlcGPurIn(arp2Plc))
					PlcSPurIn(arp2Plc);
				    (void) LstInsLast(PlcGPurIn(arp2Plc),
						      (char *)
						      PlArcCons(arpTrn,
								(unsigned long)
								1L));
				}
			}
		    if (LstNext(LacGLisTransitions(arpLaction)))
			{
			    NetOutput(netFile, " goto end; }\n");
			    goEnd = TRUE;
			}
		    else
			NetOutput(netFile, " }\n");
		}
	    if (LstNext(ActGUsed(IAction())))
		{
		    NetOutput(netFile, "        goto end;\n    g%lu: ;\n",
			      arpCounter++);
		    goEnd = TRUE;
		}
	    (void) LstDelLast(sysotList, TRUE);
	}
    if (goEnd)
	{ NetOutput(netFile, "    end: ; }\n#endtr\n"); }
    else
	{ NetOutput(netFile, "    }\n#endtr\n"); }
}

/***************************************************************************
*
* Function:  TstrTauInstances
* Purpose:   This routine constructs low and high level transitions
*            corresponding to an internal action of a tester.
* Arguments:
* Returns:
* See also:  TauInstances, Instances
*
****************************************************************************
*/
FE0(void, TstrTauInstances)
{
    arpCounter = 0L;
    goEnd = FALSE;
    HiTrNmOutput(IName(), (unsigned)
		 ((LstFirst(ActGUsed(IAction())) !=
		   LstLast(ActGUsed(IAction()))) ? 1 : 0));
    NetOutput(netFile,
	      "    in  { q: <.y.>; }\n    out { q: <.z.>; }\n    comp {\n");
    arpLaction = (LACTION *) LstLast(ActGUsed(IAction()));
    (void) LstInsLast(sysotList, (char *)
		      SysotCons(LacGSystem(arpLaction),
				(TRANS *) NULL, (DOMAIN) MINDOMAIN));
    LSTFOR (LacGLisTransitions(arpLaction), arpTrans, TRANS)
	{
	    NetOutput(netFile, "        if (y == ");
	    NetOutput(netFile, DOMFORMAT, TransGSource(arpTrans));
	    NetOutput(netFile, ") {");
	    TupSField(arpSingle, (unsigned) 0, TransGSource(arpTrans));
	    (void) TupEncode((unsigned) 0, arpSingle,
			     EngABuffer(), EngABufSize());
	    arpPlc = GetPlc(EngBuffer());
	    SysotSTrans((SYSOT *) LstLast(sysotList), arpTrans);
	    arpVars[0] = TransGSource(arpTrans);
	    LSTFOR (TransGLisTargets(arpTrans), arpTarget, DOMAIN)
		{
		    SysotSTarget((SYSOT *) LstLast(sysotList),
				 *arpTarget);
		    sysot2List = LstCopy(sysotList, TRUE);
		    if (BalTrInsert(flowTree,
				    (char *) sysot2List,
				    (char *) sysot2List, TRUE))
			Troba2("%s, internal error: unexpected duplicates",
			       myName);
		    NetOutput(netFile, " z = ");
		    NetOutput(netFile, DOMFORMAT, *arpTarget);
		    NetOutput(netFile, "; Accept();");
		    arpVars[1] = *arpTarget;
		    (void) InsEncode(MonPlInd(), (unsigned) 0,
				     arpVars, EngABuffer(),
				     EngABufSize()
#ifdef WEIRD
				     , (unsigned) 0
#endif
				     );
		    arpTrn = GetTrn((unsigned) 0, EngBuffer(), &theNovelty);
		    if (TransGSource(arpTrans) == *arpTarget)
			{
			    if (!TrnGNeutral(arpTrn))
				TrnSNeutral(arpTrn);
			    (void) LstInsLast(TrnGNeutral(arpTrn),
					      (char *)
					      TrArcCons(arpPlc,
							(unsigned long)
							1L));
			    if (!PlcGNeutral(arpPlc))
				PlcSNeutral(arpPlc);
			    (void) LstInsLast(PlcGNeutral(arpPlc),
					      (char *)
					      PlArcCons(arpTrn,
							(unsigned long)
							1L));
			}
		    else
			{
			    if (!TrnGPurIn(arpTrn))
				TrnSPurIn(arpTrn);
			    (void) LstInsLast(TrnGPurIn(arpTrn),
					      (char *)
					      TrArcCons(arpPlc,
							(unsigned long)
							1L));
			    if (!PlcGPurOut(arpPlc))
				PlcSPurOut(arpPlc);
			    (void) LstInsLast(PlcGPurOut(arpPlc),
					      (char *)
					      PlArcCons(arpTrn,
							(unsigned long)
							1L));
			    TupSField(arpSingle, (unsigned) 0, *arpTarget);
			    (void) TupEncode(MonPlInd(), arpSingle,
					     EngABuffer(), EngABufSize());
			    if (!TrnGPurOut(arpTrn))
				TrnSPurOut(arpTrn);
			    (void) LstInsLast(TrnGPurOut(arpTrn),
					      (char *)
					      TrArcCons(arp2Plc,
							(unsigned long)
							1L));
			    if (!PlcGPurIn(arp2Plc))
				PlcSPurIn(arp2Plc);
			    (void) LstInsLast(PlcGPurIn(arp2Plc),
					      (char *)
					      PlArcCons(arpTrn,
							(unsigned long)
							1L));
			}
		}
	    if (LstNext(LacGLisTransitions(arpLaction)))
		{
		    NetOutput(netFile, " goto end; }\n");
		    goEnd = TRUE;
		}
	    else
		NetOutput(netFile, " }\n");
	}
    (void) LstDelLast(sysotList, TRUE);
    if (goEnd)
	{ NetOutput(netFile, "    end: ; }\n#endtr\n"); }
    else
	{ NetOutput(netFile, "    }\n#endtr\n"); }
}

/***************************************************************************
*
* Function:  Accepted
* Purpose:   This routine takes care of the setting of the output variables
*            in a high level transition and completes the construction
*            of the corresponding low level transitions.
* Arguments: sysotCursor - the cursor of sysotList as expected by the caller
* Returns:
* See also:  Instances
*
****************************************************************************
*/
FE1(void, Accepted,
    VOID *, sysotCursor)
{
    if (LstNext(sysotList))
	{
	    LSTFOR (TransGLisTargets(SysotGTrans((SYSOT *)
						 LstThis(sysotList))),
		    arpTarget, DOMAIN)
		{
		    SysotSTarget((SYSOT *) LstThis(sysotList), *arpTarget);
		    NetOutput(netFile, " z%lu = ",
			      (unsigned long)
			      SysotGSystem((SYSOT *) LstThis(sysotList)));
		    NetOutput(netFile, DOMFORMAT, *arpTarget);
		    NetOutput(netFile, ";");
		    arpVars[arpIndex] = *arpTarget;
		    (void) LstStep(sysotList);
		    arpIndex++;
		    Accepted(LstGet(sysotList));
		    arpIndex--;
		    LstSet(sysotList, sysotCursor);
		}
	    return;
	}
    LSTFOR (TransGLisTargets(SysotGTrans((SYSOT *) LstLast(sysotList))),
	    arpTarget, DOMAIN)
	{
	    SysotSTarget((SYSOT *) LstLast(sysotList), *arpTarget);
	    if (ActHidden(arpAction))
		{
		    sysot2List = LstCopy(sysotList, TRUE);
		    if (BalTrInsert(flowTree, (char *) sysot2List,
				    (char *) sysot2List, TRUE))
			{
			    LstDestroy(sysot2List, TRUE);
			    continue;
			}
		}
	    NetOutput(netFile, " z%lu = ",
		      (unsigned long)
		      SysotGSystem((SYSOT *) LstLast(sysotList)));
	    NetOutput(netFile, DOMFORMAT, *arpTarget);
	    NetOutput(netFile, "; Accept();");
	    arpVars[arpIndex] = *arpTarget;
	    (void) InsEncode(arpHiTr, (unsigned) 0, arpVars,
			     EngABuffer(), EngABufSize()
#ifdef WEIRD
			     , (unsigned) 0
#endif
			     );
	    arpTrn = GetTrn((unsigned) 0, EngBuffer(), &theNovelty);
	    if (((!OnTheFly()) && !ActHidden(arpAction)) ||
		(OnTheFly() &&
		 (LacGSystem((LACTION *) LstLast(ActGUsed(arpAction)))
		  == tstrSys)))
		TrnSVisible(arpTrn);
	    LSTFOR (sysotList, arpSysot, SYSOT)
		{
		    if (OnTheFly() &&
			(SysotGSystem(arpSysot) == tstrSys))
			{
			    TupSField(arpSingle, (unsigned) 0,
				      TransGSource(SysotGTrans(arpSysot)));
			    (void) TupEncode(MonPlInd(), arpSingle,
					     EngABuffer(), EngABufSize());
			}
		    else
			{
			    TupSField(arpDouble, (unsigned) 0,
				      SysotGSystem(arpSysot));
			    TupSField(arpDouble, (unsigned) 1,
				      TransGSource(SysotGTrans(arpSysot)));
			    (void) TupEncode((unsigned) 0, arpDouble,
					     EngABuffer(), EngABufSize());
			}
		    arpPlc = GetPlc(EngBuffer());
		    if (TransGSource(SysotGTrans(arpSysot)) ==
			SysotGTarget(arpSysot))
			{
			    if (!TrnGNeutral(arpTrn))
				TrnSNeutral(arpTrn);
			    (void) LstInsLast(TrnGNeutral(arpTrn),
					      (char *)
					      TrArcCons(arpPlc,
							(unsigned long) 1L));
			    if (!PlcGNeutral(arpPlc))
				PlcSNeutral(arpPlc);
			    (void) LstInsLast(PlcGNeutral(arpPlc),
					      (char *)
					      PlArcCons(arpTrn,
							(unsigned long) 1L));
			}
		    else
			{
			    if (!TrnGPurIn(arpTrn))
				TrnSPurIn(arpTrn);
			    (void) LstInsLast(TrnGPurIn(arpTrn),
					      (char *)
					      TrArcCons(arpPlc,
							(unsigned long) 1L));
			    if (!PlcGPurOut(arpPlc))
				PlcSPurOut(arpPlc);
			    (void) LstInsLast(PlcGPurOut(arpPlc),
					      (char *)
					      PlArcCons(arpTrn,
							(unsigned long) 1L));
			    if (OnTheFly() &&
				(SysotGSystem(arpSysot) == tstrSys))
				{
				    TupSField(arpSingle, (unsigned) 0,
					      SysotGTarget(arpSysot));
				    (void) TupEncode(MonPlInd(), arpSingle,
						     EngABuffer(),
						     EngABufSize());
				}
			    else
				{
				    TupSField(arpDouble, (unsigned) 1,
					      SysotGTarget(arpSysot));
				    (void) TupEncode((unsigned) 0, arpDouble,
						     EngABuffer(),
						     EngABufSize());
				}
			    arp2Plc = GetPlc(EngBuffer());
			    if (!TrnGPurOut(arpTrn))
				TrnSPurOut(arpTrn);
			    (void) LstInsLast(TrnGPurOut(arpTrn),
					      (char *)
					      TrArcCons(arp2Plc,
							(unsigned long) 1L));
			    if (!PlcGPurIn(arp2Plc))
				PlcSPurIn(arp2Plc);
			    (void) LstInsLast(PlcGPurIn(arp2Plc),
					      (char *)
					      PlArcCons(arpTrn,
							(unsigned long) 1L));
			}
		}
	}
}
