/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  baltree.c
*
* Part of: PROD's programs araprod, "gphgen", probe, prpp, and strong
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     This module contains general routines for AVL balanced binary
*     trees. As usual, the key of a left-hand child of a vertex
*     is less than or equal to the key of the vertex, whereas the key
*     of a right-hand child of a vertex is greater than or equal to
*     the key of the vertex. If the official keys are not unique,
*     then unique internal keys are obtained by using the address
*     of an element as a secondary key. The AVL criterion is as
*     follows: for each vertex, the heights of child subtrees
*     either are equal or differ from each other by one.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "baltree.h"

#define AvlAlloc() ((AVL *) Allocate(sizeof(AVL)))
#define AvlGElement(avl) ((avl)->element)
#define AvlAElement(avl) (&((avl)->element))
#define AvlSElement(avl,_element) ((avl)->element = (_element))
#define AvlGBalance(avl) ((avl)->balance)
#define AvlSBalance(avl,_balance) ((avl)->balance = (_balance))
#define AvlGLeft(avl) ((avl)->left)
#define AvlALeft(avl) (&((avl)->left))
#define AvlSLeft(avl,_left) ((avl)->left = (_left))
#define AvlGRight(avl) ((avl)->right)
#define AvlARight(avl) (&((avl)->right))
#define AvlSRight(avl,_right) ((avl)->right = (_right))

/***************************************************************************
*
* Function:  AvlCons
* Purpose:   This routine constructs an AVL tree. The tree is initialized
*            to have exactly one element.
* Arguments: element - the element
* Returns:   the tree
* See also:  AvlApply, AvlBackApply, AvlCopy, AvlDupl, AvlDiscard,
*            AvlDispose, AvlDump, AvlDelete, AvlDelExact, AvlChkBal
*
****************************************************************************
*/
FS1(AVL *, AvlCons,
    char *, element)
{
    AVL *avl;

    avl = AvlAlloc();
    AvlSElement(avl, Pvoid(element));
    AvlSBalance(avl, 0);
    AvlSLeft(avl, (AVL *) NULL);
    AvlSRight(avl, (AVL *) NULL);
    return(avl);
}

static VOID **cause;

/***************************************************************************
*
* Function:  AvlApply
* Purpose:   This routine traverses an AVL tree and calls a given function
*            at each element. The vertices are visited in an "in-order"
*            from left to right. The traversal is broken if f returns TRUE.
* Arguments: avl - the tree
*            f - the function
* Returns:   the address of the breaking element if such exists,
*            and NULL otherwise
* See also:  AvlCons, AvlBackApply, AvlCopy, AvlDupl, AvlDiscard,
*            AvlDispose, AvlDump, AvlDelete, AvlDelExact, AvlChkBal
*
****************************************************************************
*/
FS2(VOID **, AvlApply,
    AVL *, avl,
    FOOL, f)
{
    return((AvlGLeft(avl) && (cause = AvlApply(AvlGLeft(avl), f))) ? cause
	   : ((*f)(AvlGElement(avl)) ? AvlAElement(avl)
	      : ((AvlGRight(avl) && (cause = AvlApply(AvlGRight(avl), f)))
		 ? cause : NULL)));
}

/***************************************************************************
*
* Function:  AvlBackApply
* Purpose:   This is like AvlApply, except that the order of the traversal
*            is reverse.
* Arguments: avl - the tree
*            f - the function
* Returns:   the address of the breaking element if such exists,
*            and NULL otherwise
* See also:  AvlCons, AvlApply, AvlCopy, AvlDupl, AvlDiscard,
*            AvlDispose, AvlDump, AvlDelete, AvlDelExact, AvlChkBal
*
****************************************************************************
*/
FS2(VOID **, AvlBackApply,
    AVL *, avl,
    FOOL, f)
{
    return((AvlGRight(avl) && (cause = AvlApply(AvlGRight(avl), f))) ? cause
	   : ((*f)(AvlGElement(avl)) ? AvlAElement(avl)
	      : ((AvlGLeft(avl) && (cause = AvlApply(AvlGLeft(avl), f)))
		 ? cause : NULL)));
}

/***************************************************************************
*
* Function:  AvlCopy
* Purpose:   This routine makes a copy of an AVL tree.
* Arguments: avl - the tree to be copied
*            fcopy - this function copies an element
* Returns:   the copy
* See also:  AvlCons, AvlApply, AvlBackApply, AvlDupl, AvlDiscard,
*            AvlDispose, AvlDump, AvlDelete, AvlDelExact, AvlChkBal
*
****************************************************************************
*/
FS2(AVL *, AvlCopy,
    AVL *, avl,
    CFUNC, fcopy)
{
    AVL *copy;

    copy = AvlAlloc();
    if (AvlGLeft(avl))
	AvlSLeft(copy, AvlCopy(AvlGLeft(avl), fcopy));
    else
	AvlSLeft(copy, (AVL *) NULL);
    AvlSBalance(copy, AvlGBalance(avl));
    AvlSElement(copy, (*fcopy)(AvlGElement(avl)));
    if (AvlGRight(avl))
	AvlSRight(copy, AvlCopy(AvlGRight(avl), fcopy));
    else
	AvlSRight(copy, (AVL *) NULL);
    return(copy);
}

/***************************************************************************
*
* Function:  AvlDupl
* Purpose:   This is like AvlCopy, except that elements are not copied.
* Arguments: avl - the tree to be copied
* Returns:   the copy
* See also:  AvlCons, AvlApply, AvlBackApply, AvlCopy, AvlDiscard,
*            AvlDispose, AvlDump, AvlDelete, AvlDelExact, AvlChkBal
*
****************************************************************************
*/
FS1(AVL *, AvlDupl,
    AVL *, avl)
{
    AVL *copy;

    copy = AvlAlloc();
    if (AvlGLeft(avl))
	AvlSLeft(copy, AvlDupl(AvlGLeft(avl)));
    else
	AvlSLeft(copy, (AVL *) NULL);
    AvlSBalance(copy, AvlGBalance(avl));
    AvlSElement(copy, AvlGElement(avl));
    if (AvlGRight(avl))
	AvlSRight(copy, AvlDupl(AvlGRight(avl)));
    else
	AvlSRight(copy, (AVL *) NULL);
    return(copy);
}

/***************************************************************************
*
* Function:  AvlDiscard
* Purpose:   This routine removes an AVL tree from the memory.
* Arguments: avl - the tree
*            discard - this function removes an element from the memory
* Returns:
* See also:  AvlCons, AvlApply, AvlBackApply, AvlCopy, AvlDupl,
*            AvlDispose, AvlDump, AvlDelete, AvlDelExact, AvlChkBal
*
****************************************************************************
*/
FS2(void, AvlDiscard,
    AVL *, avl,
    VOF, discard)
{
    if (AvlGLeft(avl))
	AvlDiscard(AvlGLeft(avl), discard);
    (*discard)(AvlGElement(avl));
    if (AvlGRight(avl))
	AvlDiscard(AvlGRight(avl), discard);
    Deallocate((char *) avl);
}

/***************************************************************************
*
* Function:  AvlDispose
* Purpose:   This is like AvlDiscard, except that the elements are not
*            touched.
* Arguments: avl - the tree
* Returns:
* See also:  AvlCons, AvlApply, AvlBackApply, AvlCopy, AvlDupl, AvlDiscard,
*            AvlDump, AvlDelete, AvlDelExact, AvlChkBal
*
****************************************************************************
*/
FS1(void, AvlDispose,
    AVL *, avl)
{
    if (AvlGLeft(avl))
	AvlDispose(AvlGLeft(avl));
    if (AvlGRight(avl))
	AvlDispose(AvlGRight(avl));
    Deallocate((char *) avl);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  AvlDump
* Purpose:   This routine displays an AVL tree.
* Arguments: avl - the tree
*            dump - this function displays an element
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  AvlCons, AvlApply, AvlBackApply, AvlCopy, AvlDupl, AvlDiscard,
*            AvlDispose, AvlDelete, AvlDelExact, AvlChkBal
*
****************************************************************************
*/
FS3(void, AvlDump,
    AVL *, avl,
    VOF, dump,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "AVL balance %d\n", AvlGBalance(avl));
    DumpIndent(depth + 1);
    (void) fputs("element\n", _err);
    (*dump)(AvlGElement(avl), depth + 2);
    DumpIndent(depth + 1);
    (void) fputs("left\n", _err);
    if (AvlGLeft(avl))
	AvlDump(AvlGLeft(avl), dump, depth + 2);
    DumpIndent(depth + 1);
    (void) fputs("right\n", _err);
    if (AvlGRight(avl))
	AvlDump(AvlGRight(avl), dump, depth + 2);
    DumpIndent(depth);
    (void) fputs("End of AVL\n", _err);
    (void) fflush(_err);
}
#endif

/***************************************************************************
*
* Function:  BalTrCreate
* Purpose:   This routine creates an AVL tree with side information.
*            The tree is initialized to have no element.
* Arguments: tools - a collection of routines for copying, deleting and
*                    displaying of an element in the tree, and for comparing
*                    two elements, and for checking whether an element
*                    has a given prefix in some context-dependent sense
* Returns:   the tree
* See also:  BalTrCopy, BalTrDestroy, BalTrCp, BalTrDc, BalTrDump,
*            BalTrApply, BalTrBackApply, BalTrImage, BalTrInsert,
*            BalTrDelete, BalTrElement, BalTrGrEqPrefix, BalTrNext,
*            BalTrFirst, BalTrLast, BalTrChkBal
*
****************************************************************************
*/
FE1(BALTREE *, BalTrCreate,
    BALTOOLS *, tools)
{
    BALTREE *balTree;

    balTree = BalTrAlloc();
    BalTrSElCount(balTree, (unsigned long) 0L);
    BalTrSRoot(balTree, (AVL *) NULL);
    BalTrSTools(balTree, tools);
    return(balTree);
}

/***************************************************************************
*
* Function:  BalTrCopy
* Purpose:   This routine makes a copy of an AVL tree (with side
*            information).
* Arguments: balTree - the tree to be copied
*            elCopy - elements are copied if and only if elCopy is TRUE
* Returns:   the copy
* See also:  BalTrCreate, BalTrDestroy, BalTrCp, BalTrDc, BalTrDump,
*            BalTrApply, BalTrBackApply, BalTrImage, BalTrInsert,
*            BalTrDelete, BalTrElement, BalTrGrEqPrefix, BalTrNext,
*            BalTrFirst, BalTrLast, BalTrChkBal, AvlCopy, AvlDupl
*
****************************************************************************
*/
FE2(BALTREE *, BalTrCopy,
    BALTREE *, balTree,
    BOOL, elCopy)
{
    BALTREE *copy;

    copy = BalTrAlloc();
    BalTrSElCount(copy, BalTrGElCount(balTree));
    BalTrSTools(copy, BalTrGTools(balTree));
    if (BalTrGRoot(balTree))
	{
	    if (elCopy)
		BalTrSRoot(copy, AvlCopy(BalTrGRoot(balTree),
					 BalElCopy(balTree)));
	    else
		BalTrSRoot(copy, AvlDupl(BalTrGRoot(balTree)));
	}
    else
	BalTrSRoot(copy, (AVL *) NULL);
    return(copy);
}

/***************************************************************************
*
* Function:  BalTrDstroy
* Purpose:   This routine removes an AVL tree (with side information) from
*            the memory.
* Arguments: balTree - the tree
*            elDiscard - elements are removed if and only if
*                        elDiscard is TRUE
* Returns:
* See also:  BalTrCreate, BalTrCopy, BalTrCp, BalTrDc, BalTrDump,
*            BalTrApply, BalTrBackApply, BalTrImage, BalTrInsert,
*            BalTrDelete, BalTrElement, BalTrGrEqPrefix, BalTrNext,
*            BalTrFirst, BalTrLast, BalTrChkBal, AvlDiscard, AvlDispose
*
****************************************************************************
*/
FE2(void, BalTrDstroy,
    BALTREE *, balTree,
    BOOL, elDiscard)
{
    if (balTree)
	{
	    if (BalTrGRoot(balTree))
		{
		    if (elDiscard)
			AvlDiscard(BalTrGRoot(balTree), BalElDiscard(balTree));
		    else
			AvlDispose(BalTrGRoot(balTree));
		}
	    Deallocate((char *) balTree);
	}
}

/***************************************************************************
*
* Function:  BalTrCp
* Purpose:   This instance of BalTrCopy is needed when copying must be
*            done by a function having one argument only.
* Arguments: balTree - the tree to be copied
* Returns:   the copy
* See also:  BalTrCreate, BalTrCopy, BalTrDestroy, BalTrDc, BalTrDump,
*            BalTrApply, BalTrBackApply, BalTrImage, BalTrInsert,
*            BalTrDelete, BalTrElement, BalTrGrEqPrefix, BalTrNext,
*            BalTrFirst, BalTrLast, BalTrChkBal
*
****************************************************************************
*/
FS1(BALTREE *, BalTrCp,
    BALTREE *, balTree)
{
    return(BalTrCopy(balTree, TRUE));
}

/***************************************************************************
*
* Function:  BalTrDc
* Purpose:   This instance of BalTrDestroy is needed when destroying
*            must be done by a function having one argument only.
* Arguments: balTree - the tree
* Returns:
* See also:  BalTrCreate, BalTrCopy, BalTrDestroy, BalTrCp, BalTrDump,
*            BalTrApply, BalTrBackApply, BalTrImage, BalTrInsert,
*            BalTrDelete, BalTrElement, BalTrGrEqPrefix, BalTrNext,
*            BalTrFirst, BalTrLast, BalTrChkBal
*
****************************************************************************
*/
FS1(void, BalTrDc,
    BALTREE *, balTree)
{
    BalTrDestroy(balTree, TRUE);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  BalTrDump
* Purpose:   This routine displays an AVL tree (with side information).
* Arguments: balTree - the tree
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  BalTrCreate, BalTrCopy, BalTrDestroy, BalTrCp, BalTrDc,
*            BalTrApply, BalTrBackApply, BalTrImage, BalTrInsert,
*            BalTrDelete, BalTrElement, BalTrGrEqPrefix, BalTrNext,
*            BalTrFirst, BalTrLast, BalTrChkBal, AvlDump
*
****************************************************************************
*/
FE2(void, BalTrDump,
    BALTREE *, balTree,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "BALTREE elCount %lu\n", BalTrGElCount(balTree));
    if (BalTrGRoot(balTree))
	AvlDump(BalTrGRoot(balTree), BalElDump(balTree), depth + 1);
    DumpIndent(depth);
    (void) fputs("End of BALTREE\n", _err);
    (void) fflush(_err);
}
#endif

/* A collection of routines for the BALTREE type to be used by
 * list routines of dynam.c and table routines of ptrbl.c. */
TOOLS balTrTools =
{ (CFUNC) BalTrCp, BalTrDc
#ifdef DEBUG
, BalTrDump
#endif
 };

/***************************************************************************
*
* Function:  BalTrApply
* Purpose:   This routine traverses an AVL tree (with side information)
*            and calls a boolean function at each element. The vertices are
*            visited in an "in-order" from left to right. The traversal
*            is broken if f returns TRUE.
* Arguments: balTree - the tree
*            f - the function
* Returns:   the address of the breaking element if such exists,
*            and NULL otherwise
* See also:  BalTrCreate, BalTrCopy, BalTrDestroy, BalTrCp, BalTrDc,
*            BalTrDump, BalTrBackApply, BalTrImage, BalTrInsert,
*            BalTrDelete, BalTrElement, BalTrGrEqPrefix, BalTrNext,
*            BalTrFirst, BalTrLast, BalTrChkBal, AvlApply
*
****************************************************************************
*/
FE2(VOID **, BalTrApply,
    BALTREE *, balTree,
    FOOL, f)
{
    return((BalTrGRoot(balTree) && (cause = AvlApply(BalTrGRoot(balTree), f)))
	   ? cause : NULL);
}

/***************************************************************************
*
* Function:  BalTrBackApply
* Purpose:   This is like BalTrApply, except that the order of the
*            traversal is reverse.
* Arguments: balTree - the tree
*            f - the function
* Returns:   the address of the breaking element if such exists,
*            and NULL otherwise
* See also:  BalTrCreate, BalTrCopy, BalTrDestroy, BalTrCp, BalTrDc,
*            BalTrDump, BalTrApply, BalTrImage, BalTrInsert,
*            BalTrDelete, BalTrElement, BalTrGrEqPrefix, BalTrNext,
*            BalTrFirst, BalTrLast, BalTrChkBal, AvlBackApply
*
****************************************************************************
*/
FE2(VOID **, BalTrBackApply,
    BALTREE *, balTree,
    FOOL, f)
{
    return((BalTrGRoot(balTree) &&
	    (cause = AvlBackApply(BalTrGRoot(balTree), f)))
	   ? cause : NULL);
}

/***************************************************************************
*
* Function:  BalTrImage
* Purpose:   This routine produces an image of an AVL tree (with side
*            information). The image is structurally the same as the
*            original tree but the elements are different.
*            The used mapping from an element to an image element must
*            be strictly increasing. (Otherwise the image tree would
*            not be a good search tree.)
* Arguments: balTree - the original tree
*            tools - a collection of routines for the image tree
*            strictlyIncreasing - the used mapping
* Returns:   the image tree
* See also:  BalTrCreate, BalTrCopy, BalTrDestroy, BalTrCp, BalTrDc,
*            BalTrDump, BalTrApply, BalTrBackApply, BalTrInsert,
*            BalTrDelete, BalTrElement, BalTrGrEqPrefix, BalTrNext,
*            BalTrFirst, BalTrLast, BalTrChkBal, AvlCopy
*
****************************************************************************
*/
FE3(BALTREE *, BalTrImage,
    BALTREE *, balTree,
    BALTOOLS *, tools,
    CFUNC, strictlyIncreasing)
{
    BALTREE *image;

    image = BalTrCreate(tools);
    BalTrSElCount(image, BalTrGElCount(balTree));
    if (BalTrGRoot(balTree))
	BalTrSRoot(image, AvlCopy(BalTrGRoot(balTree), strictlyIncreasing));
    else
	BalTrSRoot(image, (AVL *) NULL);
    return(image);
}

/***************************************************************************
*
* Function:  LeftHigher
* Purpose:   This is an auxiliary routine needed in the insertion of
*            an element into an AVL tree. (Adapted from "Etsint{- ja
*            lajittelumenetelm{t", a report written by Harri K. Laine
*            in the Computer Science Department of University of Helsinki.)
* Arguments: r - a reference to a subtree
*            s - another subtree
* Returns:
* See also:  RightHigher, BalTrInsert
*
****************************************************************************
*/
FS2(void, LeftHigher,
    AVL **, r,
    AVL *, s)
{
    AVL *c;

    if (AvlGBalance(*r) == -1)
	{
	    AvlSLeft(s, AvlGRight(*r));
	    AvlSRight(*r, s);
	    AvlSBalance(s, 0);
	    AvlSBalance(*r, 0);
	}
    else
	{
	    c = AvlGRight(*r);
	    AvlSRight(*r, AvlGLeft(c));
	    AvlSLeft(c, *r);
	    AvlSLeft(s, AvlGRight(c));
	    AvlSRight(c, s);
	    if (AvlGBalance(c) == 0)
		{
		    AvlSBalance(s, 0);
		    AvlSBalance(*r, 0);
		}
	    else if (AvlGBalance(c) == -1)
		{
		    AvlSBalance(s, 1);
		    AvlSBalance(*r, 0);
		}
	    else
		{
		    AvlSBalance(s, 0);
		    AvlSBalance(*r, -1);
		}
	    AvlSBalance(c, 0);
	    *r = c;
	}
}

/***************************************************************************
*
* Function:  RightHigher
* Purpose:   This is an auxiliary routine needed in the insertion of
*            an element into an AVL tree. (Adapted from "Etsint{- ja
*            lajittelumenetelm{t", a report written by Harri K. Laine
*            in the Computer Science Department of University of Helsinki.)
* Arguments: r - a reference to a subtree
*            s - another subtree
* Returns:
* See also:  LeftHigher, BalTrInsert
*
****************************************************************************
*/
FS2(void, RightHigher,
    AVL **, r,
    AVL *, s)
{
    AVL *c;

    if (AvlGBalance(*r) == 1)
	{
	    AvlSRight(s, AvlGLeft(*r));
	    AvlSLeft(*r, s);
	    AvlSBalance(s, 0);
	    AvlSBalance(*r, 0);
	}
    else
	{
	    c = AvlGLeft(*r);
	    AvlSLeft(*r, AvlGRight(c));
	    AvlSRight(c, *r);
	    AvlSRight(s, AvlGLeft(c));
	    AvlSLeft(c, s);
	    if (AvlGBalance(c) == 0)
		{
		    AvlSBalance(s, 0);
		    AvlSBalance(*r, 0);
		}
	    else if (AvlGBalance(c) == 1)
		{
		    AvlSBalance(s, -1);
		    AvlSBalance(*r, 0);
		}
	    else
		{
		    AvlSBalance(s, 0);
		    AvlSBalance(*r, 1);
		}
	    AvlSBalance(c, 0);
	    *r = c;
	}
}

/***************************************************************************
*
* Function:  BalTrInsert
* Purpose:   This routine inserts an element into an AVL tree. (Adapted from
*            "Etsint{- ja lajittelumenetelm{t", a report written by
*            Harri K. Laine in the Computer Science Department of
*            University of Helsinki.)
* Arguments: balTree - the tree
*            element - the intended new element
*            key - the key of the element
*            unique - if this is TRUE and there is an old element with
*                     the same key, the new element is not inserted
* Returns:   NULL if the element was inserted, and otherwise an old
*            element with the same key
* See also:  BalTrCreate, BalTrCopy, BalTrDestroy, BalTrCp, BalTrDc,
*            BalTrDump, BalTrApply, BalTrBackApply, BalTrImage,
*            BalTrDelete, BalTrElement, BalTrGrEqPrefix, BalTrNext,
*            BalTrFirst, BalTrLast, BalTrChkBal, AvlCons, LeftHigher,
*            RightHigher
*
****************************************************************************
*/
FE4(VOID *, BalTrInsert,
    BALTREE *, balTree,
    char *, element,
    char *, key,
    BOOL, unique)
{
    AVL *newAvl, *f, *s, *j, *i, *r;
    int cmpValue, b;

    if (!BalTrGRoot(balTree))
	{
	    BalTrSRoot(balTree, AvlCons(element));
	    BalTrIElCount(balTree);
	    return(NULL);
	}
    s = i = BalTrGRoot(balTree);
    f = j = NULL;
    do
	{
	    if (AvlGBalance(i) != 0)
		{
		    s = i;
		    f = j;
		}
	    /* NOTE: key may be the key of AvlGElement(i), even if
	     *       element is not AvlGElement(i). */
	    if (!(cmpValue = (*BalElKeyCmp(balTree))(AvlGElement(i), key)))
		{
		    if (unique || (element == (char *) AvlGElement(i)))
			return(AvlGElement(i));
		    else
			cmpValue =
			    (element < (char *) AvlGElement(i)) ? 1 : -1;
		}
	    j = i;
	    if (cmpValue > 0)
		i = AvlGLeft(i);
	    else
		i = AvlGRight(i);
	} while (i);
    newAvl = AvlCons(element);
    BalTrIElCount(balTree);
    if (cmpValue > 0)
	AvlSLeft(j, newAvl);
    else
	AvlSRight(j, newAvl);
    if (!(cmpValue = (*BalElKeyCmp(balTree))(AvlGElement(s), key)))
	{
	    if (element < (char *) AvlGElement(s))
		{
		    i = AvlGLeft(s);
		    b = -1;
		}
	    else
		{
		    i = AvlGRight(s);
		    b = 1;
		}
	}
    else if (cmpValue > 0)
	{
	    i = AvlGLeft(s);
	    b = -1;
	}
    else
	{
	    i = AvlGRight(s);
	    b = 1;
	}
    r = i;
    while (i != newAvl)
	{
	    if (!(cmpValue =
		  (*BalElKeyCmp(balTree))(AvlGElement(i), key)))
		{
		    if (element < (char *) AvlGElement(i))
			{
			    AvlSBalance(i, -1);
			    i = AvlGLeft(i);
			}
		    else
			{
			    AvlSBalance(i, 1);
			    i = AvlGRight(i);
			}
		}
	    else if (cmpValue > 0)
		{
		    AvlSBalance(i, -1);
		    i = AvlGLeft(i);
		}
	    else
		{
		    AvlSBalance(i, 1);
		    i = AvlGRight(i);
		}
	}
    if (AvlGBalance(s) == 0)
	{
	    AvlSBalance(s, b);
	    return(NULL);
	}
    else if (AvlGBalance(s) + b == 0)
	{
	    AvlSBalance(s, 0);
	    return(NULL);
	}
    if (b == 1)
	RightHigher(&r, s);
    else
	LeftHigher(&r, s);
    if (!f)
	BalTrSRoot(balTree, r);
    else if (s == AvlGLeft(f))
	AvlSLeft(f, r);
    else
	AvlSRight(f, r);
    return(NULL);
}

static BOOL adH;

/***************************************************************************
*
* Function:  LeftLower
* Purpose:   This is an auxiliary routine needed in the deletion of
*            an element from an AVL tree. (Adapted from "Algorithms +
*            data structures = programs", a book written by Niklaus
*            Wirth and published by Prentice-Hall in 1976.)
* Arguments: p - a subtree
* Returns:
* See also:  RightLower, AvlDelete, AvlDelExact
*
****************************************************************************
*/
FS1(void, LeftLower,
    AVL **, p)
{
    AVL *p1, *p2;
    int bp, b1, b2;

    bp = AvlGBalance(*p);
    if (bp == -1)
	AvlSBalance(*p, 0);
    else
	{
	    if (bp == 0)
		{
		    AvlSBalance(*p, 1);
		    adH = FALSE;
		}
	    else
		{
		    p1 = AvlGRight(*p);
		    b1 = AvlGBalance(p1);
		    if (b1 != -1)
			{
			    AvlSRight(*p, AvlGLeft(p1));
			    AvlSLeft(p1, *p);
			    if (b1 == 0)
				{
				    AvlSBalance(*p, 1);
				    AvlSBalance(p1, -1);
				    adH = FALSE;
				}
			    else
				{
				    AvlSBalance(*p, 0);
				    AvlSBalance(p1, 0);
				}
			    *p = p1;
			}
		    else
			{
			    p2 = AvlGLeft(p1);
			    b2 = AvlGBalance(p2);
			    AvlSLeft(p1, AvlGRight(p2));
			    AvlSRight(p2, p1);
			    AvlSRight(*p, AvlGLeft(p2));
			    AvlSLeft(p2, *p);
			    if (b2 == 1)
				AvlSBalance(*p, -1);
			    else
				AvlSBalance(*p, 0);
			    if (b2 == -1)
				AvlSBalance(p1, 1);
			    else
				AvlSBalance(p1, 0);
			    *p = p2;
			    AvlSBalance(p2, 0);
			}
		}
	}
}

/***************************************************************************
*
* Function:  RightLower
* Purpose:   This is an auxiliary routine needed in the deletion of
*            an element from an AVL tree. (Adapted from "Algorithms +
*            data structures = programs", a book written by Niklaus
*            Wirth and published by Prentice-Hall in 1976.)
* Arguments: p - a subtree
* Returns:
* See also:  LeftLower, AvlDelete, AvlDelExact
*
****************************************************************************
*/
FS1(void, RightLower,
    AVL **, p)
{
    AVL *p1, *p2;
    int bp, b1, b2;

    bp = AvlGBalance(*p);
    if (bp == 1)
	AvlSBalance(*p, 0);
    else
	{
	    if (bp == 0)
		{
		    AvlSBalance(*p, -1);
		    adH = FALSE;
		}
	    else
		{
		    p1 = AvlGLeft(*p);
		    b1 = AvlGBalance(p1);
		    if (b1 != 1)
			{
			    AvlSLeft(*p, AvlGRight(p1));
			    AvlSRight(p1, *p);
			    if (b1 == 0)
				{
				    AvlSBalance(*p, -1);
				    AvlSBalance(p1, 1);
				    adH = FALSE;
				}
			    else
				{
				    AvlSBalance(*p, 0);
				    AvlSBalance(p1, 0);
				}
			    *p = p1;
			}
		    else
			{
			    p2 = AvlGRight(p1);
			    b2 = AvlGBalance(p2);
			    AvlSRight(p1, AvlGLeft(p2));
			    AvlSLeft(p2, p1);
			    AvlSLeft(*p, AvlGRight(p2));
			    AvlSRight(p2, *p);
			    if (b2 == -1)
				AvlSBalance(*p, 1);
			    else
				AvlSBalance(*p, 0);
			    if (b2 == 1)
				AvlSBalance(p1, -1);
			    else
				AvlSBalance(p1, 0);
			    *p = p2;
			    AvlSBalance(p2, 0);
			}
		}
	}
}

static AVL *adPrev;

/***************************************************************************
*
* Function:  DelDel
* Purpose:   This is an auxiliary routine needed in the deletion of
*            an element from an AVL tree. (Adapted from "Algorithms +
*            data structures = programs", a book written by Niklaus
*            Wirth and published by Prentice-Hall in 1976.)
* Arguments: r - a reference to a subtree
* Returns:
* See also:  RightLower, AvlDelete, AvlDelExact
*
****************************************************************************
*/
FS1(void, DelDel,
    AVL **, r)
{
    if (AvlGRight(*r))
	{
	    DelDel(AvlARight(*r));
	    if (adH)
		RightLower(r);
	}
    else
	{
	    adPrev = *r;
	    *r = AvlGLeft(*r);
	    adH = TRUE;
	}
}

static BALTREE *adTree;
static char *adKey;
static VOID *adElement;
static int adcmpValue;
static AVL *adDelNode;

/***************************************************************************
*
* Function:  AvlDelete
* Purpose:   This routine deletes an element from an AVL tree.
*            (Adapted from "Algorithms + data structures = programs",
*            a book written by Niklaus Wirth and published by Prentice-Hall
*            in 1976.) The element is specified by adKey. If there is an
*            element having the key, adDelnode is set equal to a tree having
*            nothing but the deleted element. Otherwise adDelNode remains
*            NULL. If there are several elements with the same key, one of
*            them is deleted. (This makes sense if all of them are deleted
*            by calling this routine until adDelNode remains NULL.)
* Arguments: p - a reference to the tree
* Returns:
* See also:  AvlCons, AvlApply, AvlBackApply, AvlCopy, AvlDupl, AvlDiscard,
*            AvlDispose, AvlDump, AvlDelExact, AvlChkBal, LeftLower,
*            RightLower, DelDel
*
****************************************************************************
*/
FS1(void, AvlDelete,
    AVL **, p)
{
    if (!(adcmpValue = (*BalElKeyCmp(adTree))(AvlGElement(*p), adKey)))
	{
	    adDelNode = *p;
	    if (!AvlGRight(adDelNode))
		{
		    *p = AvlGLeft(adDelNode);
		    adH = TRUE;
		}
	    else if (!AvlGLeft(adDelNode))
		{
		    *p = AvlGRight(adDelNode);
		    adH = TRUE;
		}
	    else
		{
		    DelDel(AvlALeft(adDelNode));
		    *p = adPrev;
		    AvlSLeft(*p, AvlGLeft(adDelNode));
		    AvlSRight(*p, AvlGRight(adDelNode));
		    AvlSBalance(*p, AvlGBalance(adDelNode));
		    if (adH)
			LeftLower(p);
		}
	}
    else if (adcmpValue > 0)
	{
	    if (AvlGLeft(*p))
		{
		    AvlDelete(AvlALeft(*p));
		    if (adDelNode && adH)
			LeftLower(p);
		}
	}
    else
	{
	    if (AvlGRight(*p))
		{
		    AvlDelete(AvlARight(*p));
		    if (adDelNode && adH)
			RightLower(p);
		}
	}
}

/***************************************************************************
*
* Function:  AvlDelExact
* Purpose:   This is like AvlDelete, except that the element is uniquely
*            specified by adElement.
* Arguments: p - a reference to the tree
* Returns:
* See also:  AvlCons, AvlApply, AvlBackApply, AvlCopy, AvlDupl, AvlDiscard,
*            AvlDispose, AvlDump, AvlDelete, AvlChkBal, LeftLower,
*            RightLower, DelDel
*
****************************************************************************
*/
FS1(void, AvlDelExact,
    AVL **, p)
{
    if (!(adcmpValue = (*BalElKeyCmp(adTree))(AvlGElement(*p), adKey)))
	{
	    if (adElement == AvlGElement(*p))
		{
		    adDelNode = *p;
		    if (!AvlGRight(adDelNode))
			{
			    *p = AvlGLeft(adDelNode);
			    adH = TRUE;
			}
		    else if (!AvlGLeft(adDelNode))
			{
			    *p = AvlGRight(adDelNode);
			    adH = TRUE;
			}
		    else
			{
			    DelDel(AvlALeft(adDelNode));
			    *p = adPrev;
			    AvlSLeft(*p, AvlGLeft(adDelNode));
			    AvlSRight(*p, AvlGRight(adDelNode));
			    AvlSBalance(*p, AvlGBalance(adDelNode));
			    if (adH)
				LeftLower(p);
			}
		}
	    else if (adElement < AvlGElement(*p))
		{
		    if (AvlGLeft(*p))
			{
			    AvlDelExact(AvlALeft(*p));
			    if (adDelNode && adH)
				LeftLower(p);
			}
		}
	    else
		{
		    if (AvlGRight(*p))
			{
			    AvlDelExact(AvlARight(*p));
			    if (adDelNode && adH)
				RightLower(p);
			}
		}
	}
    else if (adcmpValue > 0)
	{
	    if (AvlGLeft(*p))
		{
		    AvlDelExact(AvlALeft(*p));
		    if (adDelNode && adH)
			LeftLower(p);
		}
	}
    else
	{
	    if (AvlGRight(*p))
		{
		    AvlDelExact(AvlARight(*p));
		    if (adDelNode && adH)
			RightLower(p);
		}
	}
}

/***************************************************************************
*
* Function:  BalTrDelete
* Purpose:   This routine deletes an element from an AVL tree (with side
*            information).
* Arguments: balTree - the tree
*            element - the element to be deleted (may be NULL if not
*                      known beforehand or if the given key uniquely
*                      specifies the element)
*            key - the key of the element
*            elDiscard - the element is removed from the memory if and only
*                        if elDiscard is TRUE
* Returns:   the deleted element if possible, and NULL otherwise
* See also:  BalTrCreate, BalTrCopy, BalTrDestroy, BalTrCp, BalTrDc,
*            BalTrDump, BalTrApply, BalTrBackApply, BalTrImage, BalTrInsert,
*            BalTrElement, BalTrGrEqPrefix, BalTrNext,
*            BalTrFirst, BalTrLast, BalTrChkBal, AvlDelete, AvlDelExact
*
****************************************************************************
*/
FE4(VOID *, BalTrDelete,
    BALTREE *, balTree,
    char *, element,
    char *, key,
    BOOL, elDiscard)
{
    if (BalTrGRoot(balTree))
	{
	    adTree = balTree;
	    adKey = key;
	    adDelNode = NULL;
	    if (element)
		{
		    adElement = Pvoid(element);
		    AvlDelExact(BalTrARoot(balTree));
		}
	    else
		AvlDelete(BalTrARoot(balTree));
	    if (adDelNode)
		{
		    BalTrDElCount(adTree);
		    adElement = AvlGElement(adDelNode);
		    if (elDiscard)
			(*BalElDiscard(balTree))(adElement);
		    Deallocate((char *) adDelNode);
		    return(adElement);
		}
	    else
		return(NULL);
	}
    else
	return(NULL);
}

/***************************************************************************
*
* Function:  BalTrElement
* Purpose:   This routine searches for an element having a key in an AVL
*            tree.
* Arguments: balTree - the tree
*            key - the key
* Returns:   a matching element if found, and NULL otherwise
* See also:  BalTrCreate, BalTrCopy, BalTrDestroy, BalTrCp, BalTrDc,
*            BalTrDump, BalTrApply, BalTrBackApply, BalTrImage, BalTrInsert,
*            BalTrDelete, BalTrGrEqPrefix, BalTrNext,
*            BalTrFirst, BalTrLast, BalTrChkBal
*
****************************************************************************
*/
FE2(VOID *, BalTrElement,
    BALTREE *, balTree,
    char *, key)
{
    AVL *p;
    int cmpValue;

    p = BalTrGRoot(balTree);
    while (p)
	{
	    if (!(cmpValue = (*BalElKeyCmp(balTree))(AvlGElement(p), key)))
		return(AvlGElement(p));
	    else if (cmpValue > 0)
		p = AvlGLeft(p);
	    else
		p = AvlGRight(p);
	}
    return(NULL);
}

/***************************************************************************
*
* Function:  BalTrGrEqPrefix
* Purpose:   In an AVL tree, this routine searches for the first element
*            having a key greater than or equal to a given prefix.
* Arguments: balTree - the tree
*            prefix - the prefix
*            equalPrefix - if equalPefix is TRUE, only such element
*                          is accepted which has the above prefix
* Returns:   the wanted element if such exists, and NULL otherwise
* See also:  BalTrCreate, BalTrCopy, BalTrDestroy, BalTrCp, BalTrDc,
*            BalTrDump, BalTrApply, BalTrBackApply, BalTrImage, BalTrInsert,
*            BalTrDelete, BalTrElement, BalTrNext,
*            BalTrFirst, BalTrLast, BalTrChkBal
*
****************************************************************************
*/
FE3(VOID *, BalTrGrEqPrefix,
    BALTREE *, balTree,
    char *, prefix,
    BOOL, equalPrefix)
{
    AVL *p, *cand;

    cand = NULL;
    p = BalTrGRoot(balTree);
    while (p)
	{
	    if ((*BalElKeyCmp(balTree))(AvlGElement(p), prefix) >= 0)
		{
		    cand = p;
		    p = AvlGLeft(p);
		}
	    else
		p = AvlGRight(p);
	}
    if (cand)
	{
	    if (equalPrefix)
		{
		    if ((*BalElPrefix(balTree))(AvlGElement(cand), prefix))
			return(AvlGElement(cand));
		    else
			return(NULL);
		}
	    else
		return(AvlGElement(cand));
	}
    else
	return(NULL);
}

#ifdef SMODELS
/***************************************************************************
*
* Function:  BalTrLtEqPrefix
* Purpose:   In an AVL tree, this routine searches for the last element
*            having a key less than or equal to a given prefix.
* Arguments: balTree - the tree
*            prefix - the prefix
*            equalPrefix - if equalPefix is TRUE, only such element
*                          is accepted which has the above prefix
* Returns:   the wanted element if such exists, and NULL otherwise
* See also:  BalTrCreate, BalTrCopy, BalTrDestroy, BalTrCp, BalTrDc,
*            BalTrDump, BalTrApply, BalTrBackApply, BalTrImage, BalTrInsert,
*            BalTrDelete, BalTrElement, BalTrNext,
*            BalTrFirst, BalTrLast, BalTrChkBal
*
****************************************************************************
*/
FE3(VOID *, BalTrLtEqPrefix,
    BALTREE *, balTree,
    char *, prefix,
    BOOL, equalPrefix)
{
    AVL *p, *cand;

    cand = NULL;
    p = BalTrGRoot(balTree);
    while (p)
	{
	    if ((*BalElKeyCmp(balTree))(AvlGElement(p), prefix) <= 0)
		{
		    cand = p;
		    p = AvlGRight(p);
		}
	    else
		p = AvlGLeft(p);
	}
    if (cand)
	{
	    if (equalPrefix)
		{
		    if ((*BalElPrefix(balTree))(AvlGElement(cand), prefix))
			return(AvlGElement(cand));
		    else
			return(NULL);
		}
	    else
		return(AvlGElement(cand));
	}
    else
	return(NULL);
}
#endif

/***************************************************************************
*
* Function:  BalTrNext
* Purpose:   This routine searches for an element immediately following
*            a given element in an AVL tree.
* Arguments: balTree - the tree
*            element - the current element
*            key - the key of the current element
*            prefix - the next element is acceptable only if it has this
*                     prefix or if prefix is NULL
* Returns:   the wanted prefix if such exists, and NULL otherwise
* See also:  BalTrCreate, BalTrCopy, BalTrDestroy, BalTrCp, BalTrDc,
*            BalTrDump, BalTrApply, BalTrBackApply, BalTrImage, BalTrInsert,
*            BalTrDelete, BalTrElement, BalTrGrEqPrefix,
*            BalTrFirst, BalTrLast, BalTrChkBal
*
****************************************************************************
*/
FE4(VOID *, BalTrNext,
    BALTREE *, balTree,
    char *, element,
    char *, key,
    char *, prefix)
{
    AVL *p, *cand;
    int cmpValue;

    cand = NULL;
    p = BalTrGRoot(balTree);
    while (p)
	{
	    if (!(cmpValue = (*BalElKeyCmp(balTree))(AvlGElement(p), key)))
		{
		    if (element < (char *) AvlGElement(p))
			{
			    cand = p;
			    p = AvlGLeft(p);
			}
		    else
			p = AvlGRight(p);
		}
	    else if (cmpValue > 0)
		{
		    cand = p;
		    p = AvlGLeft(p);
		}
	    else
		p = AvlGRight(p);
	}
    if (cand)
	{
	    if (prefix)
		{
		    if ((*BalElPrefix(balTree))(AvlGElement(cand), prefix))
			return(AvlGElement(cand));
		    else
			return(NULL);
		}
	    else
		return(AvlGElement(cand));
	}
    else
	return(NULL);
}

/***************************************************************************
*
* Function:  BalTrFirst
* Purpose:   This routine returns the last node of an AVL tree.
* Arguments: balTree - the tree
* Returns:   the first element if the tree is not empty, and NULL otherwise
* See also:  BalTrCreate, BalTrCopy, BalTrDestroy, BalTrCp, BalTrDc,
*            BalTrDump, BalTrApply, BalTrBackApply, BalTrImage, BalTrInsert,
*            BalTrDelete, BalTrElement, BalTrGrEqPrefix, BalTrNext,
*            BalTrLast, BalTrChkBal
*
****************************************************************************
*/
FE1(VOID *, BalTrFirst,
    BALTREE *, balTree)
{
    AVL *p;

    if ((p = BalTrGRoot(balTree)))
	{
	    while (AvlGLeft(p))
		p = AvlGLeft(p);
	    return(AvlGElement(p));
	}
    else
	return(NULL);
}

/***************************************************************************
*
* Function:  BalTrLast
* Purpose:   This routine returns the last element of an AVL tree.
* Arguments: balTree - the tree
* Returns:   the last element if the tree is not empty, and NULL otherwise
* See also:  BalTrCreate, BalTrCopy, BalTrDestroy, BalTrCp, BalTrDc,
*            BalTrDump, BalTrApply, BalTrBackApply, BalTrImage, BalTrInsert,
*            BalTrDelete, BalTrElement, BalTrGrEqPrefix, BalTrNext,
*            BalTrFirst, BalTrChkBal
*
****************************************************************************
*/
FE1(VOID *, BalTrLast,
    BALTREE *, balTree)
{
    AVL *p;

    if ((p = BalTrGRoot(balTree)))
	{
	    while (AvlGRight(p))
		p = AvlGRight(p);
	    return(AvlGElement(p));
	}
    else
	return(NULL);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  AvlChkBal
* Purpose:   This routine checks that an AVL tree satisfies the AVL
*            criterion, i.e. for each vertex, the heights of child
*            subtrees either are equal or differ from each other by one.
* Arguments: avl - the tree
* Returns:   the maximum of the heights of the child subtrees of the root
* See also:  AvlCons, AvlApply, AvlBackApply, AvlCopy, AvlDupl, AvlDiscard,
*            AvlDispose, AvlDump, AvlDelete, AvlDelExact
*
****************************************************************************
*/
FS1(unsigned long, AvlChkBal,
    AVL *, avl)
{
    unsigned long l, r;

    if (AvlGLeft(avl))
	l = AvlChkBal(AvlGLeft(avl)) + 1L;
    else
	l = 0L;
    if (AvlGRight(avl))
	r = AvlChkBal(AvlGRight(avl)) + 1L;
    else
	r = 0L;
    if (l > r)
	{
	    if ((l != r + 1L) || (AvlGBalance(avl) != -1))
		Troba4("AvlChkBal: lheight %lu, rheight %lu, balance %d",
		       l, r, AvlGBalance(avl));
	    return(l);
	}
    else if (r > l)
	{
	    if ((r != l + 1L) || (AvlGBalance(avl) != 1))
		Troba4("AvlChkBal: lheight %lu, rheight %lu, balance %d",
		       l, r, AvlGBalance(avl));
	    return(r);
	}
    else
	{
	    if (AvlGBalance(avl) != 0)
		Troba4("AvlChkBal: lheight %lu, rheight %lu, balance %d",
		       l, r, AvlGBalance(avl));
	    return(l);
	}
}

/***************************************************************************
*
* Function:  BalTrChkBal
* Purpose:   This routine checks that an AVL tree (with side information)
*            satisfies the AVL criterion, i.e. for each vertex, the heights
*            of child subtrees either are equal or differ from each other
*            by one.
* Arguments: balTree - the tree
* Returns:
* See also:  BalTrCreate, BalTrCopy, BalTrDestroy, BalTrCp, BalTrDc,
*            BalTrDump, BalTrApply, BalTrBackApply, BalTrImage, BalTrInsert,
*            BalTrDelete, BalTrElement, BalTrGrEqPrefix, BalTrNext,
*            BalTrFirst, BalTrLast, AvlChkBal
*
****************************************************************************
*/
FE1(void, BalTrChkBal,
    BALTREE *, balTree)
{
    if (BalTrGRoot(balTree))
	(void) AvlChkBal(BalTrGRoot(balTree));
}
#endif
