/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  bdd.c
*
* Part of: PROD's programs araprod and "gphgen"
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     This module contains general routines for binary decision diagrams
*     (BDD's). These routines are straightforward implementations of the
*     algorithms in "Graph-Based Algorithms for Boolean Function
*     Manipulation", an article written by Randal E. Bryant and published
*     in IEEE Transactions on Computers C-35 (1986) 8, pp. 677-691.
*     Unfortunately, everything in this module is so inefficient
*     that it is highly impractical to run applications which use this
*     module. The only such application is the "-B" option of the
*     reachability graph generator or araprod. (See the BalTrBdd call
*     in the GphPutNode routine of graph.c.) The following quotation
*     from PROD Reference Manual describes the problem.
*     "-B
*      During reachability graph generation, the set of computed states is
*      kept in a BDD (binary decision diagram) \cite{Bry86}. The current
*      implementation merely increases the time and space consumed during
*      generation, but an efficient implementation may be available in the
*      future."
*
*     Instead of commenting this module in detail, I suggest that someone
*     designs something strictly better from scratch.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "gphutil.h"

#define BddAlloc() ((BDD *) Allocate(sizeof(BDD)))
#define BddGVal(bdd) (((bdd)->flag & 2) ? 2 : ((bdd)->flag & 1))
#define BddSVal(bdd,_val) \
    ((bdd)->flag &= ~3, (bdd)->flag |= ((_val) & 2) ? 2 : ((_val) & 1))
#define BddGMark(bdd) ((bdd)->flag & 4)
#define BddCMark(bdd) ((bdd)->flag &= ~4)
#define BddSMark(bdd) ((bdd)->flag |= 4)
#define BddNMark(bdd) ((bdd)->flag ^= 4)
#define BddGIndex(bdd) ((bdd)->index)
#define BddAIndex(bdd) (&((bdd)->index))
#define BddSIndex(bdd,_index) ((bdd)->index = (_index))
#define BddGLow(bdd) ((bdd)->low)
#define BddALow(bdd) (&((bdd)->low))
#define BddSLow(bdd,_low) ((bdd)->low = (_low))
#define BddGHigh(bdd) ((bdd)->high)
#define BddAHigh(bdd) (&((bdd)->high))
#define BddSHigh(bdd,_high) ((bdd)->high = (_high))
#define BddGRepres(bdd) ((bdd)->repres)
#define BddARepres(bdd) (&((bdd)->repres))
#define BddSRepres(bdd,_repres) ((bdd)->repres = (_repres))

typedef struct DDB
{
    BDD *bdd1;
    BDD *bdd2;
    BDD *bdd3;
} DDB;

#define DdbAlloc() ((DDB *) Allocate(sizeof(DDB)))
#define DdbG1(ddb) ((ddb)->bdd1)
#define DdbA1(ddb) (&((ddb)->bdd1))
#define DdbS1(ddb,_bdd) ((ddb)->bdd1 = (_bdd))
#define DdbG2(ddb) ((ddb)->bdd2)
#define DdbA2(ddb) (&((ddb)->bdd2))
#define DdbS2(ddb,_bdd) ((ddb)->bdd2 = (_bdd))
#define DdbG3(ddb) ((ddb)->bdd3)
#define DdbA3(ddb) (&((ddb)->bdd3))
#define DdbS3(ddb,_bdd) ((ddb)->bdd3 = (_bdd))

LIST *_bddList;

FS4(BDD *, BddCons,
    unsigned, val,
    unsigned long, ind,
    BDD *, low,
    BDD *, high)
{
    BDD *bdd;

    bdd = BddAlloc();
    bdd->flag = 0;
    if (low)
	{
	    BddSVal(bdd, 2);
	    if (BddGMark(low))
		BddSMark(bdd);
	    else
		BddCMark(bdd);
	}
    else
	BddSVal(bdd, val);
    BddSIndex(bdd, ind);
    BddSLow(bdd, low);
    BddSHigh(bdd, high);
    BddSRepres(bdd, bdd);
    return(bdd);
}

#ifdef DEBUG
FS2(void, BddDump,
    BDD *, bdd,
    unsigned, depth)
{
    BddNMark(bdd);
    DumpIndent(depth);
    (void) fprintf(_err, "BDD index %lu flag %u repres 0x%lx 0x%lx\n",
		   BddGIndex(bdd), bdd->flag,
		   (unsigned long) BddGRepres(bdd), (unsigned long) bdd);
    (void) fflush(_err);
    if (BddGIndex(bdd) == ((unsigned long) -1L))
	return;
    DumpIndent(depth);
    (void) fputs("BDD low\n", _err);
    (void) fflush(_err);
    if (BddGMark(bdd) != BddGMark(BddGLow(bdd)))
	BddDump(BddGLow(bdd), depth + 1);
    else
	{
	    DumpIndent(depth + 1);
	    (void) fprintf(_err,
			   "BDD index %lu flag %u repres 0x%lx 0x%lx\n",
			   BddGIndex(BddGLow(bdd)), BddGLow(bdd)->flag,
			   (unsigned long) BddGRepres(BddGLow(bdd)),
			   (unsigned long) BddGLow(bdd));
	    (void) fflush(_err);
	}
    DumpIndent(depth);
    (void) fputs("BDD high\n", _err);
    (void) fflush(_err);
    if (BddGMark(bdd) != BddGMark(BddGHigh(bdd)))
	BddDump(BddGHigh(bdd), depth + 1);
    else
	{
	    DumpIndent(depth + 1);
	    (void) fprintf(_err,
			   "BDD index %lu flag %u repres 0x%lx 0x%lx\n",
			   BddGIndex(BddGHigh(bdd)), BddGHigh(bdd)->flag,
			   (unsigned long) BddGRepres(BddGHigh(bdd)),
			   (unsigned long) BddGHigh(bdd));
	    (void) fflush(_err);
	}
}
#endif

FS2(int, BddCompare,
    BDD *, bdd1,
    BDD *, bdd2)
{
    if (BddGRepres(BddGLow(bdd1)) < BddGRepres(BddGLow(bdd2)))
	return(-1);
    if (BddGRepres(BddGLow(bdd1)) > BddGRepres(BddGLow(bdd2)))
	return(1);
    if (BddGRepres(BddGHigh(bdd1)) < BddGRepres(BddGHigh(bdd2)))
	return(-1);
    if (BddGRepres(BddGHigh(bdd1)) > BddGRepres(BddGHigh(bdd2)))
	return(1);
    return(0);
}

TOOLS bddTools =
{ (CFUNC) NULL, Deallocate
#ifdef DEBUG
, BddDump
#endif
 };

static BALTOOLS bddBalTools =
{ (CFUNC) NULL, Deallocate
#ifdef DEBUG
, BddDump
#endif
, BddCompare, (FOOL) NULL };

FS3(DDB *, DdbCons,
    BDD *, bdd1,
    BDD *, bdd2,
    BDD *, bdd3)
{
    DDB *ddb;

    ddb = DdbAlloc();
    DdbS1(ddb, bdd1);
    DdbS2(ddb, bdd2);
    DdbS3(ddb, bdd3);
    return(ddb);
}

#ifdef DEBUG
FS2(void, DdbDump,
    DDB *, ddb,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err,
		   "DDB bdd1 index %lu flag %u\n",
		   BddGIndex(BddGHigh(DdbG1(ddb))),
		   BddGHigh(DdbG1(ddb))->flag);
    DumpIndent(depth);
    (void) fprintf(_err,
		   "DDB bdd2 index %lu flag %u\n",
		   BddGIndex(BddGHigh(DdbG2(ddb))),
		   BddGHigh(DdbG2(ddb))->flag);
    DumpIndent(depth);
    (void) fprintf(_err,
		   "DDB bdd3 index %lu flag %u\n",
		   BddGIndex(BddGHigh(DdbG3(ddb))),
		   BddGHigh(DdbG3(ddb))->flag);
    (void) fflush(_err);
}
#endif

FS2(int, DdbCompare,
    DDB *, ddb1,
    DDB *, ddb2)
{
    if (DdbG1(ddb1) < DdbG1(ddb2))
	return(-1);
    if (DdbG1(ddb1) > DdbG1(ddb2))
	return(1);
    if (DdbG2(ddb1) < DdbG2(ddb2))
	return(-1);
    if (DdbG2(ddb1) > DdbG2(ddb2))
	return(1);
    return(0);
}

static BALTOOLS ddbBalTools =
{ (CFUNC) NULL, Deallocate
#ifdef DEBUG
, DdbDump
#endif
, DdbCompare, (FOOL) NULL };


#ifdef DEBUG
FS2(void, BddRefDump,
    LIST *, bddRef,
    unsigned, depth)
{
    BDD *bdd;

    DumpIndent(depth);
    (void) fputs("BDDREF list\n", _err);
    LSTFOR (bddRef, bdd, BDD)
	{
	    DumpIndent(depth + 1);
	    (void) fprintf(_err,
			   "BDD index %lu flag %u\n",
			   BddGIndex(bdd), bdd->flag);
	}
    DumpIndent(depth);
    (void) fputs("End of BDDREF list\n", _err);
    (void) fflush(_err);
}
#endif

FS2(int, BddRefKeyCmp,
    LIST *, bddRef,
    unsigned long *, key)
{
    return((BddGIndex((BDD *) LstFirst(bddRef)) == *key) ? 0 :
	   ((BddGIndex((BDD *) LstFirst(bddRef)) < *key) ? -1 : 1));
}

static BALTOOLS bddRefBalTools =
{ (CFUNC) NULL, (VOF) NULL
#ifdef DEBUG
, BddRefDump
#endif
, BddRefKeyCmp, (FOOL) NULL };

static BALTREE *bddRefTree = NULL;
static LIST *_bddRef;
static BDD *prevBdd;
static BDD *curBdd;
static BDD *firstBdd;
static BDD *result;
static BALTREE *level;

FS1(void, BddTraverse,
    BDD *, bdd)
{
    BddNMark(bdd);
    if (BddGIndex(bdd) == ((unsigned long) -1L))
	return;
    if (BddGMark(bdd) != BddGMark(BddGLow(bdd)))
	BddTraverse(BddGLow(bdd));
    if (BddGMark(bdd) != BddGMark(BddGHigh(bdd)))
	BddTraverse(BddGHigh(bdd));
}

FE1(void, BddCat,
    BDD *, bdd)
{
    BddNMark(bdd);
    (void) LstInsLast(BddList(), (char *) bdd);
    if (BddGIndex(bdd) == ((unsigned long) -1L))
	return;
    if (BddGMark(bdd) != BddGMark(BddGLow(bdd)))
	BddCat(BddGLow(bdd));
    if (BddGMark(bdd) != BddGMark(BddGHigh(bdd)))
	BddCat(BddGHigh(bdd));
}

FS1(void, BddCollect,
    BDD *, bdd)
{
    BddNMark(bdd);
    _bddRef = (LIST *) BalTrElement(bddRefTree, (char *) BddAIndex(bdd));
    if (!_bddRef)
	{
	    _bddRef = LstCreate(&bddTools);
	    (void) BalTrInsert(bddRefTree, (char *) _bddRef,
			       (char *) BddAIndex(bdd), TRUE);
	}
    (void) LstInsLast(_bddRef, (char *) bdd);
    if (BddGIndex(bdd) == ((unsigned long) -1L))
	return;
    if (BddGMark(bdd) != BddGMark(BddGLow(bdd)))
	BddCollect(BddGLow(bdd));
    if (BddGMark(bdd) != BddGMark(BddGHigh(bdd)))
	BddCollect(BddGHigh(bdd));
}

FS1(BOOL, BddRefUpdate,
    LIST *, bddRef)
{
    if (BddGIndex((BDD *) LstFirst(bddRef)) == (unsigned long) -1L)
	{
	    firstBdd = (BDD *) LstBegin(bddRef);
	    curBdd = (BDD *) LstStep(bddRef);
	    prevBdd = NULL;
	    while (curBdd)
		{
		    if (BddGVal(curBdd) == BddGVal(firstBdd))
			BddSRepres(curBdd, firstBdd);
		    else if (prevBdd)
			BddSRepres(curBdd, prevBdd);
		    else
			prevBdd = curBdd;
		    curBdd = (BDD *) LstStep(bddRef);
		}
	    return(FALSE);
	}
    level = NULL;
    LSTFOR (bddRef, curBdd, BDD)
	{
	    if (BddGRepres(BddGLow(curBdd)) ==
		BddGRepres(BddGHigh(curBdd)))
		BddSRepres(curBdd, BddGRepres(BddGLow(curBdd)));
	    else if (!level)
		{
		    level = BalTrCreate(&bddBalTools);
		    (void) BalTrInsert(level, (char *) curBdd,
				       (char *) curBdd, TRUE);
		    BddSLow(curBdd, BddGRepres(BddGLow(curBdd)));
		    BddSHigh(curBdd, BddGRepres(BddGHigh(curBdd)));
		}
	    else if ((prevBdd = (BDD *)
		      BalTrInsert(level, (char *) curBdd,
				  (char *) curBdd, TRUE)))
		BddSRepres(curBdd, prevBdd);
	    else
		{
		    BddSLow(curBdd, BddGRepres(BddGLow(curBdd)));
		    BddSHigh(curBdd, BddGRepres(BddGHigh(curBdd)));
		}
	}
    if (level)
	BalTrDestroy(level, FALSE);
    return(FALSE);
}

FS1(BOOL, BddRefGarbage,
    LIST *, bddRef)
{
    LSTFOR (bddRef, curBdd, BDD)
	{
	    if (BddGRepres(curBdd) != curBdd)
		Deallocate((char *) curBdd);
	}
    LstDestroy(bddRef, FALSE);
    return(FALSE);
}

FS1(BDD *, BddReduce,
    BDD *, bdd)
{
    bddRefTree = BalTrCreate(&bddRefBalTools);
    BddCollect(bdd);
    (void) BalTrBackApply(bddRefTree, BddRefUpdate);
    result = BddGRepres(bdd);
    (void) BalTrBackApply(bddRefTree, BddRefGarbage);
    BalTrDestroy(bddRefTree, FALSE);
    return(result);
}

static BDD *zBdd;
static unsigned long bddIndex;
static unsigned bitCount;
static DATATYPE *dPointer;

FS2(BDD *, BddNew,
    DATATYPE *, buffer,
    unsigned long, address)
{
    bddIndex = (unsigned long) -1L;
    curBdd = BddCons(1, bddIndex, (BDD *) NULL, (BDD *) NULL);
    zBdd = BddCons(0, bddIndex, (BDD *) NULL, (BDD *) NULL);
    dPointer = buffer;
    while (*dPointer++)
	;
    bddIndex = BITSPERBYTE
	* (((unsigned long) (dPointer - buffer)) * sizeof(DATATYPE)
	   + sizeof(unsigned long));
    bitCount = BITSPERBYTE * sizeof(unsigned long);
    while (bitCount--)
	{
	    bddIndex--;
	    if ((address >> bitCount) & 1L)
		curBdd = BddCons(2, bddIndex, zBdd, curBdd);
	    else
		curBdd = BddCons(2, bddIndex, curBdd, zBdd);
	}
    while (bddIndex)
	{
	    dPointer--;
	    bitCount = BITSPERBYTE * sizeof(DATATYPE);
	    while (bitCount--)
		{
		    bddIndex--;
		    if ((*dPointer >> bitCount) & (DATATYPE) 1)
			curBdd = BddCons(2, bddIndex, zBdd, curBdd);
		    else
			curBdd = BddCons(2, bddIndex, curBdd, zBdd);
		}
	}
    return(curBdd);
}

static BALTREE *appT;
static DDB *newDdb;
static DDB *oldDdb;
static unsigned out[27];
#define DOut(i,j) (out[3 * (i) + (j)])
#define DSetOut(i,j,a) (out[3 * (i) + (j)] = (a))
#define TOut(i,j,k) (out[3 * (3 * (i) + (j)) + (k)])
#define TSetOut(i,j,k,a) (out[3 * (3 * (i) + (j)) + (k)] = (a))

FS2(BDD *, BddStepApply,
    BDD *, bdd1,
    BDD *, bdd2)
{
    BDD *aux;

    newDdb = DdbCons(bdd1, bdd2, (BDD *) NULL);
    oldDdb = (DDB *) BalTrInsert(appT, (char *) newDdb, (char *) newDdb, TRUE);
    if (oldDdb)
	{
	    Deallocate((char *) newDdb);
	    return(DdbG3(oldDdb));
	}
    aux = BddCons(0, ((unsigned long) -1L), (BDD *) NULL, (BDD *) NULL);
    DdbS3(newDdb, aux);
    BddSVal(aux, DOut(BddGVal(bdd1), BddGVal(bdd2)));
    if (BddGVal(aux) != 2)
	return(aux);
    if (BddGIndex(bdd1) < BddGIndex(bdd2))
	{
	    BddSIndex(aux, BddGIndex(bdd1));
	    BddSLow(aux, BddStepApply(BddGLow(bdd1), bdd2));
	    BddSHigh(aux, BddStepApply(BddGHigh(bdd1), bdd2));
	    return(aux);
	}
    if (BddGIndex(bdd1) > BddGIndex(bdd2))
	{
	    BddSIndex(aux, BddGIndex(bdd2));
	    BddSLow(aux, BddStepApply(bdd1, BddGLow(bdd2)));
	    BddSHigh(aux, BddStepApply(bdd1, BddGHigh(bdd2)));
	    return(aux);
	}
    BddSIndex(aux, BddGIndex(bdd1));
    BddSLow(aux, BddStepApply(BddGLow(bdd1), BddGLow(bdd2)));
    BddSHigh(aux, BddStepApply(BddGHigh(bdd1), BddGHigh(bdd2)));
    return(aux);
}

FS3(BDD *, BddApply,
    BDD *, bdd1,
    BDD *, bdd2,
    UNSF, f)
{
    static unsigned i, j;

    for (i = 0; i < 2; i++)
	{
	    for (j = 0; j < 2; j++)
		DSetOut(i, j, (f(i, j) != 0));
	    if (DOut(i, 0) == DOut(i, 1))
		DSetOut(i, 2, DOut(i, 0));
	    else
		DSetOut(i, 2, 2);
	}
    for (j = 0; j < 3; j++)
	{
	    if (DOut(0, j) == DOut(1, j))
		DSetOut(2, j, DOut(0, j));
	    else
		DSetOut(2, j, 2);
	}
    if (DOut(2, 2) != 2)
	return(BddCons(DOut(2, 2), ((unsigned long) -1L),
		       (BDD *) NULL, (BDD *) NULL));
    appT = BalTrCreate(&ddbBalTools);
    result = BddStepApply(bdd1, bdd2);
    BalTrDestroy(appT, TRUE);
    return(BddReduce(result));
}

static BALTREE *restr;
static unsigned long *bddLng;
static unsigned long bddLeastInd;

FS1(void, BddIndFind,
    BDD *, bdd)
{
    BddNMark(bdd);
    if (BddGIndex(bdd) == ((unsigned long) -1L))
	return;
    if (!bddLng)
	{
	    bddIndex = BddGIndex(bdd) << 1;
	    bddLng = (unsigned long *)
		BalTrGrEqPrefix(restr, (char *) &bddIndex, TRUE);
	}
    if (BddGMark(bdd) != BddGMark(BddGLow(bdd)))
	BddIndFind(BddGLow(bdd));
    if (BddGMark(bdd) != BddGMark(BddGHigh(bdd)))
	BddIndFind(BddGHigh(bdd));
}

FS1(void, BddIndSimpleFind,
    BDD *, bdd)
{
    BddNMark(bdd);
    if (BddGIndex(bdd) == ((unsigned long) -1L))
	return;
    if (BddGIndex(bdd) == bddIndex)
	curBdd = bdd;
    if (BddGMark(bdd) != BddGMark(BddGLow(bdd)))
	BddIndSimpleFind(BddGLow(bdd));
    if (BddGMark(bdd) != BddGMark(BddGHigh(bdd)))
	BddIndSimpleFind(BddGHigh(bdd));
}

FS1(void, BddFix,
    BDD *, bdd)
{
    BddNMark(bdd);
    if (BddGIndex(bdd) >= bddLeastInd)
	(void) LstInsLast(BddList(), (char *) bdd);
    if (BddGIndex(bdd) == ((unsigned long) -1L))
	return;
    if (BddGMark(bdd) != BddGMark(BddGLow(bdd)))
	BddFix(BddGLow(bdd));
    if (BddGMark(bdd) != BddGMark(BddGHigh(bdd)))
	BddFix(BddGHigh(bdd));
    if ((bddIndex = BddGIndex(BddGLow(bdd))) != ((unsigned long) -1L))
	{
	    bddIndex <<= 1;
	    if ((bddLng = (unsigned long *)
		 BalTrGrEqPrefix(restr, (char *) &bddIndex, TRUE)))
		{
		    if (*bddLng & 1L)
			BddSLow(bdd, BddGHigh(BddGLow(bdd)));
		    else
			BddSLow(bdd, BddGLow(BddGLow(bdd)));
		}
	}
    if ((bddIndex = BddGIndex(BddGHigh(bdd))) != ((unsigned long) -1L))
	{
	    bddIndex <<= 1;
	    if ((bddLng = (unsigned long *)
		 BalTrGrEqPrefix(restr, (char *) &bddIndex, TRUE)))
		{
		    if (*bddLng & 1L)
			BddSHigh(bdd, BddGHigh(BddGHigh(bdd)));
		    else
			BddSHigh(bdd, BddGLow(BddGHigh(bdd)));
		}
	}
}

FS2(unsigned, BinOr,
    unsigned, a,
    unsigned, b)
{
    return(a || b);
}

static LIST *oldBddList;
static LIST *newBddList;
static BDD *var1Bdd;
static BDD *var2Bdd;

FE3(BOOL, BddFind,
    BDD *, bdd,
    DATATYPE *, buffer,
    unsigned long *, address)
{
    dPointer = buffer;
    while (*dPointer++)
	;
    bddIndex = BITSPERBYTE
	* ((unsigned long) (dPointer - buffer)) * sizeof(DATATYPE);
    curBdd = bdd;
    while (BddGIndex(curBdd) < bddIndex)
	{
	    dPointer = buffer + (unsigned)
		(BddGIndex(curBdd) / (BITSPERBYTE * sizeof(DATATYPE)));
	    if ((*dPointer >> (BddGIndex(curBdd)
			       % (BITSPERBYTE * sizeof(DATATYPE))))
		& (DATATYPE) 1)
		curBdd = BddGHigh(curBdd);
	    else
		curBdd = BddGLow(curBdd);
	}
    if (!BddGVal(curBdd))
	return(FALSE);
    *address = 0L;
    while (BddGIndex(curBdd) < bddIndex + BITSPERBYTE * sizeof(unsigned long))
	{
	    if (BddGVal(BddGLow(curBdd)))
		{
		    curBdd = BddGLow(curBdd);
		    continue;
		}
	    *address |= 1L << ((BddGIndex(curBdd) - bddIndex)
			       % (BITSPERBYTE * sizeof(unsigned long)));
	    curBdd = BddGHigh(curBdd);
	}
    return(TRUE);
}

FS1(BOOL, DrefBdd,
    DREF *, dref)
{
    (void) LstInsLast(oldBddList, (char *)
		      BddNew(DrefGKey(dref), DrefGSlot(dref)));
    DrefDiscard(dref);
    return(FALSE);
}

FE2(void, BalTrBdd,
    BALTREE *, balTree,
    BDD **, bdd)
{
    oldBddList = LstCreate(&bddTools);
    (void) BalTrApply(balTree, DrefBdd);
    BalTrDestroy(balTree, FALSE);
    if (*bdd)
	(void) LstInsFirst(oldBddList, (char *) *bdd);
    while (TRUE)
	{
	    var1Bdd = (BDD *) LstBegin(oldBddList);
	    var2Bdd = (BDD *) LstStep(oldBddList);
	    if (!var2Bdd)
		{
		    LstDestroy(oldBddList, FALSE);
		    *bdd = var1Bdd;
		    return;
		}
	    newBddList = LstCreate(&bddTools);
	    while (TRUE)
		{
		    (void) LstInsLast(newBddList, (char *)
				      BddApply(var1Bdd, var2Bdd, BinOr));
		    BddSList();
		    BddCat(var1Bdd);
		    BddDList();
		    BddSList();
		    BddCat(var2Bdd);
		    BddDList();
		    var1Bdd = (BDD *) LstStep(oldBddList);
		    if (!var1Bdd)
			break;
		    var2Bdd = (BDD *) LstStep(oldBddList);
		    if (!var2Bdd)
			{
			    (void) LstInsLast(newBddList, (char *) var1Bdd);
			    break;
			}
		}
	    LstDestroy(oldBddList, FALSE);
	    oldBddList = newBddList;
	}
}
