/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  graph.c
*
* Part of: PROD's programs araprod, "gphgen", probe, and strong
*
* Authors: Marko Rauhamaa and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     The routines in this module are needed for constructing and
*     inspecting a reachability graph of a net.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "error.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "gphutil.h"
#include "mark.h"
#include "encode.h"
#include "graph.h"
#include "ptnet.h"
#include "araprod.h"
#include "cmdnm.h"

/***************************************************************************
*
* Function:  SaveLatest
* Purpose:   This routine stores into the memory the address, the record
*            and the data of a node.
* Arguments: graph - the reachability graph
*            node_addr - the address of the node
*            all - if the graph is under construction or if strongly
*                  connected components are being computed, this
*                  parameter chooses whether the data of the node
*                  is stored into the memory
* Returns:
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FS3(void, SaveLatest,
    GRAPH *, graph,
    unsigned long, node_addr,
    BOOL, all)
{
    graph->latest_node = node_addr;
    NdsSkAbs(graph, node_addr);
    NdsReLng(graph->latest_rec, graph->num_node_fields, graph);
    if (all || (graph->reason == GPH_PROBE))
	{
	    NdsReData(graph->latest_contents,
		      graph->latest_rec[graph->datalen_pos], graph);
	}
}

/***************************************************************************
*
* Function:  GphStart
* Purpose:   This routine returns the address of the root node of
*            a reachability of a net.
* Arguments: graph - the reachability graph
* Returns:   the address of the root node
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE1(unsigned long, GphStart,
    GRAPH *, graph)
{
    unsigned long node_addr;
    
    if (graph->glob[NODE_COUNT_POS] == 0L)
	Troba1("Illegal GphStart call");
    node_addr = GphRoot();
    if (node_addr != graph->latest_node) 
	SaveLatest(graph, node_addr, FALSE);
    return(node_addr);
}

/***************************************************************************
*
* Function:  GphCompNumber
* Purpose:   This routine returns the number of the strongly connected
*            component containing a given node in a reachability graph
*            of a net. There is no need to call this routine if the
*            strongly connected components have not been computed.
* Arguments: graph - the reachability graph
*            node_addr - the address of the node
* Returns:   the number explained above
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE2(unsigned long, GphCompNumber,
    GRAPH *, graph,
    unsigned long, node_addr)
{
    if (graph->glob[SGC_COUNT_POS] == 0L)
	Troba1("Illegal GphCompNumber call");
    if (node_addr >= graph->glob[EOF_NODES_POS])
	Troba2("Illegal node address for GphCompNumber: %lu", node_addr);
    if (node_addr != graph->latest_node)
	SaveLatest(graph, node_addr, FALSE);
    return(((unsigned long) -1L) - graph->latest_rec[graph->sgc_val_pos]);
}

/***************************************************************************
*
* Function:  GphOrdinal
* Purpose:   This routine returns the ordinal number of a node
*            in a reachability graph of a net. The ordinal number
*            of the first node is 0L.
* Arguments: graph - the reachability graph
*            node_addr - the address of the node
* Returns:   the ordinal number of the node
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE2(unsigned long, GphOrdinal,
    GRAPH *, graph,
    unsigned long, node_addr)
{
    if (node_addr >= graph->glob[EOF_NODES_POS])
	Troba2("Illegal node address for GphOrdinal: %lu", node_addr);
    if (node_addr != graph->latest_node)
	SaveLatest(graph, node_addr, FALSE);
    return(graph->latest_rec[ORDINAL_POS]);
}

/***************************************************************************
*
* Function:  GphId
* Purpose:   This routine returns the address of a node in a reachability
*            graph of a net.
* Arguments: graph - the reachability graph
*            ordinal - the ordinal number of the node
* Returns:   the address of the node
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE2(unsigned long, GphId,
    GRAPH *, graph,
    unsigned long, ordinal)
{
    unsigned long node_addr = 0L;
    
    if (ordinal >= graph->glob[NODE_COUNT_POS])
	Troba2("Illegal node ordinal number for GphId: %lu", ordinal);
    if ((graph->latest_node != (unsigned long) -1L) &&
	(ordinal == graph->latest_rec[ORDINAL_POS]))
	return(graph->latest_node);
    AdrSkAbs(graph, ordinal * sizeof(unsigned long));
    AdrReLng(&node_addr, 1, graph);
    SaveLatest(graph, node_addr, FALSE);
    return(node_addr);
}

/***************************************************************************
*
* Function:  GphNdCount
* Purpose:   This routine returns the number of nodes in a reachability
*            graph of a net.
* Arguments: graph - the reachability graph
* Returns:   the number of nodes in the graph
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE1(unsigned long, GphNdCount,
    GRAPH *, graph)
{
    return(graph->glob[NODE_COUNT_POS]);
}

/***************************************************************************
*
* Function:  GphGlAwCount
* Purpose:   This routine returns the number of arrows in a reachability
*            graph of a net.
* Arguments: graph - the reachability graph
* Returns:   the number of arrows in the graph
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE1(unsigned long, GphGlAwCount,
    GRAPH *, graph)
{
    return(graph->glob[ARROW_COUNT_POS]);
}

/***************************************************************************
*
* Function:  GphAwCount
* Purpose:   This routine returns the number of immediate successor
*            arrows of a node in a reachability graph of a net.
* Arguments: graph - the reachability graph
*            node_addr - the address of the node
* Returns:   the number explained above, or, in the case of an untouched
*            node, -1L
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE2(unsigned long, GphAwCount,
    GRAPH *, graph,
    unsigned long, node_addr)
{
    if (node_addr >= graph->glob[EOF_NODES_POS])
	Troba2("Illegal node address for GphAwCount: %lu", node_addr);
    if (node_addr != graph->latest_node)
	SaveLatest(graph, node_addr, FALSE);
    if ((graph->latest_rec[SUCCESSOR_COUNT_POS] != (unsigned long) -1L) &&
	(graph->latest_rec[SUCCESSOR_COUNT_POS] >
	 graph->glob[ARROW_COUNT_POS]))
	Troba4("GphAwCount: node %lu, internal error, %lu > %lu",
	       node_addr, graph->latest_rec[SUCCESSOR_COUNT_POS],
	       graph->glob[ARROW_COUNT_POS]);
    return(graph->latest_rec[SUCCESSOR_COUNT_POS]);
}

/***************************************************************************
*
* Function:  GphFresh
* Purpose:   This routine checks whether a node in a reachability graph
*            of a net is fresh or not. Freshness is a notion used
*            in on-the-fly verification where the graph may have to be
*            traversed twice during its construction. There are different
*            levels of freshness.
* Arguments: graph - the reachability graph
*            node_addr - the address of the node
*            mask - this indicates the level of freshness in question
* Returns:   TRUE if the node is fresh, and FALSE otherwise
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE3(BOOL, GphFresh,
    GRAPH *, graph,
    unsigned long, node_addr,
    unsigned long, mask)
{
    if (!_inftr(graph))
	Troba1("Illegal GphFresh call");
    if (node_addr >= graph->glob[EOF_NODES_POS])
	Troba2("Illegal node address for GphFresh: %lu", node_addr);
    if (node_addr != graph->latest_node)
	SaveLatest(graph, node_addr, FALSE);
    return(!(graph->latest_rec[graph->check_pos] & mask));
}

/***************************************************************************
*
* Function:  GphConsume
* Purpose:   This routine writes down that a node in a reachability graph
*            of a net is not fresh any more. Freshness is a notion used
*            in on-the-fly verification where the graph may have to be
*            traversed twice during its construction. There are different
*            levels of freshness.
* Arguments: graph - the reachability graph
*            node_addr - the address of the node
*            mask - this indicates the level of freshness in question
* Returns:
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE3(void, GphConsume,
    GRAPH *, graph,
    unsigned long, node_addr,
    unsigned long, mask)
{
    if (!_inftr(graph))
	Troba1("Illegal GphConsume call");
    if (node_addr >= graph->glob[EOF_NODES_POS])
	Troba2("Illegal node address for GphConsume: %lu", node_addr);
    if (node_addr != graph->latest_node)
	SaveLatest(graph, node_addr, FALSE);
    if ((graph->latest_rec[graph->check_pos] & mask) == mask)
	Troba1("Useless GphConsume call");
    graph->latest_rec[graph->check_pos] |= mask;
    NdsSkAbs(graph, node_addr + graph->check_pos * sizeof(unsigned long));
    NdsWrLng(graph->latest_rec + graph->check_pos, 1, graph);
}

/***************************************************************************
*
* Function:  GphFlFirst
* Purpose:   This routine tells us the first immediate successor arrow
*            of a node in a reachability graph of a net. There is no need
*            to call this routine if the node has no immediate successor
*            arrow.
* Arguments: graph - the reachability graph
*            node_addr - the address of the node
*            awaddr - a reference to the address of the arrow
*            succ - a reference to the address of the target node
*                   of the arrow
* Returns:
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE4(void, GphFlFirst,
    GRAPH *, graph,
    unsigned long, node_addr,
    unsigned long *, awaddr,
    unsigned long *, succ)
{
    if (node_addr >= graph->glob[EOF_NODES_POS])
	Troba2("Illegal node address for GphFlFirst: %lu", node_addr);
    if (node_addr != graph->latest_node)
	SaveLatest(graph, node_addr, FALSE);
    if ((graph->latest_rec[SUCCESSOR_COUNT_POS] == 0L) ||
	(graph->latest_rec[SUCCESSOR_COUNT_POS] == (unsigned long) -1L))
	Troba1("Illegal GphFlFirst call");
    *awaddr = graph->latest_rec[FOLLOWERS_POS];
    if (succ)
	{
	    AwsSkAbs(graph, *awaddr + TARGET_POS * sizeof(unsigned long));
	    AwsReLng(succ, 1, graph);
	}
}

/***************************************************************************
*
* Function:  GphFlNext
* Purpose:   This routine tells us the next immediate successor arrow
*            of a node in a reachability graph of a net. There is no
*            need to call this routine if such arrow does not exist.
* Arguments: graph - the reachability graph
*            awaddr - a reference to the address of the current
*                     immediate successor arrow of the node
*            succ - a reference to the target node of the arrow
* Returns:
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE3(void, GphFlNext,
    GRAPH *, graph,
    unsigned long *, awaddr,
    unsigned long *, succ)
{
    unsigned long arrow_addr = 0L;
    unsigned long awdatalen = 0L;
    
    if (*awaddr >= graph->glob[EOF_ARROWS_POS])
	Troba2("Illegal arrow address for GphFlNext: %lu", *awaddr);
    if (_frexpansion(graph))
	{
	    AwsSkAbs(graph, *awaddr + AW_NEXT_POS * sizeof(unsigned long));
	    AwsReLng(&arrow_addr, 1, graph);
	}
    else
	{
	    AwsSkAbs(graph,
		     *awaddr + graph->aw_datalen_pos * sizeof(unsigned long));
	    AwsReLng(&awdatalen, 1, graph);
	    arrow_addr = 
		*awaddr + graph->num_arrow_fields * sizeof(unsigned long) +
		    awdatalen * sizeof(DATATYPE);
	}
    if (arrow_addr >= graph->glob[EOF_ARROWS_POS])
	Troba2("Illegal arrow address for GphFlNext: %lu", arrow_addr);
    AwsSkAbs(graph, arrow_addr + TARGET_POS * sizeof(unsigned long));
    AwsReLng(succ, 1, graph);
    *awaddr = arrow_addr;
}

/***************************************************************************
*
* Function:  GphPrLast
* Purpose:   This routine tells us the (so far) last immediate predecessor
*            arrow of a node in a reachability graph of a net.
* Arguments: graph - the reachability graph
*            node_addr - the address of the node
*            awaddr - a reference to the address of the arrow
*            pred - a reference to the address of the source node
*                   of the arrow
* Returns:
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE4(void, GphPrLast,
    GRAPH *, graph,
    unsigned long, node_addr,
    unsigned long *, awaddr,
    unsigned long *, pred)
{
    if (node_addr >= graph->glob[EOF_NODES_POS])
	Troba2("Illegal node address for GphPrLast: %lu", node_addr);
    if (node_addr != graph->latest_node)
	SaveLatest(graph, node_addr, FALSE);
    *awaddr = graph->latest_rec[graph->predecessors_pos];
    if (*awaddr >= graph->glob[EOF_ARROWS_POS])
	{
	    if ((*awaddr != (unsigned long) -1L) || (node_addr != GphRoot()))
		Troba3("GphPrLast: internal error, back %lu >= %lu",
		       *awaddr, graph->glob[EOF_ARROWS_POS]);
	    if (pred)
		*pred = (unsigned long) -1L;
	    return;
	}
    if (pred)
	{
	    AwsSkAbs(graph, *awaddr + SOURCE_POS * sizeof(unsigned long));
	    AwsReLng(pred, 1, graph);
	    if (*pred >= graph->glob[EOF_NODES_POS])
		Troba3("GphPrLast: internal error, source %lu >= %lu",
		       *pred, graph->glob[EOF_NODES_POS]);
	}
}

/***************************************************************************
*
* Function:  GphPrPrec
* Purpose:   This routine tells us the preceding immediate predecessor arrow
*            of a node in a reachability graph of a net.
* Arguments: graph - the reachability graph
*            awaddr - a reference to the address of the current
*                     immediate successor arrow of the node
*            succ - a reference to the target node of the arrow
* Returns:
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE3(void, GphPrPrec,
    GRAPH *, graph,
    unsigned long *, awaddr,
    unsigned long *, pred)
{
    unsigned long arrow_addr = 0L;
    
    if (*awaddr >= graph->glob[EOF_ARROWS_POS])
	Troba2("Illegal arrow address for GphPrPrec: %lu", *awaddr);
    AwsSkAbs(graph, *awaddr + AW_BACK_POS * sizeof(unsigned long));
    AwsReLng(&arrow_addr, 1, graph);
    if (arrow_addr >= graph->glob[EOF_ARROWS_POS])
	{
	    if (arrow_addr != (unsigned long) -1L)
		Troba2("Illegal arrow address for GphPrPrec: %lu", arrow_addr);
	    *pred = (unsigned long) -1L;
	}
    else
	{
	    AwsSkAbs(graph, arrow_addr + SOURCE_POS * sizeof(unsigned long));
	    AwsReLng(pred, 1, graph);
	    if (*pred >= graph->glob[EOF_NODES_POS])
		Troba3("GphPrPrec: internal error, source %lu >= %lu",
		       *pred, graph->glob[EOF_NODES_POS]);
	}
    *awaddr = arrow_addr;
}

/***************************************************************************
*
* Function:  GphNode
* Purpose:   This routine returns a pointer to a buffer containing an
*            encoded marking corresponding to a node in a reachability
*            graph of a net. The buffer should not be modified or
*            deallocated.
* Arguments: graph - the reachability graph
*            node_addr - the address of the node
* Returns:   the pointer explained above
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE2(DATATYPE *, GphNode,
    GRAPH *, graph,
    unsigned long, node_addr)
{
    if (node_addr >= graph->glob[EOF_NODES_POS])
	Troba2("Illegal node address for GphNode: %lu", node_addr);
    if ((node_addr != graph->latest_node) || (graph->reason != GPH_PROBE))
	SaveLatest(graph, node_addr, TRUE);
    return(graph->latest_contents);
}

/***************************************************************************
*
* Function:  GphArrow
* Purpose:   This routine returns a pointer to a buffer containing
*            the information of an arrow in a reachability graph of a net.
*            The buffer should neither be modified nor deallocated.
* Arguments: graph - the reachability graph
*            awaddr - the address of the arrow
* Returns:   the pointer explained above
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE2(DATATYPE *, GphArrow,
    GRAPH *, graph,
    unsigned long, awaddr)
{
    unsigned long awdatalen = 0L;
    
    if (awaddr >= graph->glob[EOF_ARROWS_POS])
	Troba2("Illegal arrow address for GphArrow: %lu", awaddr);
    AwsSkAbs(graph, awaddr + graph->aw_datalen_pos * sizeof(unsigned long));
    AwsReLng(&awdatalen, 1, graph);
    AwsReData(graph->latest_contents, awdatalen, graph);
    if (graph->reason == GPH_PROBE)
	graph->latest_node = ((unsigned long) -1L);
    return(graph->latest_contents);
}

/***************************************************************************
*
* Function:  GphChLastArrow
* Purpose:   This routine modifies the latest arrow of a reachability graph
*            of a net.
* Arguments: graph - the reachability graph
*            arrow - new information
*            len - the length of the arrow string
*            target - a new target node
*            aw_back_addr - the address of the latest incoming arrow
*                           of the target node if such arrow exists,
*                           and (unsigned long) -1L otherwise
*            back - returns the address of the arrow
*                   (exception: a NULL pointer returns nothing)
*            icold - TRUE if and only if the arrow is invisible and cold
* Returns:
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE7(void, GphChLastArrow,
    GRAPH *, graph,
    DATATYPE *, arrow,
    unsigned, len,
    unsigned long, target,
    unsigned long, aw_back_addr,
    unsigned long *, back,
    BOOL, icold)
{
    if (((unsigned long *) NULL) != back)
	*back = graph->glob[LAST_ARROW_POS];
    if (graph->glob[LAST_ARROW_POS] >= graph->glob[EOF_ARROWS_POS])
	Troba3("Illegal last for GphChLastArrow: %lu >= end %lu",
	       graph->glob[LAST_ARROW_POS], graph->glob[EOF_ARROWS_POS]);
    if (target >= graph->glob[EOF_NODES_POS])
	Troba3("Illegal target for GphChLastArrow: %lu >= end %lu",
	       target, graph->glob[EOF_NODES_POS]);
    AwsSkAbs(graph, graph->glob[LAST_ARROW_POS]);
    AwsReLng(graph->arrow_rec, graph->aw_datalen_pos, graph);
    if (graph->arrow_rec[SOURCE_POS] >= graph->glob[EOF_NODES_POS])
	Troba3("Illegal source for GphChLastArrow: %lu >= end %lu",
	       graph->arrow_rec[SOURCE_POS], graph->glob[EOF_NODES_POS]);
    if (icold)
	{
	    graph->arrow_rec[AW_ORDINAL_POS] &= ~3L;
	    graph->arrow_rec[AW_BACK_POS] = aw_back_addr;
	    graph->arrow_rec[TARGET_POS] = target;
	    graph->arrow_rec[graph->aw_datalen_pos] = (unsigned long) len;
	    AwsSkAbs(graph, graph->glob[LAST_ARROW_POS]);
	    AwsWrLng(graph->arrow_rec, graph->num_arrow_fields, graph);
	}
    else
	{
	    if (graph->arrow_rec[TARGET_POS] != target)
		{
		    if (graph->arrow_rec[AW_BACK_POS] != aw_back_addr)
			{
			    graph->arrow_rec[AW_BACK_POS] = aw_back_addr;
			    graph->arrow_rec[TARGET_POS] = target;
			    AwsSkAbs(graph,
				     graph->glob[LAST_ARROW_POS] +
				     AW_BACK_POS * sizeof(unsigned long));
			    AwsWrLng(graph->arrow_rec + AW_BACK_POS,
				     2, graph);
			}
		    else
			{
			    AwsSkAbs(graph,
				     graph->glob[LAST_ARROW_POS] +
				     TARGET_POS * sizeof(unsigned long));
			    AwsWrLng(&target, 1, graph);
			}
		}
	    else if (graph->arrow_rec[AW_BACK_POS] != aw_back_addr)
		{
		    AwsSkAbs(graph,
			     graph->glob[LAST_ARROW_POS] +
			     AW_BACK_POS * sizeof(unsigned long));
		    AwsWrLng(&aw_back_addr, 1, graph);
		}
	    if (graph->arrow_rec[graph->aw_datalen_pos] !=
		(unsigned long) len)
		{
		    graph->arrow_rec[graph->aw_datalen_pos] =
			(unsigned long) len;
		    AwsSkAbs(graph, graph->glob[LAST_ARROW_POS]
			     + graph->aw_datalen_pos * sizeof(unsigned long));
		    AwsWrLng(graph->arrow_rec + graph->aw_datalen_pos,
			     1, graph);
		}
	    else
		{
		    AwsSkAbs(graph, graph->glob[LAST_ARROW_POS]
			     + graph->num_arrow_fields
			     * sizeof(unsigned long));
		}
	}
    AwsWrData(arrow, len, graph);
    if (len > (unsigned) graph->glob[MAX_DATALEN_POS])
	graph->glob[MAX_DATALEN_POS] = (unsigned long) len;
    graph->glob[EOF_ARROWS_POS] = graph->glob[LAST_ARROW_POS] +
	len * sizeof(DATATYPE) +
	    graph->num_arrow_fields * sizeof(unsigned long);
    NdsSkAbs(graph, target + graph->predecessors_pos * sizeof(unsigned long));
    NdsWrLng(graph->glob + LAST_ARROW_POS, 1, graph);
}

/***************************************************************************
*
* Function:  GphAwOrdinal
* Purpose:   Given the nth immediate successor arrow of a node in
*            a reachability graph of a net, this routine returns n - 1L.
* Arguments: graph - the reachability graph
*            awaddr - the address of the arrow
* Returns:   the number explained above
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE2(unsigned long, GphAwOrdinal,
    GRAPH *, graph,
    unsigned long, awaddr)
{
    unsigned long ordinal = 0L;
    
    if (awaddr >= graph->glob[EOF_ARROWS_POS])
	Troba2("Illegal arrow address for GphAwOrdinal: %lu", awaddr);
    AwsSkAbs(graph, awaddr + AW_ORDINAL_POS * sizeof(unsigned long));
    AwsReLng(&ordinal, 1, graph);
    ordinal = (ordinal >> 2);
    if (ordinal >= graph->glob[ARROW_COUNT_POS])
	Troba4("GphAwOrdinal: addr %lu, internal error, %lu >= %lu",
	       awaddr, ordinal, graph->glob[ARROW_COUNT_POS]);
    return(ordinal);
}

/***************************************************************************
*
* Function:  GphAwInvisible
* Purpose:   This routine tells if or not an arrow in a reachability graph
*            of a net is invisible.
* Arguments: graph - the reachability graph
*            awaddr - the address of the arrow
* Returns:   TRUE if the arrow is visible, and FALSE otherwise
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE2(BOOL, GphAwInvisible,
    GRAPH *, graph,
    unsigned long, awaddr)
{
    unsigned long ordinal = 0L;
    
    if (awaddr >= graph->glob[EOF_ARROWS_POS])
	Troba2("Illegal arrow address for GphAwInvisible: %lu", awaddr);
    AwsSkAbs(graph, awaddr + AW_ORDINAL_POS * sizeof(unsigned long));
    AwsReLng(&ordinal, 1, graph);
    return((ordinal & 1L) == 0L);
}

/***************************************************************************
*
* Function:  GphAwHot
* Purpose:   This routine tells if or not an arrow in a reachability graph
*            of a net is hot.
* Arguments: graph - the reachability graph
*            awaddr - the address of the arrow
* Returns:   TRUE if the arrow is visible, and FALSE otherwise
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE2(BOOL, GphAwHot,
    GRAPH *, graph,
    unsigned long, awaddr)
{
    unsigned long ordinal = 0L;
    
    if (awaddr >= graph->glob[EOF_ARROWS_POS])
	Troba2("Illegal arrow address for GphAwHot: %lu", awaddr);
    AwsSkAbs(graph, awaddr + AW_ORDINAL_POS * sizeof(unsigned long));
    AwsReLng(&ordinal, 1, graph);
    return((ordinal & 3L) == 2L);
}

/***************************************************************************
*
* Function:  GphStatistics
* Purpose:   This routine displays global information about a reachability
*            graph of a net.
* Arguments: graph - the reachability graph
*            file - the file for the output
* Returns:
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE2(void, GphStatistics,
    GRAPH *, graph,
    FILE *, file)
{
    (void) fprintf(file, "Number of nodes: %lu\n",
		   graph->glob[NODE_COUNT_POS]);
    (void) fprintf(file, "Number of arrows: %lu\n",
		   graph->glob[ARROW_COUNT_POS]);
    (void) fprintf(file, "Number of terminal nodes: %lu\n",
		   graph->glob[TERMNODE_COUNT_POS]);
    (void) fprintf(file, "Number of %s: %lu\n",
		   _s_l_e_e_p(graph)
		   ? "poppings from node stack during graph generation"
		   : "nodes that have been completely processed",
		   graph->glob[GEN_COUNT_POS]);
    if (((unsigned long) 0L) == graph->glob[SGC_COUNT_POS])
	{
	    StrongNotComp(file);
	    return;
	}
    (void) fprintf(file, "Number of strongly connected components: %lu\n",
		   graph->glob[SGC_COUNT_POS]);
    (void)
	fprintf
	(file,
	 "Number of nontrivial terminal strongly connected components: %lu\n",
	 graph->glob[TERMCOMP_COUNT_POS]);
    (void) fflush(file);
}

/***************************************************************************
*
* Function:  GphSgCount
* Purpose:   This routine returns the number of strongly connected
*            components in a reachability graph of a net if the components
*            have been computed. Otherwise it returns 0L.
* Arguments: graph - the reachability graph
* Returns:   the number explained above
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE1(unsigned long, GphSgCount,
    GRAPH *, graph)
{
    return(graph->glob[SGC_COUNT_POS]);
}

/* The GV type is needed by the Visit routine. Recursion was eliminated
 * from the Visit routine by a stack of elements of GV. (In the versions
 * where recursion was used, overflows in the recursion stack caused
 * mysterious behaviours.) */
typedef struct GV
{
    unsigned long node_addr;
    LIST **succ_comp;
    unsigned long node_val;
    unsigned long min;
    unsigned long arrow_addr;
    unsigned long i;
    LIST *tmp;
} GV;

/***************************************************************************
*
* Function:  Visit
* Purpose:   This routine computes the strongly connected components
*            of a reachability graph of a net, using Tarjan's algorithm.
*            Moreover, a directed acyclic graph corresponding to
*            the accessibility relation between the components is
*             constructed.
* Arguments: graph - the reachability graph
*            node_addr - the address of the node that is being visited
*            succ_comp - a reference to a successor list needed in the
*                        construction of the component graph
* Returns:
* See also:  GphSgSearch
*
****************************************************************************
*/
FS3(void, Visit,
    GRAPH *, graph,
    unsigned long, node_addr,
    LIST **, succ_comp)
{
    unsigned long node_val;
    unsigned long min;
    unsigned long arrow_addr;
    unsigned long i;
    LIST *tmp;
    GV *gv;

 BEGIN: ;    
    (void) LstInsFirst(graph->list, (char *) LngCons(node_addr));
    (graph->now)++;
    node_val = graph->now;
    min = graph->now;
    NdsSkAbs(graph,
	     node_addr + graph->sgc_val_pos * sizeof(unsigned long));
    NdsWrLng(&node_val, 1, graph);
    NdsSkAbs(graph, 
	     node_addr + SUCCESSOR_COUNT_POS * sizeof(unsigned long));
    NdsReLng(graph->latest_rec + SUCCESSOR_COUNT_POS, 2, graph);
    i = graph->latest_rec[SUCCESSOR_COUNT_POS];
    if (i == 0L)
	goto CONCLUDE;
    arrow_addr = graph->latest_rec[FOLLOWERS_POS];
 ARROW: ;
    i--;
    AwsSkAbs(graph, arrow_addr + TARGET_POS * sizeof(unsigned long));
    AwsReLng(graph->arrow_rec + TARGET_POS, 2, graph);
    if (_frexpansion(graph))
	arrow_addr = graph->arrow_rec[AW_NEXT_POS];
    else
	arrow_addr +=
	    graph->arrow_rec[graph->aw_datalen_pos]
		* sizeof(DATATYPE) +
		    graph->num_arrow_fields * sizeof(unsigned long);
    NdsSkAbs(graph,
	     graph->arrow_rec[TARGET_POS] + 
	     graph->sgc_val_pos * sizeof(unsigned long));
    NdsReLng(&(graph->m), 1, graph);
    if (graph->m != 0L)
	goto PROCEED;
    gv = (GV *) Allocate(sizeof(GV));
    gv->node_addr = node_addr;
    gv->succ_comp = succ_comp;
    gv->node_val = node_val;
    gv->min = min;
    gv->arrow_addr = arrow_addr;
    gv->i = i;
    gv->tmp = NULL;
    (void) LstInsFirst(graph->vis_stack, (char *) gv);
    node_addr = graph->arrow_rec[TARGET_POS];
    succ_comp = &(gv->tmp);
    goto BEGIN;
 CONT: ;
    gv = (GV *) LstFirst(graph->vis_stack);
    node_addr = gv->node_addr;
    succ_comp = gv->succ_comp;
    node_val = gv->node_val;
    min = gv->min;
    arrow_addr = gv->arrow_addr;
    i = gv->i;
    tmp = gv->tmp;
    (void) LstDelFirst(graph->vis_stack, TRUE);
    if ((tmp != NULL) && (graph->m <= graph->now))
	{
	    if (*succ_comp == NULL)
		*succ_comp = tmp;
	    else
		{
		    while (!LstEmpty(tmp))
			{
			    graph->p = (unsigned long *) 
				LstDelFirst(tmp, FALSE);
			    for (graph->q = (unsigned long *) 
				 LstBegin(*succ_comp);
				 (graph->q != NULL) && (*(graph->q) !=
							*(graph->p));
				 graph->q = (unsigned long *)
				 LstStep(*succ_comp))
				;
			    if (graph->q == NULL)
				(void) LstInsLast(*succ_comp, (char *)
						  (graph->p));
			    else
				Deallocate((char *) (graph->p));
			}
		    LstDestroy(tmp, TRUE);
		}
	}
    else if (tmp != NULL)
	LstDestroy(tmp, TRUE);
 PROCEED: ;
    if (graph->m > graph->now)
	{
	    graph->m = ((unsigned long) -1L) - graph->m;
	    if (*succ_comp == NULL)
		{
		    *succ_comp = LstCreate(&lngTools);
		    (void) LstInsLast(*succ_comp, 
				      (char *) LngCons(graph->m));
		}
	    else
		{
		    for (graph->q = (unsigned long *)
			 LstBegin(*succ_comp);
			 (graph->q != NULL) &&
			 (*(graph->q) != graph->m);
			 graph->q = (unsigned long *)
			 LstStep(*succ_comp))
			;
		    if (graph->q == NULL)
			(void) LstInsLast(*succ_comp, 
					  (char *) LngCons(graph->m));
		}
	}
    else if (graph->m < min)
	min = graph->m;
    if (i != 0L)
	goto ARROW;
 CONCLUDE: ;
    if (min == node_val)
	{
	    graph->sgc_rec[FIRST_NODE_POS] = node_addr;
	    graph->p = (unsigned long *) LstDelFirst(graph->list, FALSE);
	    if ((*succ_comp == NULL) || LstEmpty(*succ_comp))
		{
		    graph->sgc_rec[SGC_SUCC_COUNT_POS] = 0L;
		    if ((*(graph->p) != node_addr) ||
			(graph->latest_rec[SUCCESSOR_COUNT_POS] != 0L))
			/* If *succ_comp is empty and *(graph->p) is node_addr
			 * then the component is either a terminal node or
			 * a node having itself as only successor node.
			 * graph->latest_rec has thus not been modified
			 * during the while (i) loop. */
			{
			    graph->sgc_rec[FIRST_SGC_SUCC_POS] = 
				graph->glob[LATEST_TERMCOMP_POS];
			    graph->glob[LATEST_TERMCOMP_POS] = 
				graph->glob[SGC_COUNT_POS];
			    (graph->glob[TERMCOMP_COUNT_POS])++;
			}
		    else
			graph->sgc_rec[FIRST_SGC_SUCC_POS] =
			    (unsigned long) -1L;
		}
	    else
		{
		    AwsSkAbs(graph, graph->glob[EOF_ARROWS_POS]);
		    graph->sgc_rec[FIRST_SGC_SUCC_POS] = 
			graph->glob[EOF_ARROWS_POS];
		    i = 0L;
		    if (*succ_comp != NULL)
			{
			    while (!LstEmpty(*succ_comp))
				{
				    graph->q = (unsigned long *) 
					LstDelFirst(*succ_comp, FALSE);
				    AwsWrLng(graph->q, 1, graph);
				    Deallocate((char *) (graph->q));
				    i++;
				}
			}
		    graph->sgc_rec[SGC_SUCC_COUNT_POS] = i;
		    graph->glob[EOF_ARROWS_POS] += i * sizeof(unsigned long);
		}
	    AdrWrLng(graph->sgc_rec, NUM_SGC_FIELDS, graph);
	    graph->latest_rec[graph->sgc_val_pos] =
		((unsigned long) -1L) - graph->glob[SGC_COUNT_POS];
	    graph->latest_rec[graph->left_child_pos] = (unsigned long) -1L;
	    for (;;)
		{
		    NdsSkAbs(graph, 
			     *(graph->p) +
			     graph->sgc_val_pos * sizeof(unsigned long));
		    NdsWrLng(graph->latest_rec + graph->sgc_val_pos,
			     1, graph);
		    NdsSkAbs(graph, 
			     *(graph->p) +
			     graph->left_child_pos * sizeof(unsigned long));
		    NdsWrLng(graph->latest_rec + graph->left_child_pos,
			     1, graph);
		    graph->latest_rec[graph->left_child_pos] = *(graph->p);
		    if (node_addr == *(graph->p))
			{
			    Deallocate((char *) (graph->p));
			    break;
			}
		    Deallocate((char *) (graph->p));
		    graph->p = (unsigned long *) 
			LstDelFirst(graph->list, FALSE);
		}
	    (graph->glob[SGC_COUNT_POS])++;
	    graph->m = graph->latest_rec[graph->sgc_val_pos];
	}
    else
	graph->m = min;
    if (!LstEmpty(graph->vis_stack))
	goto CONT;
}

/***************************************************************************
*
* Function:  GphSgSearch
* Purpose:   This routine is an interface to the Visit routine that
*            computes the strongly connected components of a reachability
*            graph of a net and constructs a directed acyclic graph
*            corresponding to the accessibility relation between the
*            components. There is no need to call this routine if
*            some nodes are incompletely processed or the components
*            have already been computed.
* Arguments: graph - the reachability graph
* Returns:
* See also:  Visit, all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE1(void, GphSgSearch,
    GRAPH *, graph)
{
    LIST *succ_comp;
    
    if ((graph->glob[SGC_COUNT_POS] != 0L) ||
	(graph->glob[NODE_COUNT_POS] == 0L) ||
	(_s_l_e_e_p(graph) &&
	 (graph->glob[TOPMOST_POS] != (unsigned long) -1L)) ||
	((!_s_l_e_e_p(graph)) &&
	 (graph->glob[GEN_COUNT_POS] != graph->glob[NODE_COUNT_POS])))
	Troba1("Illegal GphSgSearch call");
    graph->now = 0L;
    graph->list = LstCreate(&lngTools);
    graph->vis_stack = LstCreate(&strTools);
    succ_comp = NULL;
    AdrSkAbs(graph,
	     graph->glob[NODE_COUNT_POS] * sizeof(unsigned long));
    Visit(graph, GphStart(graph), &succ_comp);
    graph->latest_node = (unsigned long) -1L;
    LstDestroy(graph->list, TRUE);
    LstDestroy(graph->vis_stack, TRUE);
    if (succ_comp != NULL)
	LstDestroy(succ_comp, TRUE);
    if (graph->glob[SGC_COUNT_POS] == 1L)
	{
	    graph->glob[TERMCOMP_COUNT_POS] = 0L;
	    graph->glob[LATEST_TERMCOMP_POS] = ((unsigned long) -1L);
	}
    NdsSkAbs(graph,
	     (unsigned long) (SGC_COUNT_POS * sizeof(unsigned long)));
    NdsWrLng(graph->glob + SGC_COUNT_POS, 3, graph);
    NdsSkAbs(graph,
	     (unsigned long) (EOF_ARROWS_POS * sizeof(unsigned long)));
    NdsWrLng(graph->glob + EOF_ARROWS_POS, 1, graph);
    graph->latest_node = (unsigned long) -1L;
}

/***************************************************************************
*
* Function:  GphSgGet
* Purpose:   This routine returns the nodes in a strongly connected
*            component of a reachability graph of a net.
* Arguments: graph - the reachability graph
*            component - the number of the component
* Returns:   a tree containing the addresses of the nodes in the component
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE2(BALTREE *, GphSgGet,
    GRAPH *, graph,
    unsigned long, component)
{
    unsigned long node_addr = 0L;
    BALTREE *balTree;
    
    if (component >= graph->glob[SGC_COUNT_POS])
	Troba2("Illegal component number for GphSgGet: %lu", component);
    balTree = BalTrCreate(&lngBalTools);
    AdrSkAbs(graph, 
	     (graph->glob[NODE_COUNT_POS] + 
	      component * NUM_SGC_FIELDS + FIRST_NODE_POS) *
	     sizeof(unsigned long));
    AdrReLng(&node_addr, 1, graph);
    while (node_addr != (unsigned long) -1L)
	{
	    (void) BalTrInsert(balTree, (char *) LngCons(node_addr),
			       (char *) (&node_addr), TRUE);
	    NdsSkAbs(graph, 
		     node_addr +
		     graph->left_child_pos * sizeof(unsigned long));
	    NdsReLng(&node_addr, 1, graph);
	}
    return(balTree);
}

/***************************************************************************
*
* Function:  GphSgFlCount
* Purpose:   This routine returns the number of immediate successor
*            components of a strongly connected component in a reachability
*            graph of a net
* Arguments: graph - the reachability graph
*            component - the number of the component
* Returns:   the number explained above
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE2(unsigned long, GphSgFlCount,
    GRAPH *, graph,
    unsigned long, component)
{
    unsigned long count = 0L;
    
    if (component >= graph->glob[SGC_COUNT_POS])
	Troba2("Illegal component number for GphSgFlCount: %lu", component);
    AdrSkAbs(graph, 
	     (graph->glob[NODE_COUNT_POS] +
	      component * NUM_SGC_FIELDS + SGC_SUCC_COUNT_POS) *
	     sizeof(unsigned long));
    AdrReLng(&count, 1, graph);
    return(count);
}

/***************************************************************************
*
* Function:  GphSgSucc
* Purpose:   This routine returns the immediate successor components of
*            a strongly connected component in a reachability graph of
*            a net.
* Arguments: graph - the reachability graph
*            component - the number of the component
* Returns:   a tree of numbers corresponding to the successor components
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE2(BALTREE *, GphSgSucc,
    GRAPH *, graph,
    unsigned long, component)
{
    unsigned long count;
    unsigned long i;
    unsigned long succ = 0L;
    BALTREE *balTree;
    
    if (component >= graph->glob[SGC_COUNT_POS])
	Troba2("Illegal component number for GphSgSucc: %lu", component);
    balTree = BalTrCreate(&lngBalTools);
    AdrSkAbs(graph, 
	     (graph->glob[NODE_COUNT_POS] +
	      component * NUM_SGC_FIELDS + FIRST_SGC_SUCC_POS) *
	     sizeof(unsigned long));
    AdrReLng(graph->sgc_rec + FIRST_SGC_SUCC_POS, 2, graph);
    count = graph->sgc_rec[SGC_SUCC_COUNT_POS];
    if (count == 0L)
	return(balTree);
    AwsSkAbs(graph, graph->sgc_rec[FIRST_SGC_SUCC_POS]);
    for (i = 0L; i < count; i++)
	{
	    AwsReLng(&succ, 1, graph);
	    (void) BalTrInsert(balTree, (char *) LngCons(succ),
			       (char *) (&succ), TRUE);
	}
    return(balTree);
}

/***************************************************************************
*
* Function:  GphTNGet
* Purpose:   This routine returns the terminal nodes of a reachability
*            graph of a net.
* Arguments: graph - the reachability graph
* Returns:   a tree containing the terminal nodes
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE1(BALTREE *, GphTNGet,
    GRAPH *, graph)
{
    unsigned long node_addr, count, i;
    BALTREE *balTree;
    
    balTree = BalTrCreate(&lngBalTools);
    count = graph->glob[TERMNODE_COUNT_POS];
    node_addr = graph->glob[LATEST_TERMNODE_POS];
    for (i = 0L; i < count; i++)
	{
	    (void) BalTrInsert(balTree, (char *) LngCons(node_addr),
			       (char *) (&node_addr), TRUE);
	    if (i + 1L < count)
		{
		    NdsSkAbs(graph, node_addr
			     + FOLLOWERS_POS * sizeof(unsigned long));
		    NdsReLng(&node_addr, 1, graph);
		}
	}
    return(balTree);
}

/***************************************************************************
*
* Function:  GphTmSgCount
* Purpose:   This routine returns the number of nontrivial terminal
*            strongly connected components in a reachability graph
*            of a net if the strongly connected components of the graph
*            have been computed. Otherwise it returns 0L.
* Arguments: graph - the reachability graph
* Returns:
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE1(unsigned long, GphTmSgCount,
    GRAPH *, graph)
{
    return(graph->glob[TERMCOMP_COUNT_POS]);
}

/***************************************************************************
*
* Function:  GphTmSgGet
* Purpose:   This routine returns the nontrivial terminal strongly
*            connected components of a reachability graph of a net.
* Arguments: graph - the reachability graph
* Returns:   a tree containing the numbers of the components
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE1(BALTREE *, GphTmSgGet,
    GRAPH *, graph)
{
    unsigned long component, count, i, initial_addr;
    BALTREE *balTree;
    
    if (graph->glob[SGC_COUNT_POS] == 0L)
	Troba1("Illegal GphTmSgGet call");
    balTree = BalTrCreate(&lngBalTools);
    count = graph->glob[TERMCOMP_COUNT_POS];
    component = graph->glob[LATEST_TERMCOMP_POS];
    initial_addr =
	(graph->glob[NODE_COUNT_POS] + FIRST_SGC_SUCC_POS) *
	    sizeof(unsigned long);
    for (i = 0L; i < count; i++)
	{
	    (void) BalTrInsert(balTree, (char *) LngCons(component),
			       (char *) (&component), TRUE);
	    if (i + 1L < count)
		{
		    AdrSkAbs(graph,
			     initial_addr +
			     component *
			     (NUM_SGC_FIELDS * sizeof(unsigned long)));
		    AdrReLng(&component, 1, graph);
		}
	}
    return(balTree);
}
