/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  graph.c
*
* Part of: PROD's programs araprod, "gphgen", probe, and strong
*
* Authors: Marko Rauhamaa and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     The routines in this module are needed for constructing and
*     inspecting a reachability graph of a net.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "error.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "gphutil.h"
#include "mark.h"
#include "encode.h"
#include "graph.h"
#include "ptnet.h"
#include "araprod.h"
#include "cmdnm.h"

/***************************************************************************
 *
 * Function:  GphIncGenCount
 * Purpose:   This routine increments the number of completely processed
 *            nodes (nodes having all their successors generated).
 *            However, in the case of the sleep set method, the
 *            number of poppings from the search stack is incremented
 *            instead.
 * Arguments: graph - the reachability graph
 * Returns:
 * See also:  all routines having a name of the form Gph[A-Za-z]+
 *
 ****************************************************************************
 */
FE1(void, GphIncGenCount,
    GRAPH *, graph)
{
    if ((graph->glob[GEN_COUNT_POS] < graph->glob[NODE_COUNT_POS]) ||
	(_s_l_e_e_p(graph) &&
	 (graph->glob[GEN_COUNT_POS] < graph->glob[ARROW_COUNT_POS] + 2L)))
	(graph->glob[GEN_COUNT_POS])++;
    else
	Troba1("Illegal GphIncGenCount call");
}

/***************************************************************************
 *
 * Function:  GphGenCount
 * Purpose:   This routine returns the number of completely processed
 *            nodes (nodes having all their successors generated).
 *            However, in the case of the sleep set method, the
 *            number of poppings from the search stack is returned
 *            instead.
 * Arguments: graph - the reachability graph
 * Returns:   the number explained above
 * See also:  all routines having a name of the form Gph[A-Za-z]+
 *
 ****************************************************************************
 */
FE1(unsigned long, GphGenCount,
    GRAPH *, graph)
{
    return(graph->glob[GEN_COUNT_POS]);
}

/***************************************************************************
 *
 * Function:  GphOpen
 * Purpose:   This routine opens an existing reachability graph of a net
 *            for reading or writing.
 * Arguments: name - the name of the graph
 *            reason - this is one of GPH_GENERATE, GPH_COMPS and
 *                     GPH_PROBE, indicating how the graph is used
 *                     (see graph.h)
 *            rotSuff - TRUE if and only if the graph is an auxiliary
 *                      graph (then the suffix of a file name
 *                      is obtained from the default suffix by
 *                      a rotation to left)
 * Returns:   the graph
 * See also:  all routines having a name of the form Gph[A-Za-z]+
 *
 ****************************************************************************
 */
FE3(GRAPH *, GphOpen,
    char *, name,
    int, reason,
    BOOL, rotSuff)
{
    GRAPH *graph;
    char *s;
    BINARYFILE file = NonBinaryFile();
    
    if ((reason != GPH_GENERATE) && (reason != GPH_COMPS) &&
	(reason != GPH_PROBE))
	{
	    (void) fprintf(_err,
			   "\nIllegal reason for GphOpen: %d\n", reason);
	    (void) fflush(_err);
            return(NULL);
	}
    graph = (GRAPH *) Allocate(sizeof(GRAPH));
    graph->reason = reason;
    graph->mmnodes = NULL;
    graph->mmarrows = NULL;
    graph->mmaddrs = NULL;
    graph->mmstack = NULL;
    s = FileName(name, rotSuff ? "phg" : "gph");
    if (reason == GPH_PROBE)
	graph->nodes = OpenRead(s, (BOOL *) NULL);
    else
	graph->nodes = OpenUpdate(s);
    if (NotBinaryFile(graph->nodes))
	{
	    ClDeallocate(s);
	    ClDeallocate(graph);
	    return(NULL);
	}
    ClDeallocate(s);
    graph->glob = (unsigned long *)
        AllocArray (NUM_GRAPH_GLOBALS, sizeof(unsigned long));
    SeekAbs(graph->nodes, (unsigned long) 0L);
    ReadLng(graph->glob, (unsigned) NUM_GRAPH_GLOBALS, graph->nodes);
    if (graph->glob[ARROW_COUNT_POS] + 1L < graph->glob[NODE_COUNT_POS])
	Troba3("GphOpen: internal error, %lu + 1 < %lu",
	       graph->glob[ARROW_COUNT_POS], graph->glob[NODE_COUNT_POS]);
    if (reason == GPH_PROBE)
	graph->now = (graph->glob[NODE_COUNT_POS]
		      + NUM_SGC_FIELDS * graph->glob[SGC_COUNT_POS]) *
	    sizeof(unsigned long);
    if (reason == GPH_GENERATE)
	{
	    if ((graph->glob[NODE_COUNT_POS] != 0L) &&
		((_s_l_e_e_p(graph) &&
		  (graph->glob[TOPMOST_POS] == (unsigned long) -1L)) ||
		 ((!_s_l_e_e_p(graph)) &&
		  (graph->glob[GEN_COUNT_POS] ==
		   graph->glob[NODE_COUNT_POS]))))
		{
		    (void)
			fprintf
			(_err, 
			 "\nReachability graph %s has been completely generated\n",
			 name);
		    (void) fflush(_err);
		    OwnClose(&(graph->nodes));
		    ClDeallocate((graph->glob));
		    ClDeallocate(graph);
		    return(NULL);
		}
	    if (_gerr(graph))
		{
		    (void)
			fputs
			("\nGraph generation will not be continued because\nan illegal path has already been found\n",
			 _err);
		    (void) fflush(_err);
		    OwnClose(&(graph->nodes));
		    ClDeallocate((graph->glob));
		    ClDeallocate(graph);
		    return(NULL);
		}
	    if (_bdd(graph))
		{
		    (void)
			fputs
			("\nBDD-based graph generation will not be continued\n",
			 _err);
		    (void) fflush(_err);
		    OwnClose(&(graph->nodes));
		    ClDeallocate((graph->glob));
		    ClDeallocate(graph);
		    return(NULL);
		}
	    if (_inftr(graph))
		{
		    (void)
			fputs("\nGraph generation will not be continued because\nan infinite path monitor state has been defined\n",
			      _err);
		    (void) fflush(_err);
		    OwnClose(&(graph->nodes));
		    ClDeallocate((graph->glob));
		    ClDeallocate(graph);
		    return(NULL);
		}
	    s = FileName(name, rotSuff ? "shh" : "hsh");
	    if (NotBinaryFile(file = OpenRead(s, (BOOL *) NULL)))
		{
		    OwnClose(&(graph->nodes));
		    ClDeallocate(s);
		    ClDeallocate((graph->glob));
		    ClDeallocate(graph);
		    return(NULL);
		}
	    ClDeallocate(s);
	    graph->hash = HashCreate(graph->glob[HASH_SIZE_POS]);
	    HashRead(graph->hash, graph->glob[HASH_SIZE_POS], file);
	    OwnClose(&file);
	    if (_s_l_e_e_p(graph))
		{
		    s = FileName(name, rotSuff ? "tks" : "stk");
		    if (NotBinaryFile(graph->stack = OpenUpdate(s)))
			{
			    OwnClose(&(graph->nodes));
			    ClDeallocate(s);
			    ClDeallocate((graph->glob));
			    ClDeallocate(graph);
			    return(NULL);
			}
		    ClDeallocate(s);
		}
	}
    else 
	{
	    graph->hash = NULL;
	    if (reason == GPH_COMPS) 
		{
		    if (graph->glob[SGC_COUNT_POS] != 0L)
			{
			    (void)
				fprintf
				(_err,
				 "\nStrongly connected components of %s have already been computed\n",
				 name);
			    (void) fflush(_err);
			    OwnClose(&(graph->nodes));
			    ClDeallocate((graph->glob));
			    ClDeallocate(graph);
			    return(NULL);
			}
		    if (graph->glob[NODE_COUNT_POS] == 0L)
			{
			    (void)
				fprintf
				(_err,
				 "\nComputation of strongly connected components of %s is not allowed\nuntil all nodes of %s have been completely processed\nand there is at least one node in %s\n",
				 name, name, name);
			    (void) fflush(_err);
			    OwnClose(&(graph->nodes));
			    ClDeallocate((graph->glob));
			    ClDeallocate(graph);
			    return(NULL);
			}
		    if ((_s_l_e_e_p(graph) &&
			 (graph->glob[TOPMOST_POS] != (unsigned long) -1L)) ||
			((!_s_l_e_e_p(graph)) &&
			 (graph->glob[GEN_COUNT_POS] !=
			  graph->glob[NODE_COUNT_POS])))
			{
			    (void)
				fprintf
				(_err,
				 "\nComputation of strongly connected components of %s is not allowed\nuntil all nodes of %s have been completely processed\n",
				 name, name);
			    (void) fflush(_err);
			    OwnClose(&(graph->nodes));
			    ClDeallocate((graph->glob));
			    ClDeallocate(graph);
			    return(NULL);
			}
		}
	}
    s = FileName(name, rotSuff ? "wsa" : "aws");
    if (reason == GPH_PROBE)
	graph->arrows = OpenRead(s, (BOOL *) NULL);
    else
	graph->arrows = OpenUpdate(s);
    if (NotBinaryFile(graph->arrows))
	{
	    OwnClose(&(graph->nodes));
	    ClDeallocate(s);
	    ClDeallocate((graph->glob));
	    if (graph->hash != NULL)
		HashDiscard(graph->hash);
	    ClDeallocate(graph);
	    return(NULL);
	}
    ClDeallocate(s);
    s = FileName(name, rotSuff ? "dra" : "adr");
    graph->addrs = OpenUpdate(s);
    if (NotBinaryFile(graph->addrs))
	{
	    OwnClose(&(graph->nodes));
	    OwnClose(&(graph->arrows));
	    ClDeallocate(s);
	    ClDeallocate((graph->glob));
	    if (graph->hash != NULL)
		HashDiscard(graph->hash);
	    ClDeallocate(graph);
	    return(NULL);
	}
    ClDeallocate(s);
    graph->bdd = NULL;
    graph->balTree = NULL;
    if (_dfs(graph))
	{
	    graph->next_lower_pos = FOLLOWERS_POS + 1;
	    if (_inftr(graph))
		{
		    graph->check_pos = graph->next_lower_pos + 1;
		    graph->sgc_val_pos = graph->check_pos + 1;
		}
	    else
		graph->sgc_val_pos = graph->next_lower_pos + 1;
	}
    else
	{
	    graph->next_lower_pos = (unsigned) -1;
	    graph->sgc_val_pos = FOLLOWERS_POS + 1;
	}
    graph->predecessors_pos = graph->sgc_val_pos + 1;
    graph->left_child_pos = graph->predecessors_pos + 1;
    if (_bdd(graph))
	graph->datalen_pos = graph->left_child_pos + 1;
    else
	{
	    graph->right_child_pos = graph->left_child_pos + 1;
	    graph->datalen_pos = graph->right_child_pos + 1;
	}
    graph->num_node_fields = graph->datalen_pos + 1;
    if (_frexpansion(graph))
	graph->aw_datalen_pos = AW_NEXT_POS + 1;
    else
	graph->aw_datalen_pos = TARGET_POS + 1;
    graph->num_arrow_fields = graph->aw_datalen_pos + 1;
    graph->latest_node = (unsigned long) -1L;
    graph->latest_rec = (unsigned long *)
        AllocArray(graph->num_node_fields, sizeof(unsigned long));
    if (reason == GPH_COMPS)
	graph->latest_contents = NULL;
    else if (reason == GPH_GENERATE)
	graph->latest_contents = (DATATYPE *)
	    AllocArray((unsigned) (graph->glob[BUF_SIZE_POS]),
		       sizeof(DATATYPE));
    else
	graph->latest_contents = (DATATYPE *)
	    AllocArray(1 + (unsigned) (graph->glob[MAX_DATALEN_POS]),
		       sizeof(DATATYPE));
    graph->arrow_rec = (unsigned long *)
	AllocArray(graph->num_arrow_fields, sizeof(unsigned long));
    graph->sgc_rec = (unsigned long *)
	AllocArray(NUM_SGC_FIELDS, sizeof(unsigned long));
    graph->name = StrCopy(name);
    if (reason == GPH_GENERATE)
	{
	    SeekAbs(graph->arrows, graph->glob[EOF_ARROWS_POS]);
	    SeekAbs(graph->addrs, 
		    graph->glob[NODE_COUNT_POS] * sizeof(unsigned long));
	}
    if ((!rotSuff) && ((GPH_GENERATE == reason) || (GPH_COMPS == reason)))
        {
            s = FileName(name, "cad");
            (void) unlink(s);
	    ClDeallocate(s);
            s = FileName(name, "caw");
            (void) unlink(s);
	    ClDeallocate(s);
        }
    return(graph);
}

/***************************************************************************
 *
 * Function:  GphFlush
 * Purpose:   This routine updates files associated with a reachability
 *            graph of a net.
 * Arguments: graph - the reachability graph
 * Returns:
 * See also:  all routines having a name of the form Gph[A-Za-z]+
 *
 ****************************************************************************
 */
FE1(void, GphFlush,
    GRAPH *, graph)
{
    NdsSkAbs(graph, (unsigned long) 0L);
    NdsWrLng(graph->glob, NUM_GRAPH_GLOBALS, graph);
    if (graph->mmnodes)
	MapFlush(graph->mmnodes);
    if (graph->mmarrows)
	MapFlush(graph->mmarrows);
    if (graph->mmaddrs)
	MapFlush(graph->mmaddrs);
    if (graph->mmstack)
	MapFlush(graph->mmstack);
#ifndef UNIX
    OwnFlush(graph->nodes);
    OwnFlush(graph->arrows);
    OwnFlush(graph->addrs);
    OwnFlush(graph->stack);
#endif
}

/***************************************************************************
 *
 * Function:  GphMapToMem
 * Purpose:   If this routine is called, the construction of a reachability
 *            graph of a net takes place in the virtual memory, instead of
 *            the file system. If a space reservation for a buffer
 *            corresponding to a file fails, the file is manipulated
 *            directly. Direct manipulation of a file is also started
 *            if the corresponding buffer becomes full. A buffer that
 *            has become fill is not expanded because an excessive use
 *            of the virtual memory may have serious side effects in
 *            a system where many applications are run simultaneously.
 * Arguments: graph - the reachability graph
 * Returns:
 * See also:  all routines having a name of the form Gph[A-Za-z]+
 *
 ****************************************************************************
 */
FE2(void, GphMapToMem,
    GRAPH *, graph,
    unsigned long, len)
{
    char *s;

    s = FileName(graph->name, _rotsuff(graph) ? "phg" : "gph");
    if (graph->reason == GPH_GENERATE)
	graph->mmnodes =
	    MapToMem(s, graph->nodes, len, graph->glob[EOF_NODES_POS]);
    else
	graph->mmnodes =
	    MapToMem(s, graph->nodes, graph->glob[EOF_NODES_POS],
		     graph->glob[EOF_NODES_POS]);
    if (_s_l_e_e_p(graph) && (graph->reason == GPH_GENERATE))
	graph->mmstack =
	    MapToMem(FileName(graph->name, _rotsuff(graph) ? "tks" : "stk"),
		     graph->stack, len, graph->glob[EOF_STACK_POS]);
    s = FileName(graph->name, _rotsuff(graph) ? "dra" : "adr");
    if (graph->reason == GPH_GENERATE)
	graph->mmaddrs =
	    MapToMem(s, graph->addrs,
		     (len == 0L) ? 0L
		     : 1L + (len - 1L) / graph->num_node_fields,
		     ((graph->glob[NODE_COUNT_POS] + NUM_SGC_FIELDS
		       * graph->glob[SGC_COUNT_POS])
		      * sizeof(unsigned long)));
    else if (graph->reason == GPH_PROBE)
	graph->mmaddrs =
	    MapToMem(s, graph->addrs,
		     (graph->glob[NODE_COUNT_POS] * 5
		      + NUM_SGC_FIELDS
		      * graph->glob[SGC_COUNT_POS])
		     * sizeof(unsigned long),
		     (graph->glob[NODE_COUNT_POS]
		      + NUM_SGC_FIELDS
		      * graph->glob[SGC_COUNT_POS])
		     * sizeof(unsigned long));
    else
	graph->mmaddrs =
	    MapToMem(s, graph->addrs,
		     (graph->glob[NODE_COUNT_POS]
		      + NUM_SGC_FIELDS
		      * graph->glob[NODE_COUNT_POS])
		     * sizeof(unsigned long),
		     graph->glob[NODE_COUNT_POS]
		     * sizeof(unsigned long));
    s = FileName(graph->name, _rotsuff(graph) ? "wsa" : "aws");
    if (graph->reason == GPH_GENERATE)
	graph->mmarrows =
	    MapToMem(s, graph->arrows, len, graph->glob[EOF_ARROWS_POS]);
    else if (graph->reason == GPH_PROBE)
	graph->mmarrows =
	    MapToMem(s, graph->arrows, graph->glob[EOF_ARROWS_POS],
		     graph->glob[EOF_ARROWS_POS]);
    else
	graph->mmarrows =
	    MapToMem(s, graph->arrows, 2L * graph->glob[EOF_ARROWS_POS],
		     graph->glob[EOF_ARROWS_POS]);
    if (graph->reason != GPH_GENERATE)
	return;
    AwsSkAbs(graph, graph->glob[EOF_ARROWS_POS]);
    AdrSkAbs(graph, 
	     graph->glob[NODE_COUNT_POS] * sizeof(unsigned long));
}

/***************************************************************************
 *
 * Function:  GphEsolc
 * Purpose:   This routine closes a reachability graph of a net. If the
 *            graph was under construction and it is possible to continue
 *            the construction later, the hash table associated with the
 *            graph is written to a file.
 * Arguments: graph - the reachability graph
 * Returns:
 * See also:  all routines having a name of the form Gph[A-Za-z]+
 *
 ****************************************************************************
 */
FE1(void, GphEsolc,
    GRAPH *, graph)
{
    char *s;
    BINARYFILE file = NonBinaryFile();

    if (!graph)
	return;
    if (graph->mmarrows)
	Unmap(graph->mmarrows, graph->reason != GPH_PROBE);
    if (graph->mmaddrs)
	Unmap(graph->mmaddrs, graph->reason != GPH_PROBE);
    OwnClose(&(graph->arrows));
    OwnClose(&(graph->addrs));
    if (graph->reason == GPH_GENERATE)
	{
	    NdsSkAbs((graph), (unsigned long) 0L);
	    NdsWrLng(graph->glob, NUM_GRAPH_GLOBALS, graph);
	    if (graph->mmnodes)
		Unmap(graph->mmnodes, TRUE);
	    OwnClose(&(graph->nodes));
	    if (_s_l_e_e_p(graph))
		{
		    if (graph->mmstack &&
			((graph->glob[NODE_COUNT_POS] == 0L) ||
			 (graph->glob[TOPMOST_POS] != (unsigned long) -1L)))
			Unmap(graph->mmstack,
			      (!_gerr(graph)) && (!_inftr(graph)) &&
			      (!_bdd(graph)));
		    OwnClose(&(graph->stack));
		}
	    if (_bdd(graph))
		{
		    if (graph->balTree != NULL)
			{
			    BalTrDestroy(graph->balTree, TRUE);
			    graph->balTree = NULL;
			}
		    if (graph->bdd != NULL)
			{
			    BddSList();
			    BddCat(graph->bdd);
			    BddDList();
			    graph->bdd = NULL;
			}
		}
	    else if (graph->hash != NULL)
		{
		    s = FileName(graph->name, _rotsuff(graph) ? "shh" : "hsh");
		    if ((!_gerr(graph)) && (!_inftr(graph)) &&
			((graph->glob[NODE_COUNT_POS] == 0L) ||
			 (_s_l_e_e_p(graph) &&
			  (graph->glob[TOPMOST_POS] != (unsigned long) -1L)) ||
			 ((!_s_l_e_e_p(graph) &&
			   (graph->glob[GEN_COUNT_POS] !=
			    graph->glob[NODE_COUNT_POS])))))
			{
			    if (NotBinaryFile(file = CreateWrite(s)))
				{
				    (void)
					fputs
					("No panic. The graph itself should be OK.\n", 
					 _err);
				    (void) fflush(_err);
				}
			    else
				{
				    HashWrite(graph->hash,
					      graph->glob[HASH_SIZE_POS],
					      file);
				    OwnClose(&file);
				}
			}
		    else
			(void) unlink(s);
		    ClDeallocate(s);
		    HashDiscard(graph->hash);
		}
	}
    else
	{
	    if (graph->mmnodes)
		Unmap(graph->mmnodes, graph->reason != GPH_PROBE);
	    OwnClose(&(graph->nodes));
	}
    ClDeallocate((graph->glob));
    ClDeallocate((graph->latest_rec));
    if (graph->latest_contents != NULL)
	ClDeallocate((graph->latest_contents));
    ClDeallocate((graph->arrow_rec));
    ClDeallocate((graph->sgc_rec));
    ClDeallocate(graph->name);
    if (graph == theGraph)
	theGraph = (GRAPH *) NULL;
    if (graph == the2Graph)
	the2Graph = (GRAPH *) NULL;
    ClDeallocate(graph);
}

/***************************************************************************
 *
 * Function:  GphBufSize
 * Purpose:   This routine returns the size of a special encoding buffer
 *            used in the construction of a reachability graph of a net.
 * Arguments: graph - the reachability graph
 * Returns:   the size explained above
 * See also:  all routines having a name of the form Gph[A-Za-z]+
 *
 ****************************************************************************
 */
FE1(unsigned, GphBufSize,
    GRAPH *, graph)
{
    return((unsigned) (graph->glob[BUF_SIZE_POS]));
}

/***************************************************************************
 *
 * Function:  GphBits
 * Purpose:   This routine returns the hash_bits that was chosen for
 *            a reachability graph of a net by GphCreate.
 * Arguments: graph - the reachability graph
 * Returns:   see above
 * See also:  all routines having a name of the form Gph[A-Za-z]+
 *
 ****************************************************************************
 */
FE1(unsigned, GphBits,
    GRAPH *, graph)
{
    return((unsigned) (graph->glob[HASH_BITS_POS]));
}

/***************************************************************************
 *
 * Function:  GphSeed
 * Purpose:   This routine returns the seed number of a random number
 *            generator associated with a reachability graph of a net.
 * Arguments: graph - the reachability graph
 * Returns:   the number explained above
 * See also:  all routines having a name of the form Gph[A-Za-z]+
 *
 ****************************************************************************
 */
FE1(UNS32, GphSeed,
    GRAPH *, graph)
{
    return((UNS32) (graph->glob[SEED_POS]));
}

/***************************************************************************
 *
 * Function:  GphSetSeed
 * Purpose:   This routine sets the seed number of a random number generator
 *            associated with a reachability graph of a net.
 * Arguments: graph - the reachability graph
 *            seed - the new seed value
 * Returns:
 * See also:  all routines having a name of the form Gph[A-Za-z]+
 *
 ****************************************************************************
 */
FE2(void, GphSetSeed,
    GRAPH *, graph,
    UNS32, seed)
{
    graph->glob[SEED_POS] = (unsigned long) seed;
}

/***************************************************************************
 *
 * Function:  GphUnfold
 * Purpose:   This routine returns the unfold that was chosen for
 *            a reachability graph of a net by GphCreate.
 * Arguments: graph - the reachability graph
 * Returns:   see above
 * See also:  all routines having a name of the form Gph[A-Za-z]+
 *
 ****************************************************************************
 */
FE1(BOOL, GphUnfold,
    GRAPH *, graph)
{
    return(_unfold(graph) != 0L);
}

/***************************************************************************
 *
 * Function:  GphSymm
 * Purpose:   This routine returns the symm that was chosen for
 *            a reachability graph of a net by GphCreate.
 * Arguments: graph - the reachability graph
 * Returns:   see above
 * See also:  all routines having a name of the form Gph[A-Za-z]+
 *
 ****************************************************************************
 */
FE1(BOOL, GphSymm,
    GRAPH *, graph)
{
    return(_symm(graph) != 0L);
}

/***************************************************************************
 *
 * Function:  GphIMkSymm
 * Purpose:   This routine returns the iMkSymm that was chosen for
 *            a reachability graph of a net by GphCreate.
 * Arguments: graph - the reachability graph
 * Returns:   see above
 * Purpose:
 * Arguments: graph - the reachability graph
 * Returns:
 * See also:  all routines having a name of the form Gph[A-Za-z]+
 *
 ****************************************************************************
 */
FE1(BOOL, GphIMkSymm,
    GRAPH *, graph)
{
    return(_imksymm(graph) != 0L);
}

/***************************************************************************
 *
 * Function:  GphStub
 * Purpose:   This routine returns the stub that was chosen for
 *            a reachability graph of a net by GphCreate.
 * Arguments: graph - the reachability graph
 * Returns:   see above
 * See also:  all routines having a name of the form Gph[A-Za-z]+
 *
 ****************************************************************************
 */
FE1(BOOL, GphStub,
    GRAPH *, graph)
{
    return(_stub(graph) != 0L);
}

/***************************************************************************
 *
 * Function:  GphDelAlg
 * Purpose:   This routine returns the delAlg that was chosen for
 *            a reachability graph of a net by GphCreate.
 * Arguments: graph - the reachability graph
 * Returns:   see above
 * See also:  all routines having a name of the form Gph[A-Za-z]+
 *
 ****************************************************************************
 */
FE1(BOOL, GphDelAlg,
    GRAPH *, graph)
{
    return(_delalg(graph) != 0L);
}

/***************************************************************************
 *
 * Function:  GphCFFD
 * Purpose:   This routine returns the cffd that was chosen for
 *            a reachability graph of a net by GphCreate.
 * Arguments: graph - the reachability graph
 * Returns:   see above
 * See also:  all routines having a name of the form Gph[A-Za-z]+
 *
 ****************************************************************************
 */
FE1(BOOL, GphCFFD,
    GRAPH *, graph)
{
    return(_cffd(graph) != 0L);
}

/***************************************************************************
 *
 * Function:  GphDefVis
 * Purpose:   This routine returns the defVis that was chosen for
 *            a reachability graph of a net by GphCreate.
 * Arguments: graph - the reachability graph
 * Returns:   see above
 * See also:  all routines having a name of the form Gph[A-Za-z]+
 *
 ****************************************************************************
 */
FE1(BOOL, GphDefVis,
    GRAPH *, graph)
{
    return(_defvis(graph) != 0L);
}

/***************************************************************************
 *
 * Function:  GphSleep
 * Purpose:   This routine returns the sleep that was chosen for
 *            a reachability graph of a net by GphCreate.
 * Arguments: graph - the reachability graph
 * Returns:   see above
 * See also:  all routines having a name of the form Gph[A-Za-z]+
 *
 ****************************************************************************
 */
FE1(BOOL, GphSleep,
    GRAPH *, graph)
{
    return(_s_l_e_e_p(graph) != 0L);
}

/***************************************************************************
 *
 * Function:  GphDfs
 * Purpose:   This routine returns the dfs that was chosen for
 *            a reachability graph of a net by GphCreate.
 * Arguments: graph - the reachability graph
 * Returns:   see above
 * See also:  all routines having a name of the form Gph[A-Za-z]+
 *
 ****************************************************************************
 */
FE1(BOOL, GphDfs,
    GRAPH *, graph)
{
    return(_dfs(graph) != 0L);
}

/***************************************************************************
 *
 * Function:  GphFreeExpansion
 * Purpose:   This routine returns the freeExpansion that was chosen for
 *            a reachability graph of a net by GphCreate.
 * Arguments: graph - the reachability graph
 * Returns:   see above
 * See also:  all routines having a name of the form Gph[A-Za-z]+
 *
 ****************************************************************************
 */
FE1(BOOL, GphFreeExpansion,
    GRAPH *, graph)
{
    return(_frexpansion(graph) != 0L);
}

/***************************************************************************
 *
 * Function:  GphErr
 * Purpose:   This routine returns TRUE if on-the-fly verification has
 *            been done and an error has been found during the construction
 *            of a reachability graph of a net. Otherwise it returns FALSE.
 * Arguments: graph - the reachability graph
 * Returns:   see above
 * See also:  all routines having a name of the form Gph[A-Za-z]+
 *
 ****************************************************************************
 */
FE1(BOOL, GphErr,
    GRAPH *, graph)
{
    return(_gerr(graph) != 0L);
}

GRAPH *_graph[2] = { (GRAPH *) NULL, (GRAPH *) NULL };
/* For a net, _graph[0] is the reachability graph
 * to be generated. If a linear time temporal
 * formula is to be verified with the aid of
 * stubborn sets, then an auxiliary graph,
 * _graph[1], is constructed simultaneously
 * with _graph[0]. */
unsigned long _ndAddr;  /* An address of a node in a reachability graph
                         * of a net. */
unsigned long _ndFirst; /* The address of the target node of the first
                         * immediate successor arrow of a node in a
                         * reachability graph of a net. */
unsigned long _awCount; /* The number of immediate successor arrows of
                         * a node in a reachability graph of a net. */
unsigned long _awAddr;  /* The address of an immediate successor arrow
                         * of a node in a reachability graph of a net. */ 
unsigned long _awFirst; /* The address of the first immediate successor
                         * arrow of a node in a reachability graph of
                         * a net. */
LIST *_rootList = NULL;        /* A list of root nodes of LTS's. */
LIST *_actList = NULL;         /* A list of actions. */
ACTION *_iAction = NULL; /* This is an invisible action for all LTS's. For
                          * compilation reasons, the external declaration
                          * of this variable is in araprod.h. */
PTRBL *_arpFoldings = NULL; 
/* This table maps each transition of a low level
 * net to the corresponding action. */
unsigned _iCode;        /* This is a special action number reserved
                         * for _iAction. */
char *_iName = NULL;    /* This is a special action name reserved
                         * for _iAction. */
BOOL _araProd;          /* As far as reachability graph generation is
                         * concerned, this flag is TRUE if and only if
                         * the araprod program is being run .*/

/***************************************************************************
 *
 * Function:  LTSOutput
 * Purpose:   This routine constructs an LTS corresponding to
 *            a reachability graph of a net.
 * Arguments: graph - the reachability graph
 * Returns:
 * See also:  main (araprod.c), LTSBegin (ptelem.c),
 *            PtAraOutput (ptelem3.c), LTSEnd (ptelem.c),
 *            AraStep (inst.c), TrnFire (ptelem.c)
 *
 ****************************************************************************
 */
FE1(void, LTSOutput,
    GRAPH *, graph)
{
    unsigned i;
    unsigned long ndCount;
    unsigned long node;
    unsigned long ordNode;
    unsigned long awCount;
    unsigned long arrow = 0L;
    unsigned long ordArrow;
    unsigned long succ = 0L;
    BOOL faithful;
    ACTION *action;
    char *name;
    FILE *file;
    register DATATYPE *pointer;
    register unsigned long a;

    name = FileName(graph->name, "lts");
    file = MyFopen(name, "w", (BOOL *) NULL);
    ClDeallocate(name);
    if (!file)
	Exit(2);
    if (!RootList())
	{
	    (void)
		fprintf
		(file,
		 "LTS |B|\nROOT |0|\nNODE_COUNT |1|\nARC_COUNT |0|\nSYMBOL_COUNT |0|\nHISTORY |created by %s|\nEND_HEADER\nSYMBOLS\nEND_SYMBOLS\nARCS\nEND_ARCS\nNODES\nEND_NODES\nEND_LTS\n",
		 myName);
	    MyFclose(file);
	    return;
	}
    if (IAction())
	i = 1;
    else
	{
	    i = 0;
	    LSTFOR (ActList(), action, ACTION)
		    {
			if (ActHidden(action))
			    {
				SetIAction(action);
				i = 1;
				break;
			    }
		    }
	}
    LSTFOR (ActList(), action, ACTION)
	    {
		if ((action != IAction()) && !ActHidden(action))
		    i++;
	    }
    ndCount = GphNdCount(graph);
    (void)
	fprintf
	(file,
	 "LTS |B|\nROOT |%lu|\nNODE_COUNT |%lu|\nARC_COUNT |%lu|\nSYMBOL_COUNT |%u|\nHISTORY |created by %s|\nEND_HEADER\nSYMBOLS\n",
	 GphOrdinal(graph, GphStart(graph)), ndCount,
	 GphGlAwCount(graph), i, myName);
    i = 0;
    LSTFOR (ActList(), action, ACTION)
	    {
		if (i == ICode())
		    {
			if (IAction())
			    (void) fprintf(file, "%u %s\n", ICode(), IName());
			i++;
		    }
		if ((action == IAction()) || ActHidden(action))
		    continue;
		(void) fprintf(file, "%u %s\n", i, ActGName(action));
		i++;
	    }
    (void) fputs("END_SYMBOLS\nARCS\n", file);
    faithful = TRUE;
    for (ordNode = 0L; ordNode < ndCount; ordNode++)
	{
	    node = GphId(graph, ordNode);
	    awCount = GphAwCount(graph, node);
	    if ((awCount == 0L) || (awCount == (unsigned long) -1L))
		continue;
	    GphFlFirst(graph, node, &arrow, &succ);
	    if (_frexpansion(graph))
		{
		    for (ordArrow = awCount; ordArrow != 0L;)
			{
			    ordArrow--;
			    pointer = GphArrow(graph, arrow);
			    GetValue(unsigned long, a, pointer);
			    if (_symm(graph))
				{
				    if (a & 1L)
					faithful = FALSE;
				    else
					faithful = TRUE;
				    a >>= 1;
				}
			    if (a >= PtrBlGCount(ArpFoldings()))
				Troba3
				    ("Illegal data for LTSOutput: %lu >= %lu",
				     a, PtrBlGCount(ArpFoldings()));
			    (void)
				fprintf
				(file, "%lu %lu %u",
				 ordNode, GphOrdinal(graph, succ),
				 ActGNumber((ACTION *)
					    PtrBlGet(ArpFoldings(), a)));
			    if (!faithful)
				(void)
				    fprintf
				    (file,
				     " -- arc explained by the state equivalence relation");
			    (void) fputc((int) '\n', file);
			    if (ordArrow != 0L)
				GphFlNext(graph, &arrow, &succ);
			}
		}
	    else
		{
		    for (ordArrow = 0L; ordArrow < awCount; ordArrow++)
			{
			    pointer = GphArrow(graph, arrow);
			    GetValue(unsigned long, a, pointer);
			    if (_symm(graph))
				{
				    if (a & 1L)
					faithful = FALSE;
				    else
					faithful = TRUE;
				    a >>= 1;
				}
			    if (a >= PtrBlGCount(ArpFoldings()))
				Troba3
				    ("Illegal data for LTSOutput: %lu >= %lu",
				     a, PtrBlGCount(ArpFoldings()));
			    (void)
				fprintf
				(file, "%lu %lu %u",
				 ordNode, GphOrdinal(graph, succ),
				 ActGNumber((ACTION *)
					    PtrBlGet(ArpFoldings(), a)));
			    if (!faithful)
				(void)
				    fprintf
				    (file,
				     " -- arc explained by the state equivalence relation");
			    (void) fputc((int) '\n', file);
			    if (ordArrow + 1L < awCount)
				GphFlNext(graph, &arrow, &succ);
			}
		}
	}
    (void) fputs("END_ARCS\nNODES\nEND_NODES\nEND_LTS\n", file);
    MyFclose(file);
}

/****************************************************************************
 *
 * Function:  ExOnTheFly
 * Purpose:   This routine is used for exit when on-the-fly verification
 *            has found an error. The error, also reported by this routine,
 *            is either a loop or a node.
 * Arguments: graph - a reachability graph of a net
 *            path - an erroneous path in the graph
 *            loop - this is TRUE if and only if the path represents a loop
 * Returns:
 * See also:  IllTrace (inst.c), IllStabFail (inst.c), IllDivTrace (inst.c),
 *            IllInfTrace (inst.c), PtC1 (ptnet.c), PtC2 (ptnet.c),
 *            PtC3 (ptnet.c), PtC4 (ptnet.c)
 *
 ****************************************************************************
 */
FE3(void, ExOnTheFly,
    GRAPH *, graph,
    PATH *, path,
    BOOL, loop)
{
    BINARYFILE bFile = NonBinaryFile();
    char *str;
    STEP *step;
    register DATATYPE *pointer;
    register unsigned long a;
    ACTION *action;

    if (loop)
	{
	    (void) fputs("Loop ", _out);
	    LSTNEST(PthGSteps(path), step, STEP)
		    {
			if (StpGWorld(step) == PthEnd(path))
			    {
				(void)
				    fprintf
				    (_out, "%lu [%lu> ",
				     GphOrdinal(graph, StpGWorld(step)),
				     GphAwOrdinal(graph, StpGAwAddr(step)));
				break;
			    }
		    }
	    LSTUNNEST (PthGSteps(path), step, STEP)
		(void) fprintf(_out, "%lu [%lu> ",
			       GphOrdinal(graph, StpGWorld(step)),
			       GphAwOrdinal(graph, StpGAwAddr(step)));
	    if (AraProd())
		{
		    (void) fprintf(_out, "%lu\nThe same for ARA users:\n",
				   GphOrdinal(graph, PthEnd(path)));
		    LSTNEST(PthGSteps(path), step, STEP)
			    {
				if (StpGWorld(step) == PthEnd(path))
				    {
					pointer = GphArrow(graph,
							   StpGAwAddr(step));
					GetValue(unsigned long, a, pointer);
					if (a >= PtrBlGCount(ArpFoldings()))
					    Troba3
						("Illegal data for ExOnTheFly: %lu >= %lu",
						 a, PtrBlGCount(ArpFoldings()));
					action = (ACTION *)
					    PtrBlGet(ArpFoldings(), a);
					(void)
					    fprintf
					    (_out, "%lu %s(%u)\n",
					     GphOrdinal(graph,
							StpGWorld(step)),
					     ActGName(action),
					     ActGNumber(action));
					break;
				    }
			    }
		    LSTUNNEST (PthGSteps(path), step, STEP)
			    {
				pointer = GphArrow(graph,
						   StpGAwAddr(step));
				GetValue(unsigned long, a, pointer);
				if (a >= PtrBlGCount(ArpFoldings()))
				    Troba3
					("Illegal data for ExOnTheFly: %lu >= %lu",
					 a, PtrBlGCount(ArpFoldings()));
				action = (ACTION *)
				    PtrBlGet(ArpFoldings(), a);
				(void)
				    fprintf
				    (_out, "%lu %s(%u)\n",
				     GphOrdinal(graph, StpGWorld(step)),
				     ActGName(action),
				     ActGNumber(action));
			    }
		}
	}
    else
	(void) fputs("Node ", _out);
    (void)
	fprintf
	(_out,
	 "%lu\nFor more information, start \"probe %s\"\nand look at the set ",
	 GphOrdinal(graph, PthEnd(path)), graph->name);
    SetDenote(_out, (unsigned long) 1L);
    (void) fputc((int) '\n', _out);
    (void) fflush(_out);
    if (AraProd())
	LTSOutput(graph);
    str = FileName(graph->name, "err");
    _serr(graph);
    if ((((GRAPH *) NULL) != the2Graph) && (graph != the2Graph))
	GphClose(the2Graph);
    GphClose(graph);
    bFile = CreateWrite(str);
    ClDeallocate(str);
    if (NotBinaryFile(bFile))
	Exit(5);
    if (loop)
	{
	    LSTFOR (PthGSteps(path), step, STEP)
		    {
			WriteLng(StpAWorld(step), (unsigned) 1, bFile);
			WriteLng(StpAAwAddr(step), (unsigned) 1, bFile);
			if (StpGWorld(step) == PthEnd(path))
			    break;
		    }
	}
    WriteLng(PthAEnd(path), (unsigned) 1, bFile);
    PthSEnd(path, (unsigned long) -1L);
    WriteLng(PthAEnd(path), (unsigned) 1, bFile);
    PthDiscard(path);
    OwnClose(&bFile);
    Exit(1);
}

FE10(PATH *, GphShortest,
     GRAPH *, graph,
     unsigned long *, present,
     unsigned long *, head,
     unsigned long *, tail,
     PATH **, path,
     PATH **, next,
     BOOL *, cont,
     VOID *, formula,
     FOOL, along,
     PTHF, admit)
{
    unsigned long counter = 0L;
    unsigned long awCount;
    unsigned long awAddr;
    unsigned long back;
    unsigned long succ;
    unsigned long dist = 0L;
    unsigned long refDist = 0L;

    *next = (PATH *) NULL;
    NdsSkAbs(graph, *present);
    NdsReLng(graph->latest_rec, graph->num_node_fields, graph);
    graph->latest_node = (unsigned long) -1L;
    succ =
	graph->now + graph->latest_rec[ORDINAL_POS] * sizeof(unsigned long);
    if (((PATH *) NULL) == *path)
	{
	    *head = graph->glob[NODE_COUNT_POS] * sizeof(unsigned long);
	    *tail = *head;
	    AdrSkAbs(graph, graph->now);
	    dist = (unsigned long) ~0L;
	    refDist = (unsigned long) ~(graph->glob[NODE_COUNT_POS]);
	    for (counter = 0L; counter != graph->glob[NODE_COUNT_POS];
		 counter++)
		{
		    if (!*cont)
			goto E_N_D;
		    if (counter == graph->latest_rec[ORDINAL_POS])
			{ AdrWrLng(&dist, 1, graph); }
		    else
			{ AdrWrLng(&refDist, 1, graph);	}
		}
	    while (counter--)
		{
		    if (!*cont)
			goto E_N_D;
		    AdrWrLng(&(graph->latest_node), 1, graph);
		}
	}
    else
	{
	    AdrSkAbs(graph, succ);
	    AdrReLng(&dist, 1, graph);
	}
    for (;;)
	{
	    if (!*cont)
		goto E_N_D;
	    dist = (unsigned long) ~dist;
	    if (dist > graph->glob[NODE_COUNT_POS])
		Troba1("GphShortest: internal error, unexpected dist");
	    awCount = graph->latest_rec[SUCCESSOR_COUNT_POS];
	    awAddr = graph->latest_rec[FOLLOWERS_POS];
	    back = graph->latest_rec[graph->predecessors_pos];
	    graph->now +=
		graph->glob[NODE_COUNT_POS] * sizeof(unsigned long) * 2L;
	    *next = (*admit)(formula, graph, *present);
	    graph->now -=
		graph->glob[NODE_COUNT_POS] * sizeof(unsigned long) * 2L;
	    if (((PATH *) NULL) != *next)
		{
		    if (((PATH *) NULL) == *path)
			{
			    *path = PthCreate(*present);
			    while (dist--)
				{
				    for (;;)
					{
					    if (!*cont)
						goto E_N_D;
					    if (graph->glob[EOF_ARROWS_POS]
						<= back)
						Troba4
						    ("GphShortest: internal error, back %lu >= %lu, dist %lu",
						     back,
						     graph->glob[EOF_ARROWS_POS],
						     dist);
					    AwsSkAbs(graph, back +
						     SOURCE_POS *
						     sizeof(unsigned long));
					    AwsReLng(graph->arrow_rec +
						     SOURCE_POS, 2, graph);
					    if (graph->glob[EOF_NODES_POS]
						<=
						graph->arrow_rec[SOURCE_POS]
						)
						Troba3
						    ("GphShortest: internal error, pred %lu >= %lu",
						     graph->arrow_rec[SOURCE_POS],
						     graph->glob[EOF_NODES_POS]);
					    NdsSkAbs(graph,
						     graph->arrow_rec[SOURCE_POS]);
					    NdsReLng(graph->latest_rec,
						     graph->num_node_fields,
						     graph);
					    graph->latest_node =
						(unsigned long) -1L;
					    AdrSkAbs(graph,
						     graph->now +
						     graph->latest_rec[ORDINAL_POS] *
						     sizeof(unsigned long));
					    AdrReLng(&refDist, 1, graph);
#if 0
					    (void)
						fprintf
						(_out,
						 "ndaddr %lu, ord %lu, pred %lu, awaddr %lu, back %lu, refDist %lu, dist %lu\n",
						 graph->arrow_rec[SOURCE_POS],
						 graph->latest_rec[ORDINAL_POS],
						 graph->latest_rec[graph->predecessors_pos],
						 back,
						 graph->arrow_rec[AW_BACK_POS],
						 refDist, dist);
#endif
					    if (refDist == dist)
						break;
					    back =
						graph->arrow_rec[AW_BACK_POS];
					}
				    PthInsFirst(*path,
						graph->arrow_rec[SOURCE_POS],
						back);
				    back =
					graph->latest_rec[graph->predecessors_pos];
				    if (!(*along)(formula, graph,
						  graph->
						  arrow_rec[SOURCE_POS]))
					Troba1
					    ("GphShortest: internal error, unexpexted along");
				}
			}
		    return(PthCatenate(PthCopy(*path), *next));
		}
	    if (((PATH *) NULL) != *path)
		PthDiscard(*path);
	    if (awCount == (unsigned long) -1L)
		awCount = 0L;
	    else if (awCount != 0L)
		{
		    graph->now +=
			graph->glob[NODE_COUNT_POS] *
			sizeof(unsigned long) * 2L;
		    if (!(*along)(formula, graph, *present))
			awCount = 0L;
		    graph->now -=
			graph->glob[NODE_COUNT_POS] *
			sizeof(unsigned long) * 2L;
		}
	    AdrSkAbs(graph, succ);
	    if (0L == awCount)
		{
		    refDist = graph->glob[NODE_COUNT_POS] + 1L;
		    AdrWrLng(&refDist, 1, graph);
		}
	    else
		{
		    AdrWrLng(&dist, 1, graph);
		    refDist = dist + 1L;
		    AwsSkAbs(graph,
			     awAddr + TARGET_POS * sizeof(unsigned long));
		    AwsReLng(&succ, 1, graph);
		    do
			{
			    if (!*cont)
				goto E_N_D;
			    awCount--;
			    NdsSkAbs(graph,
				     succ + ORDINAL_POS *
				     sizeof(unsigned long));
			    NdsReLng(&counter, 1, graph);
			    AdrSkAbs(graph,
				     graph->now + counter *
				     sizeof(unsigned long));
			    AdrReLng(&dist, 1, graph);
			    if (dist > (unsigned long) ~dist)
				{
				    if (graph->glob[NODE_COUNT_POS] ==
					(unsigned long) ~dist)
					{
					    AdrSkAbs(graph,
						     graph->now + *tail);
					    AdrWrLng(&succ, 1, graph);
					    *tail += (unsigned long)
						sizeof(unsigned long);
					}
				    if (refDist	< (unsigned long) ~dist)
					{
					    dist = (unsigned long) ~refDist;
					    AdrSkAbs(graph,
						     graph->now + counter *
						     sizeof(unsigned long));
					    AdrWrLng(&dist, 1, graph);
					}
				}
			    if (awCount != 0L)
				GphFlNext(graph, &awAddr, &succ);
			} while (awCount != 0L);
		}
	    if (*head == *tail)
		break;
	    AdrSkAbs(graph, graph->now + *head);
	    AdrReLng(present, 1, graph);
	    *head += (unsigned long) sizeof(unsigned long);
	    NdsSkAbs(graph, *present);
	    NdsReLng(graph->latest_rec, graph->num_node_fields,
		     graph);
	    graph->latest_node = (unsigned long) -1L;
	    succ =
		graph->now + graph->latest_rec[ORDINAL_POS] *
		sizeof(unsigned long);
	    AdrSkAbs(graph, succ);
	    AdrReLng(&dist, 1, graph);
	}
E_N_D: ;
    if (((PATH *) NULL) != *path)
	PthDiscard(*path);
    if (((PATH *) NULL) != *next)
	PthDiscard(*next);
    return((PATH *) NULL);
}

FE4(void, IsntArrow,
    FILE *, file,
    GRAPH *, graph,
    unsigned long, world,
    unsigned long, awNr)
{
    (void) fprintf(file, "Node %lu does not have successor arrow %lu\n",
		   GphOrdinal(graph, world), awNr);
    (void) fflush(file);
}

FE3(void, DoesntKnow,
    FILE *, file,
    GRAPH *, graph,
    unsigned long, world)
{
    (void) fprintf(file,
		   "%s does not know whether node %lu has successors.\n",
		   myName, GphOrdinal(graph, world));
    (void)
	fputs
	("The reachability graph has not been completely generated.\n",
	 file);
    (void) fflush(file);
}

FE3(void, NodeBelongs,
    FILE *, file,
    GRAPH *, graph,
    unsigned long, world)
{
    (void) fprintf(file, "Node %lu", GphOrdinal(graph, world));
    if (((unsigned long) 0L) != graph->glob[SGC_COUNT_POS])
	{
	    (void) fputs(", belongs to strongly connected component ",
			 file);
	    SgcDenote(file, GphCompNumber(graph, world));
	}
    (void) fputc((int) '\n', file);
    (void) fflush(file);
}
