/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  mark.c
*
* Part of: PROD's programs araprod, "gphgen", probe, and prpp
*
* Authors: Marko Rauhamaa and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     This module contains routines for the manipulation of markings.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "mark.h"

unsigned _magNofPlaces;       /* the number of places in a net */
unsigned *_magProfile = NULL; /* _magProfile[i] is the default arity
                               * for tuples in the (i + 1)th place */
STATE *_state = NULL;        /* the current marking of the net */
STATE *_folState = NULL;     /* a computed successor marking of _state */
STATE *_evalState = NULL;    /* a marking of the net to be compared to
                              * the label of a state of a Buchi automaton */

FE0(STATE *, StaAlloc)
{
    STATE *state;
    register unsigned place;

    if (!MagGNofPlaces())
	return((STATE *) NULL);
    state = (STATE *) AllocArray(MagGNofPlaces(), sizeof(MARKING));
    for (place = 0; place < MagGNofPlaces(); place++)
	state[place] = (MARKING *) NULL;
    return(state);
}

/***************************************************************************
*
* Function:  StaFrCopy
* Purpose:   This routine makes a copy of the "free part" of a marking of
*            a net. The free part is the part which remains when the effect
*            of input arcs of a transition is taken into account, without
*            taking into account the effect of output arcs.
* Arguments: state - the original marking
* Returns:   a copy of the free part
* See also:  StaDiscard, StaDump
*
****************************************************************************
*/
FE1(STATE *, StaFrCopy,
    STATE *, state)
{
    STATE *copyState;
    register unsigned place;
    MARKING *marking, *copyMarking;
    MULTUP *multup;
    VOID *cursor;

    copyState = StaAlloc();
    for (place = 0; place < MagGNofPlaces(); place++)
        {
	    copyState[place] = MarCreate();
	    if (MarEmpty(state[place]))
		continue;
            marking = state[place];
            copyMarking = copyState[place];
	    cursor = MarGet(marking);
            MARFOR (marking, multup)
		{
		    if (MultupGFreeCount(multup) != 0L)
			MarInsLast(copyMarking,
				   MultupCons(MultupGFreeCount(multup),
					   TupCopy(MultupGTuple(multup))));
		}
	    MarSet(marking, cursor);
        }
    return(copyState);
}

/***************************************************************************
*
* Function:  StaDracsid
* Purpose:   This routine removes a marking of a net from the memory.
* Arguments: state - the marking
* Returns:
* See also:  StaFrCopy, StaDump
*
****************************************************************************
*/
FE1(void, StaDracsid,
    STATE *, state)
{
    register unsigned place;

    if (state)
	{
	    if (state == _state)
		_state = (STATE *) NULL;
	    if (state == _folState)
		_folState = (STATE *) NULL;
	    if (state == _evalState)
		_evalState = (STATE *) NULL;
	    for (place = 0; place < MagGNofPlaces(); place++)
		MarDiscard(state[place]);
	    Deallocate((char *) state);
	}
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  StaDump
* Purpose:   This routine displays a marking of a net.
* Arguments: state - the marking
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  StaFrCopy, StaDiscard
*
****************************************************************************
*/
FE2(void, StaDump,
    STATE *, state,
    unsigned, depth)
{
    register unsigned place;

    DumpIndent(depth);
    (void) fprintf(_err, "STATE (%u places)\n", MagGNofPlaces());
    for (place = 0; place < MagGNofPlaces(); place++)
	MarDump(state[place], depth + 1);
    DumpIndent(depth);
    (void) fputs("End of STATE\n", _err);
    (void) fflush(_err);
}

/***************************************************************************
*
* Function:  MarDump
* Purpose:   This routine displays a marking of a place.
* Arguments: marking - the marking
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  all routines having a name of the form Mar[A-Za-z]+
*
****************************************************************************
*/
FE2(void, MarDump,
    MARKING *, marking,
    unsigned, depth)
{
    MULTUP *multup;
    VOID *cursor;

    DumpIndent(depth);
    (void) fputs("MARKING\n", _err);
    cursor = MarGet(marking);
    MARFOR (marking, multup)
	MultupDump(multup, depth + 1);
    MarSet(marking, cursor);
    DumpIndent(depth);
    (void) fputs("End of MARKING\n", _err);
    (void) fflush(_err);
}
#endif

/***************************************************************************
*
* Function:  MarDelete
* Purpose:   This routine subtracts a multituple from a marking of a place.
* Arguments: marking - the marking
*            multup - the multituple
* Returns:
* See also:  all routines having a name of the form Mar[A-Za-z]+
*
****************************************************************************
*/
FE2(void, MarDelete,
    MARKING *, marking,
    MULTUP *, multup)
{
    MULTUP *multup2;
    VOID *cursor;
    int cmpVal = 0;

    cursor = NULL;
    MARFOR (marking, multup2)
	{
	    cmpVal = TupComp(MultupGTuple(multup), MultupGTuple(multup2));
	    if (cmpVal <= 0)
		break;
	    cursor = MarGet(marking);
	}
    if (multup2 && (cmpVal == 0))
	{
	    if (MultupGCount(multup2) <= MultupGCount(multup))
		{
		    MarSet(marking, cursor);
		    MarDelNext(marking);
		}
	    else
		MultupDCount(multup2, MultupGCount(multup));
	}
}

/***************************************************************************
*
* Function:  MarInsert
* Purpose:   This routine adds a multituple to a marking of a place.
* Arguments: marking - the marking
*            multup - the multituple
* Returns:
* See also:  all routines having a name of the form Mar[A-Za-z]+
*
****************************************************************************
*/
FE2(void, MarInsert,
    MARKING *, marking,
    MULTUP *, multup)
{
    MULTUP *multup2;
    VOID *cursor;
    int cmpVal = 0;

    cursor = NULL;
    MARFOR (marking, multup2)
	{
	    cmpVal = TupComp(MultupGTuple(multup), MultupGTuple(multup2));
	    if (cmpVal <= 0)
		break;
	    cursor = MarGet(marking);
	}
    if ((!multup2) || (cmpVal < 0))
	{
	    MarSet(marking, cursor);
	    MarInsNext(marking, multup);
	}
    else
	{
	    MultupICount(multup2, MultupGCount(multup));
	    MultupDiscard(multup);
	}
}

/***************************************************************************
*
* Function:  MarLess
* Purpose:   This routine checks whether a marking of a place is less than
*            another.
* Arguments: marking1 - the first marking
*            marking2 - the second marking
* Returns:   TRUE if and only if each tuple in marking1 occurs at least
*            equally many times in marking2 and the markings are not equal
* See also:  all routines having a name of the form Mar[A-Za-z]+
*
****************************************************************************
*/
FE2(BOOL, MarLess,
    MARKING *, marking1,
    MARKING *, marking2)
{
    MULTUP *multup1, *multup2;
    BOOL flag = FALSE;
    int cmpVal = 0;

    multup2 = MarBegin(marking2);
    MARFOR (marking1, multup1)
	{
	    while (multup2)
		{
		    cmpVal = TupComp(MultupGTuple(multup1), MultupGTuple(multup2));
		    if (cmpVal <= 0)
			break;
		    multup2 = MarStep(marking2);
		    flag = TRUE;
		}
	    if ((!multup2) || (cmpVal < 0) ||
		(MultupGCount(multup1) > MultupGCount(multup2)))
		return(FALSE);
	    if (MultupGCount(multup1) < MultupGCount(multup2))
		flag = TRUE;
	    multup2 = MarStep(marking2);
	}
    return(multup2 || flag);
}

/***************************************************************************
*
* Function:  MarLequal
* Purpose:   This routine checks whether a marking of a place is less than
*            or equal to another.
* Arguments: marking1 - the first marking
*            marking2 - the second marking
* Returns:   TRUE if and only if each tuple in marking1 occurs at least
*            equally many times in marking2
* See also:  all routines having a name of the form Mar[A-Za-z]+
*
****************************************************************************
*/
FE2(BOOL, MarLequal,
    MARKING *, marking1,
    MARKING *, marking2)
{
    MULTUP *multup1, *multup2;
    int cmpVal = 0;

    multup2 = MarBegin(marking2);
    MARFOR (marking1, multup1)
	{
	    while (multup2)
		{
		    cmpVal = TupComp(MultupGTuple(multup1), MultupGTuple(multup2));
		    if (cmpVal <= 0)
			break;
		    multup2 = MarStep(marking2);
		}
	    if ((!multup2) || (cmpVal < 0) ||
		(MultupGCount(multup1) > MultupGCount(multup2)))
		return(FALSE);
	    multup2 = MarStep(marking2);
	}
    return(TRUE);
}

/***************************************************************************
*
* Function:  MarEqual
* Purpose:   This routine checks whether a marking of a place is equal
*            to another.
* Arguments: marking1 - the first marking
*            marking2 - the second marking
* Returns:   TRUE if and only if the markings are equal
* See also:  all routines having a name of the form Mar[A-Za-z]+
*
****************************************************************************
*/
FE2(BOOL, MarEqual,
    MARKING *, marking1,
    MARKING *, marking2)
{
    MULTUP *multup1, *multup2;

    if (marking1 == marking2)
	return(TRUE);
    for (multup1 = MarBegin(marking1), multup2 = MarBegin(marking2);
	 multup1 && multup2;
	 multup1 = MarStep(marking1), multup2 = MarStep(marking2))
	{
	    if ((MultupGCount(multup1) != MultupGCount(multup2))
		|| (TupComp(MultupGTuple(multup1), MultupGTuple(multup2)) != 0))
		return(FALSE);
	}
    return((!multup1) && (!multup2));
}

/***************************************************************************
*
* Function:  MarProduct
* Purpose:   This routine computes the product of a number and a marking
*            of a place.
* Arguments: marking - the marking (will contain the result)
*            coeff - the number
* Returns:   the product
* See also:  all routines having a name of the form Mar[A-Za-z]+
*
****************************************************************************
*/
FE2(MARKING *, MarProduct,
    MARKING *, marking,
    unsigned long, coeff)
{
    MULTUP *multup;
    
    if (coeff == 0L)
	{
	    MarDiscard(marking);
	    return(MarCreate());
	}
    if (coeff != 1L)
	{
	    MARFOR (marking, multup)
		MultupSCount(multup, coeff * MultupGCount(multup));
	}
    return(marking);
}

/***************************************************************************
*
* Function:  MarCartes
* Purpose:   This routine computes the Cartesian product of two
*            place markings.
* Arguments: marking1 - the first marking
*            marking2 - the second marking
* Returns:   the Cartesian product
* See also:  all routines having a name of the form Mar[A-Za-z]+
*
****************************************************************************
*/
FE2(MARKING *, MarCartes,
    MARKING *, marking1,
    MARKING *, marking2)
{
    MULTUP *multup1, *multup2, *multup3;
    MARKING *marking3;
    unsigned i;

    marking3 = MarCreate();
    MARFOR (marking1, multup1)
        MARFOR (marking2, multup2)
            {
                multup3 = MultupCons(MultupGCount(multup1) *
                                     MultupGCount(multup2),
                                     TupCons(MultupGArity(multup1) +
                                             MultupGArity(multup2)));
                for (i = 0; i < MultupGArity(multup1); i++)
                    MultupSField(multup3, i, MultupGField(multup1, i));
                for (i = 0; i < MultupGArity(multup2); i++)
                    MultupSField(multup3, i + MultupGArity(multup1),
                                 MultupGField(multup2, i));
                MarInsert(marking3, multup3);
            }
    return(marking3);
}

/***************************************************************************
*
* Function:  MarAdd
* Purpose:   This routine computes the sum of two place markings.
* Arguments: marking1 - the first marking (will contain the result)
*            marking2 - the second marking
* Returns:   the sum
* See also:  all routines having a name of the form Mar[A-Za-z]+
*
****************************************************************************
*/
FE2(MARKING *, MarAdd,
    MARKING *, marking1,
    MARKING *, marking2)
{
    MULTUP *multup;

    MARFOR (marking2, multup)
	MarInsert(marking1, MultupCpy(multup));
    return(marking1);
}

/***************************************************************************
*
* Function:  MarSubtract
* Purpose:   This routine computes the complement of a marking of
*            a place with respect to another.
* Arguments: marking1 - a marking (will contain the result)
*            marking2 - a marking
* Returns:   the complement of marking2 with respect to marking1
* See also:  all routines having a name of the form Mar[A-Za-z]+
*
****************************************************************************
*/
FE2(MARKING *, MarSubtract,
    MARKING *, marking1,
    MARKING *, marking2)
{
    MULTUP *multup;

    MARFOR (marking2, multup)
	MarDelete(marking1, multup);
    return(marking1);
}

/***************************************************************************
*
* Function:  MarIntersect
* Purpose:   This routine computes the intersections of two place markings.
* Arguments: marking1 - the first marking (will contain the result)
*            marking2 - the second marking
* Returns:   the intersection
* See also:  all routines having a name of the form Mar[A-Za-z]+
*
****************************************************************************
*/
FE2(MARKING *, MarIntersect,
    MARKING *, marking1,
    MARKING *, marking2)
{
    MARKING *difference;

    difference = MarCopy(marking1);
    (void) MarSubtract(difference, marking2);
    (void) MarSubtract(marking1, difference);
    MarDiscard(difference);
    return(marking1);
}

/***************************************************************************
*
* Function:  MarCatenate
* Purpose:   This routine concatenates two place markings. It is meaningful
*            to call this routine only if it is guaranteed that the
*            result of the concatenation is the sum of the markings.
* Arguments: marking1 - the first marking (will contain the result)
*            marking2 - the second marking (will be deleted here)
* Returns:   the result of concatenating marking2 after marking1
* See also:  all routines having a name of the form Mar[A-Za-z]+
*
****************************************************************************
*/
FE2(MARKING *, MarCatenate,
    MARKING *, marking1,
    MARKING *, marking2)
{
    (void) LstCatenate(marking1, marking2);
    MarDiscard(marking2);
    return(marking1);
}

/***************************************************************************
*
* Function:  MultupCons
* Purpose:   This routine constructs a multituple.
* Arguments: count - the number of occurrences
*                    of the tuple of the multituple
*            tuple - the tuple of the multituple
* Returns:   the multituple
* See also:  MultupCpy, MultupDiscard, MultupDump
*
****************************************************************************
*/
FE2(MULTUP *, MultupCons,
    unsigned long, count,
    TUPLE *, tuple)
{
    MULTUP *multup;
    
    multup = MultupAlloc();
    MultupSCount(multup, count);
    MultupSFreeCount(multup, count);
    MultupSTuple(multup, tuple);
    return(multup);
}

/***************************************************************************
*
* Function:  MultupCpy
* Purpose:   This routine makes a copy of a multituple.
* Arguments: multup - the multituple to be copied
* Returns:   the copy
* See also:  MultupCons, MultupDiscard, MultupDump
*
****************************************************************************
*/
FE1(MULTUP *, MultupCpy,
    MULTUP *, multup)
{
    MULTUP *copy;
    
    copy = MultupAlloc();
    MultupSCount(copy, MultupGCount(multup));
    MultupSFreeCount(copy, MultupGFreeCount(multup));
    MultupSTuple(copy, TupCopy(MultupGTuple(multup)));
    return(copy);
}

/***************************************************************************
*
* Function:  MultupDracsid
* Purpose:   This routine removes a multituple from the memory.
* Arguments: multup - the multituple
* Returns:
* See also:  MultupCons, MultupCpy, MultupDump
*
****************************************************************************
*/
FE1(void, MultupDracsid,
    MULTUP *, multup)
{
    if (multup)
	{
	    TupDiscard(MultupGTuple(multup));
	    Deallocate((char *) multup);
	}
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  MultupDump
* Purpose:   Thus routine displays a multituple.
* Arguments: multup - the multituple
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  MultupCons, MultupCpy, MultupDiscard
*
****************************************************************************
*/
FE2(void, MultupDump,
    MULTUP *, multup,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "MULTUP count %lu freeCount %lu\n",
		   MultupGCount(multup), MultupGFreeCount(multup));
    TupDump(MultupGTuple(multup), depth + 1);
}
#endif

/* A collection of routines for multituples to be used by
 * list routines of dynam.c and table routines of ptrbl.c. */
TOOLS multupTools =
{ (CFUNC) MultupCpy, MultupDracsid
#ifdef DEBUG
, MultupDump
#endif
 };

/***************************************************************************
*
* Function:  TupCons
* Purpose:   This routine constructs a tuple. The fields are not
*            initialized.
* Arguments: arity - the arity of the tuple
* Returns:   the tuple
* See also:  TupCopy, TupDiscard, TupDump, TupComp
*
****************************************************************************
*/
FE1(TUPLE *, TupCons,
    unsigned, arity)
{
    TUPLE *tuple;

    tuple = TupAlloc();
    TupSArity(tuple, arity);
    if (TupGArity(tuple))
	TupSArray(tuple, (DOMAIN *)
		  AllocArray(TupGArity(tuple), sizeof(DOMAIN)));
    else
	TupSArray(tuple, (DOMAIN *) NULL);
    return(tuple);
}

/***************************************************************************
*
* Function:  TupCopy
* Purpose:   This routine makes a copy of a tuple.
* Arguments: tuple - the tuple to be copied
* Returns:   the copy
* See also:  TupCons, TupDiscard, TupDump, TupComp
*
****************************************************************************
*/
FE1(TUPLE *, TupCopy,
    TUPLE *, tuple)
{
    TUPLE *copy;
    register unsigned i;
    register DOMAIN *d, *s;

    copy = TupAlloc();
    TupSArity(copy, TupGArity(tuple));
    if (TupGArity(tuple))
	{
	    TupSArray(copy, (DOMAIN *)
		      AllocArray(TupGArity(tuple), sizeof(DOMAIN)));
	    for (d = TupGArray(copy), s = TupGArray(tuple), i = 0;
		 i < TupGArity(tuple); *d++ = *s++, i++)
		;
	}
    else
	TupSArray(copy, (DOMAIN *) NULL);
    return(copy);
}

/***************************************************************************
*
* Function:  TupDracsid
* Purpose:   This routine removes a tuple from the memory.
* Arguments: tuple - the tuple
* Returns:
* See also:  TupCons, TupCopy, TupDump, TupComp
*
****************************************************************************
*/
FE1(void, TupDracsid,
    TUPLE *, tuple)
{
    if (tuple)
	{
	    if (TupGArity(tuple))
		Deallocate((char *) TupGArray(tuple));
	    Deallocate((char *) tuple);
	}
}

/***************************************************************************
*
* Function:  TupDump
* Purpose:   This routine displays a tuple.
* Arguments: tuple - the tuple
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  TupCons, TupCopy, TupDiscard, TupComp
*
****************************************************************************
*/
FE2(void, TupDump,
    TUPLE *, tuple,
    unsigned, depth)
{
    register unsigned i;
    register DOMAIN *s;

    DumpIndent(depth);
    (void) fputs("TUPLE <.", _err);
    if (TupGArity(tuple))
	for (s = TupGArray(tuple), i = 0; i < TupGArity(tuple); s++, i++)
	    {
		if (i)
		    (void) fputs(", ", _err);
		(void) fprintf(_err, DOMFORMAT, *s);
	    }
    (void) fputs(".>\n", _err);
    (void) fflush(_err);
}

/***************************************************************************
*
* Function:  TupComp
* Purpose:   This routine compares two tuples to each other.
* Arguments: tuple1 - the first tuple
*            tuple2 - the second tuple
* Returns:   -1 if the first is less than the second,
*             1 if the first is greater than the second, and
*             0 otherwise
* See also:  TupCons, TupCopy, TupDiscard, TupDump
*
****************************************************************************
*/
FE2(int, TupComp,
    TUPLE *, tuple1,
    TUPLE *, tuple2)
{
    register unsigned i;
    register DOMAIN *f1, *f2;

    if (TupGArity(tuple1) < TupGArity(tuple2))
	return(-1);
    if (TupGArity(tuple1) > TupGArity(tuple2))
	return(1);
    if (!TupGArity(tuple1))
	return(0);
    for (f1 = TupGArray(tuple1), f2 = TupGArray(tuple2), i = 0;
	 i < TupGArity(tuple1); f1++, f2++, i++)
	if (*f1 < *f2)
	    return(-1);
	else if (*f1 > *f2)
	    return(1);
    return(0);
}

/* A collection of routines for tuples to be used by
 * list routines of dynam.c and table routines of ptrbl.c. */
TOOLS tupTools =
{ (CFUNC) TupCopy, TupDracsid
#ifdef DEBUG
, TupDump
#endif
 };
