/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  prepctl.c
*
* Part of: PROD's program probe
*
* Authors: Keijo Heljanko
*
* Current version:
*     $Id: prepctl.c,v 3.1 1997/04/17 15:44:04 kepa Exp $
*
* Description:
*     This module implements the preprocessing of the aws file
*     to a form better suited for CTL model-checking. The files
*     created end in .caw and .cad .
*
* Version history:
*     $Log: prepctl.c,v $
 * Revision 3.1  1997/04/17  15:44:04  kepa
 * *** empty log message ***
 *
***************************************************************************/


/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "gphutil.h"
#include "bdd.h"
#include "graph.h"
#include "error.h"
#include "dassert.h"
#include "mark.h"
#include "expr.h"
#include "formula.h"
#include "bitarr.h"
#include "stackpro.h"
#include "temporal.h"
#include "checkctl.h"
#include "prepctl.h"


static BOOL CTLArrowsPrepared = FALSE;
/***************************************************************************
*
* Function:  PrepareCTL
* Purpose:   Prepares datastructures and files ready for
*            CTL model checking.
* Arguments: graph - the reachability graph
* Returns:   TRUE on success, FALSE on failure.
* See also:
*
****************************************************************************
*/
FE1(BOOL, PrepareCTL,
    GRAPH *, graph)
{
    BOOL result = TRUE;
    
    if(CTLArrowsPrepared == FALSE) {
	result = AwsPrepareCTL(graph);

	if(result != FALSE) {
	    CTLArrowsPrepared = TRUE;
	}

    }
    
    return result;
}



/***************************************************************************
*
* Function:  AwsPrepareCTL
* Purpose:   Creates the files .cad and .caw ,which are needed in CTL
*            model checking.
* Arguments: graph - the reachabilty graph
*            name  - the name of the graph
* Returns:   TRUE on success, FALSE on failure.
* See also:
*
****************************************************************************
*/
FE1(BOOL, AwsPrepareCTL,
    GRAPH *, graph)
{
    BINARYFILE cad = NonBinaryFile();
    BINARYFILE caw = NonBinaryFile();
    char *name;
    char *cad_name;
    char *caw_name;
    
    unsigned long i;
    unsigned long caw_offset;

    unsigned long num_nodes;
    unsigned long num_arrows;
    unsigned long node_addr;
    unsigned long aw_count;
    
    unsigned long aw_addr = 0L;
    unsigned long succ = 0L;
    unsigned long succ_ord;

    unsigned long terminator;
    int oldreason;

    unsigned long cad_len;
    unsigned long caw_len;
    
    terminator = ~((unsigned long) 0L);

    name = graph->name;
    num_nodes = GphNdCount(graph);
    num_arrows = GphGlAwCount(graph);
    caw_offset = 0;

    /* This stuff is NOT done with speed in mind, therefore
       it should be rewritten later.

       kepa 1996/07/16 */

    
    /* Add here checks for creating the adress file only
       when the graph has changed after the .cad and .caw
       files have been created.

       Maybe we could use something as simple as node and
       arrow count compared to the size of the .cad and
       .caw file respectively. */

    /* Here is a simple size check only. */
    cad_name = FileName(name, "cad");
    caw_name = FileName(name, "caw");

    cad_len = GetFileLen(cad_name);
    caw_len = GetFileLen(caw_name);
    
    if ((cad_len != (unsigned long) -1L) &&
	(cad_len == (unsigned long)
	 (num_nodes * sizeof(unsigned long))) &&
	(caw_len != (unsigned long) -1L) &&
	(caw_len == (unsigned long)
	 ((num_nodes + num_arrows) * sizeof(unsigned long))))
	{
	    /* The files exist and the sizes match, exit. */
	    return(TRUE);
	}
    
    /* Create the CTL arrow address file. */

    cad = CreateWrite(cad_name);

    if(NotBinaryFile(cad)) {
	(void) fprintf(_err,
			   "\nCould not create file: %s\n", cad_name);
	(void) fflush(_err);
	Deallocate((char *)cad_name);
	Deallocate((char *)caw_name);
	return(FALSE);
    }

    Deallocate((char *)cad_name);


    /* Create the CTL arrow file. */
    
    caw = CreateWrite(caw_name);

    if(NotBinaryFile(caw)) {
	(void) fprintf(_err,
			   "\nCould not create file: %s\n", caw_name);
	(void) fflush(_err);
	Deallocate((char *)caw_name);
	OwnClose(&cad);
	return(FALSE);
    }

    Deallocate((char *)caw_name);
    

    /* Both the .cad and .caw files have unsigned long data
       elements only.

       
       The .cad CTL address file contains offsets to the .caw
       CTL address file.

       The .cad file contains one unsigned long for each node
       ordered in the ordinal number order. (Ordinal numbers
       range from 0 to num_nodes-1.) This is the offset to
       .caw file, and if you want to get a byte offset from it,
       it must be multiplied by sizeof(unsigned long).

       
       The .caw CTL arrow file contains a lists of ordinal
       numbers of successor nodes. This list is terminated
       with a ~(0UL). 

       The lists in .caw file are in increasing ordinal
       number order. So the list of successors for the
       node 0 starts at offset 0, followed by successors
       of the node 1 etc.


       The order of successors in the lists is the same as
       they are returned from the graph. This may change if
       some other order speeds up the model checking process. */


    /* Change the graph->reason to speed up the
       graph processing. (GPH_PROBE reason will read
       the marking of the state into memory also. The
       marking isn't needed here. */
    
    oldreason = graph->reason;
    graph->reason = GPH_PREPARE_CTL;
    
    for(i=0; i<num_nodes; i++) {

	/* Get the node address and as a side-effect
	   reads the node fields to memory. */

	node_addr = GphId(graph, i);

	/* Add the node into the .cad offset file. */
	
	WriteLng(&caw_offset, 1, cad);


	/* Check if this node really has successors. */

	aw_count = GphAwCount(graph, node_addr);
	
	if (aw_count == 0 ||
	    aw_count == ((unsigned long) -1L)) {

	    /* No successor nodes, add nothing. */
	    ;

	} else {

	    /* Successors found, add them. */

	    /* Get the first successor arrow. */

	    GphFlFirst(graph, node_addr, &aw_addr, &succ);

	    /* Get the ordinal number of the successor. */
	    
	    succ_ord = GphOrdinal(graph, succ);

	    /* Add the ordinal number to the caw file. */

	    WriteLng(&succ_ord, 1, caw);
	    caw_offset++;
	    aw_count--;

	    while(aw_count != 0) {

		/* Get next successor arrow. */

		GphFlNext(graph, &aw_addr, &succ);
		
		/* Get the ordinal number of the successor. */
	    
		succ_ord = GphOrdinal(graph, succ);

		/* Add the ordinal number to the caw file. */

		WriteLng(&succ_ord, 1, caw);
		caw_offset++;
		aw_count--;

	    }
	    
	}

	/* Add a successor list terminator next. */
	
	WriteLng(&terminator, 1, caw);
	caw_offset++;
	
    }

    /* Restore the graph->reason. */
    
    graph->reason = oldreason;

    OwnClose(&cad);
    OwnClose(&caw);

    return(TRUE);
}

/***************************************************************************
*
* Function:  EvalAtomicFormulas
* Purpose:   Evaluates the atomic formulas in all the states of the
*            graph.
* Arguments: graph - the reachability graph
* Returns:   TRUE on success, FALSE on failure.
* See also:
*
****************************************************************************
*/
FE2(BOOL, EvalAtomicFormulas,
    GRAPH *, graph,
    LIST *, formulas)
{
    LIST *atomic;
    unsigned long num_nodes;
    unsigned long i;
    unsigned long node_addr;
    CTL_FORMULA *node;
    
    atomic = LstCreate(&CTLFormulaTools);

    LSTFOR(formulas, node, CTL_FORMULA) {
	if(node->formulatype == ATOMIC_FORMULA) {
	    (void) LstInsLast(atomic, (char *)node);
	}
    }

    if(!LstEmpty(atomic)) {

	/* There are some atomic subformulas to evaluate. */

	/* Loop through the states in ordinal number order. */

	num_nodes = GphNdCount(graph);
	
	for(i=0; i<num_nodes; i++) {
	    /* Get the node address and as a side-effect
	       reads the node fields to memory. If
	       graph->reason == GPH_PROBE it also gets
	       the marking info into memory. */

	    node_addr = GphId(graph, i);
	    FogSGraph(graph);
	    FogSWorld(node_addr);

	    /* Evaluate all atomic formulas in it. */

	    /* All the formulas are initialized to 0,
	       so we only need to set them to TRUE
	       when the atomic proposition is TRUE. */
	    
	    LSTFOR(atomic, node, CTL_FORMULA) {

		if(XprEval(node->p)) { 
		    SetBit(node->result, i);
		}

	    }
	    
	}

	/* Mark the formulas evaluated. */
    
	LSTFOR(atomic, node, CTL_FORMULA) {
	    node->formulatype = EVALUATED_FORMULA;
	}

    }

    LstDestroy(atomic, FALSE);

    return TRUE;
}

/***************************************************************************
*
* Function:  GetFile
* Purpose:   Get the contents of the file and return a pointer
*            to the start of the newly allocated memory area.
* Arguments: fileName - the name of the file to be loaded
* Returns:   a pointer to a newly allocated memory area or
*            NULL on failure.
* See also:
*
****************************************************************************
*/
FE1(VOID *, GetFile,
    char *, fileName)
{
    VOID *mem = NULL;
    BINARYFILE file = NonBinaryFile();
    unsigned long bytes;
    
    file = OpenRead(fileName, (BOOL *) NULL);
    if (NotBinaryFile(file))
	return(NULL);
    bytes = SeekEnd(file);
    if ((((unsigned long) -1L) == bytes) || (0L == bytes) ||
	!(mem = Hallocate(bytes)))
	{
	    OwnClose(&file);
	    return(NULL);
	}
    SeekAbs(file, (unsigned long) 0L);
    ReadBytes((char *) mem, bytes, file);
    OwnClose(&file);
    return(mem);
}

/***************************************************************************
*
* Function:  GetFileLen
* Purpose:   Get the size of the file
* Arguments: fileName - the name of the file to be checked
* Returns:   the file size, -1 on failure
* See also:
*
****************************************************************************
*/
FE1(unsigned long, GetFileLen,
    char *, fileName)
{
    BINARYFILE file = NonBinaryFile();
    long bytes;
    BOOL rel = TRUE;
    
    file = OpenRead(fileName, &rel);
    if (NotBinaryFile(file))
	return((unsigned long) -1L);
    bytes = SeekEnd(file);
    OwnClose(&file);
    return(bytes);
}
