/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  ptrbl.c
*
* Part of: PROD's programs araprod, "gphgen", probe, and strong
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     This module contains general routines for segmented tables.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ptrbl.h"

/* An array of pointers to segments is of size power(2, SEGBITS) or less. */
#define SEGBITS 11
#define SEGMASK ((((unsigned) 1) << SEGBITS) - 1)

/***************************************************************************
*
* Function:  SpaAlloc
* Purpose:   This routine reserves space for an array of pointers.
*            If all pointers do not fit in a single array,
*            a new level of segments is created.
* Arguments: ptrCount - the number of leaf pointers
*            bits - an array of leaf pointers is created if bits is 0,
*                   otherwise a new level of segments is created
* Returns:   an pointer array of pointers, possibly segmented
* See also:  SpaApply, SpaBackApply, SpaAddrApply, SpaBkdrApply,
*            SpaCopy, SpaDupl, SpaDiscard, SpaDispose, SpaDump
*
****************************************************************************
*/
FS2(VOID **, SpaAlloc,
    unsigned long, ptrCount,
    unsigned, bits)
{
    VOID **spa;
    register unsigned a;
    register unsigned long n;

    a = (unsigned) ((ptrCount - 1L) >> bits);
    spa = (VOID **) AllocArray(a + 1, sizeof(VOID *));
    if (bits)
	{
	    n = ((unsigned long) 1L) << bits;
	    bits -= SEGBITS;
	    spa[a] = (VOID *) SpaAlloc(((ptrCount - 1L) & (n - 1L)) + 1L, 
				       bits);
	    while (a)
		spa[--a] = (VOID *) SpaAlloc(n, bits);
	}
    return(spa);
}

static VOID **cause;

/***************************************************************************
*
* Function:  SpaApply
* Purpose:   This routine traverses an array of pointers and calls
*            a boolean function at every leaf pointer. The leaf pointers
*            are visited in a growing order with respect to the order
*            numbers of the leaf pointers. The traversal is broken
*            if f returns TRUE.
* Arguments: spa - an array of pointers, possibly segmented
*            ptrCount - the number of leaf pointers
*            bits - an array of leaf pointers is in question if this is 0,
*                   otherwise the array is segmented
*            f - the function
* Returns:   the address of the breaking leaf pointer if such exists,
*            and NULL otherwise
* See also:  SpaAlloc, SpaBackApply, SpaAddrApply, SpaBkdrApply,
*            SpaCopy, SpaDupl, SpaDiscard, SpaDispose, SpaDump
*
****************************************************************************
*/
FS4(VOID **, SpaApply,
    VOID **, spa,
    unsigned long, ptrCount,
    unsigned, bits,
    FOOL, f)
{
    register unsigned a, i;
    register unsigned long n;

    a = (unsigned) ((ptrCount - 1L) >> bits);
    if (bits)
	{
	    n = ((unsigned long) 1L) << bits;
	    bits -= SEGBITS;
	    i = 0;
	    while (i < a)
		if ((cause = SpaApply((VOID **) (spa[i++]), n, bits, f)))
		    return(cause);
	    if ((cause = SpaApply((VOID **) (spa[a]),
				  ((ptrCount - 1L) & (n - 1L)) + 1L,
				  bits, f)))
		return(cause);
	}
    else
	for (i = 0; ; )
	    {
		if ((*f)(spa[i]))
		    return(spa + i);
		if (i == a)
		    break;
		i++;
	    }
    return(NULL);
}

/***************************************************************************
*
* Function:  SpaBackApply
* Purpose:   This is like SpaApply, except that the order of the traversal
*            is reverse.
* Arguments: spa - an array of pointers, possibly segmented
*            ptrCount - the number of leaf pointers
*            bits - an array of leaf pointers is in question if this is 0,
*                   otherwise the array is segmented
*            f - the function
* Returns:   the address of the breaking leaf pointer if such exists,
*            and NULL otherwise
* See also:  SpaAlloc, SpaApply, SpaAddrApply, SpaBkdrApply,
*            SpaCopy, SpaDupl, SpaDiscard, SpaDispose, SpaDump
*
****************************************************************************
*/
FS4(VOID **, SpaBackApply,
    VOID **, spa,
    unsigned long, ptrCount,
    unsigned, bits,
    FOOL, f)
{
    register unsigned a;
    register unsigned long n;

    a = (unsigned) ((ptrCount - 1L) >> bits);
    if (bits)
	{
	    n = ((unsigned long) 1L) << bits;
	    bits -= SEGBITS;
	    if ((cause = SpaBackApply((VOID **) (spa[a]),
				      ((ptrCount - 1L) & (n - 1L)) + 1L,
				      bits, f)))
		return(cause);
	    while (a)
		if ((cause = SpaBackApply((VOID **) (spa[--a]), n, bits, f)))
		    return(cause);
	}
    else
	while (TRUE)
	    {
		if ((*f)(spa[a]))
		    return(spa + a);
		if (!a)
		    break;
		a--;
	    }
    return(NULL);
}

/***************************************************************************
*
* Function:  SpaAddrApply
* Purpose:   This is like SpaApply, except that the argument of
*            the given function is one addressing level higher.
* Arguments: spa - an array of pointers, possibly segmented
*            ptrCount - the number of leaf pointers
*            bits - an array of leaf pointers is in question if this is 0,
*                   otherwise the array is segmented
*            f - the function
* Returns:   the address of the breaking leaf pointer if such exists,
*            and NULL otherwise
* See also:  SpaAlloc, SpaApply, SpaBackApply, SpaBkdrApply,
*            SpaCopy, SpaDupl, SpaDiscard, SpaDispose, SpaDump
*
****************************************************************************
*/
FS4(VOID **, SpaAddrApply,
    VOID **, spa,
    unsigned long, ptrCount,
    unsigned, bits,
    FOOL, f)
{
    register unsigned a, i;
    register unsigned long n;

    a = (unsigned) ((ptrCount - 1L) >> bits);
    if (bits)
	{
	    n = ((unsigned long) 1L) << bits;
	    bits -= SEGBITS;
	    i = 0;
	    while (i < a)
		if ((cause = SpaAddrApply((VOID **) (spa[i++]), n, bits, f)))
		    return(cause);
	    if ((cause = SpaAddrApply((VOID **) (spa[a]),
				      ((ptrCount - 1L) & (n - 1L)) + 1L,
				      bits, f)))
		return(cause);
	}
    else
	for (i = 0; ; )
	    {
		if ((*f)(spa + i))
		    return(spa + i);
		if (i == a)
		    break;
		i++;
	    }
    return(NULL);
}

/***************************************************************************
*
* Function:  SpaBkdrApply
* Purpose:   This is like SpaBackApply, except that the argument of
*            the given function is one addressing level higher.
* Arguments: spa - an array of pointers, possibly segmented
*            ptrCount - the number of leaf pointers
*            bits - an array of leaf pointers is in question if this is 0,
*                   otherwise the array is segmented
*            f - the function
* Returns:   the address of the breaking leaf pointer if such exists,
*            and NULL otherwise
* See also:  SpaAlloc, SpaApply, SpaBackApply, SpaAddrApply,
*            SpaCopy, SpaDupl, SpaDiscard, SpaDispose, SpaDump
*
****************************************************************************
*/
FS4(VOID **, SpaBkdrApply,
    VOID **, spa,
    unsigned long, ptrCount,
    unsigned, bits,
    FOOL, f)
{
    register unsigned a;
    register unsigned long n;

    a = (unsigned) ((ptrCount - 1L) >> bits);
    if (bits)
	{
	    n = ((unsigned long) 1L) << bits;
	    bits -= SEGBITS;
	    if ((cause = SpaBkdrApply((VOID **) (spa[a]),
				      ((ptrCount - 1L) & (n - 1L)) + 1L,
				      bits, f)))
		return(cause);
	    while (a)
		if ((cause = SpaBkdrApply((VOID **) (spa[--a]), n, bits, f)))
		    return(cause);
	}
    else
	while (TRUE)
	    {
		if ((*f)(spa + a))
		    return(spa + a);
		if (!a)
		    break;
		a--;
	    }
    return(NULL);
}

/***************************************************************************
*
* Function:  SpaCopy
* Purpose:   This routine makes a copy of an array of pointers.
* Arguments: spa - an array of pointers, possibly segmented
*            ptrCount - the number of leaf pointers
*            bits - an array of leaf pointers is in question if this is 0,
*                   otherwise the array is segmented
*            fcopy - this function copies the stuff pointed by a leaf
*                    pointer
* Returns:   the copy
* See also:  SpaAlloc, SpaApply, SpaBackApply, SpaAddrApply, SpaBkdrApply,
*            SpaDupl, SpaDiscard, SpaDispose, SpaDump
*
****************************************************************************
*/
FS4(VOID **, SpaCopy,
    VOID **, spa,
    unsigned long, ptrCount,
    unsigned, bits,
    CFUNC, fcopy)
{
    VOID **copy;
    register unsigned a;
    register unsigned long n;

    a = (unsigned) ((ptrCount - 1L) >> bits);
    copy = (VOID **) AllocArray(a + 1, sizeof(VOID *));
    if (bits)
	{
	    n = ((unsigned long) 1L) << bits;
	    bits -= SEGBITS;
	    copy[a] = (VOID *) SpaCopy((VOID **) (spa[a]), 
				       ((ptrCount - 1L) & (n - 1L)) + 1L,
				       bits, fcopy);
	    while (a--)
		copy[a] = (VOID *) SpaCopy((VOID **) (spa[a]), n, 
					   bits, fcopy);
	}
    else
	{
	    copy[a] = (*fcopy)(spa[a]);
	    while (a--) 
		copy[a] = (*fcopy)(spa[a]);
	}
    return(copy);
}

/***************************************************************************
*
* Function:  SpaDupl
* Purpose:   This is like SpaCopy except that the stuff pointed by
*            the leaf pointers is not copied.
* Arguments: spa - an array of pointers, possibly segmented
*            ptrCount - the number of leaf pointers
*            bits - an array of leaf pointers is in question if this is 0,
*                   otherwise the array is segmented
* Returns:   the copy
* See also:  SpaAlloc, SpaApply, SpaBackApply, SpaAddrApply, SpaBkdrApply,
*            SpaCopy, SpaDiscard, SpaDispose, SpaDump
*
****************************************************************************
*/
FS3(VOID **, SpaDupl,
    VOID **, spa,
    unsigned long, ptrCount,
    unsigned, bits)
{
    VOID **copy;
    register unsigned a;
    register unsigned long n;

    a = (unsigned) ((ptrCount - 1L) >> bits);
    copy = (VOID **) AllocArray(a + 1, sizeof(VOID *));
    if (bits)
	{
	    n = ((unsigned long) 1L) << bits;
	    bits -= SEGBITS;
	    copy[a] = (VOID *) SpaDupl((VOID **) (spa[a]), 
				       ((ptrCount - 1L) & (n - 1L)) + 1L,
				       bits);
	    while (a--)
		copy[a] = (VOID *) SpaDupl((VOID **) (spa[a]), n, bits);
	}
    else
	{
	    copy[a] = spa[a];
	    while (a--) 
		copy[a] = spa[a];
	}
    return(copy);
}

/***************************************************************************
*
* Function:  SpaDiscard
* Purpose:   This routine removes an array of pointers from the memory.
* Arguments: spa - an array of pointers, possibly segmented
*            ptrCount - the number of leaf pointers
*            bits - an array of leaf pointers is in question if this is 0,
*                   otherwise the array is segmented
*            discard - this function removes the stuff pointed by a leaf
*                      pointer from the memory
* Returns:
* See also:  SpaAlloc, SpaApply, SpaBackApply, SpaAddrApply, SpaBkdrApply,
*            SpaCopy, SpaDupl, SpaDispose, SpaDump
*
****************************************************************************
*/
FS4(void, SpaDiscard,
    VOID **, spa,
    unsigned long, ptrCount,
    unsigned, bits,
    VOF, discard)
{
    register unsigned a;
    register unsigned long n;

    a = (unsigned) ((ptrCount - 1L) >> bits);
    if (bits)
	{
	    n = ((unsigned long) 1L) << bits;
	    bits -= SEGBITS;
	    SpaDiscard((VOID **) (spa[a]),
		       ((ptrCount - 1L) & (n - 1L)) + 1L,
		       bits, discard);
	    while (a)
		SpaDiscard((VOID **) (spa[--a]), n, bits, discard);
	}
    else
	{
	    (*discard)(spa[a]);
	    while (a)
		(*discard)(spa[--a]);
	}
    Deallocate((char *) spa);
}

/***************************************************************************
*
* Function:  SpaDispose
* Purpose:   This is like SpaDiscard except that the stuff pointed by
*            leaf pointers is not touched.
* Arguments: spa - an array of pointers, possibly segmented
*            ptrCount - the number of leaf pointers
*            bits - an array of leaf pointers is in question if this is 0,
*                   otherwise the array is segmented
* Returns:
* See also:  SpaAlloc, SpaApply, SpaBackApply, SpaAddrApply, SpaBkdrApply,
*            SpaCopy, SpaDupl, SpaDiscard, SpaDump
*
****************************************************************************
*/
FS3(void, SpaDispose,
    VOID **, spa,
    unsigned long, ptrCount,
    unsigned, bits)
{
    register unsigned a;
    register unsigned long n;

    a = (unsigned) ((ptrCount - 1L) >> bits);
    if (bits)
	{
	    n = ((unsigned long) 1L) << bits;
	    bits -= SEGBITS;
	    SpaDispose((VOID **) (spa[a]),
		       ((ptrCount - 1L) & (n - 1L)) + 1L,
		       bits);
	    while (a)
		SpaDispose((VOID **) (spa[--a]), n, bits);
	}
    Deallocate((char *) spa);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  SpaDump
* Purpose:   This routine displays an array of pointers.
* Arguments: spa - an array of pointers, possibly segmented
*            ptrCount - the number of leaf pointers
*            bits - an array of leaf pointers is in question if this is 0,
*                   otherwise the array is segmented
*            dump - this function displays the stuff pointed by a leaf
*                   pointer
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  SpaAlloc, SpaApply, SpaBackApply, SpaAddrApply, SpaBkdrApply,
*            SpaCopy, SpaDupl, SpaDiscard, SpaDispose
*
****************************************************************************
*/
FS5(void, SpaDump,
    VOID **, spa,
    unsigned long, ptrCount,
    unsigned, bits,
    VOF, dump,
    unsigned, depth)
{
    register unsigned a, i;
    register unsigned long n;

    a = (unsigned) ((ptrCount - 1L) >> bits);
    if (bits)
	{
	    n = ((unsigned long) 1L) << bits;
	    bits -= SEGBITS;
	    i = 0;
	    while (i < a)
		{
		    DumpIndent(depth);
		    (void) fputs("SEGMENT\n", _err);
		    SpaDump((VOID **) (spa[i++]), n, bits, dump, depth + 1);
		    DumpIndent(depth);
		    (void) fputs("End of SEGMENT\n", _err);
		}
	    DumpIndent(depth);
	    (void) fputs("SEGMENT\n", _err);
	    SpaDump((VOID **) (spa[a]),
		    ((ptrCount - 1L) & (n - 1L)) + 1L,
		    bits, dump, depth + 1);
	    DumpIndent(depth);
	    (void) fputs("End of SEGMENT\n", _err);
	}
    else
	{
	    i = 0;
	    while (i < a)
		(*dump)(spa[i++], depth);
	    (*dump)(spa[a], depth);
	}
}
#endif

/***************************************************************************
*
* Function:  PtrBlCreate
* Purpose:   This routine creates a segmented pointer table.
* Arguments: ptrCount - the number of leaf pointers
*            tools - a collection of routines needed for copying,
*                    deleting and displaying of the stuff pointed
*                    by the leaf pointers
* Returns:   the table
* See also:  PtrBlCopy, PtrBlDestroy, PtrBlDump, PtrBlPut, PtrBlGet,
*            PtrBlApply, PtrBlBackApply, PtrBlAddrApply, PtrBlBkdrApply,
*            SpaAlloc
*
****************************************************************************
*/
FE2(PTRBL *, PtrBlCreate,
    unsigned long, ptrCount,
    TOOLS *, tools)
{
    PTRBL *ptrbl;
    register unsigned bits;

    bits = 0;
    if (ptrCount != 0L)
	{
	    while (((ptrCount - 1L) >> (bits + SEGBITS)) != 0L)
		bits += SEGBITS;
	}
    ptrbl = PtrBlAlloc();
    ptrbl->ptrCount = ptrCount;
    ptrbl->bits = bits;
    ptrbl->tools = tools;
    if (ptrCount != 0L)
	ptrbl->root = SpaAlloc(ptrCount, bits);
    else
	ptrbl->root = NULL;
    return(ptrbl);
}

/***************************************************************************
*
* Function:  PtrBlCopy
* Purpose:   This routine makes a copy of a segmented pointer table.
* Arguments: ptrbl - the table to be copied
*            elCopy - the stuff pointed by the leaf pointers is copied
*                     if and only if elCopy is TRUE
* Returns:   the copy
* See also:  PtrBlCreate, PtrBlDestroy, PtrBlDump, PtrBlPut, PtrBlGet,
*            PtrBlApply, PtrBlBackApply, PtrBlAddrApply, PtrBlBkdrApply,
*            SpaCopy, SpaDupl
*
****************************************************************************
*/
FE2(PTRBL *, PtrBlCopy,
    PTRBL *, ptrbl,
    BOOL, elCopy)
{
    PTRBL *copy;

    copy = PtrBlAlloc();
    copy->ptrCount = ptrbl->ptrCount;
    copy->bits = ptrbl->bits;
    copy->tools = ptrbl->tools;
    if (ptrbl->root)
	{
	    if (elCopy)
		copy->root = SpaCopy(ptrbl->root, ptrbl->ptrCount,
				     ptrbl->bits, ptrbl->tools->_copy);
	    else
		copy->root = SpaDupl(ptrbl->root, ptrbl->ptrCount,
				     ptrbl->bits);
	}
    else
	copy->root = NULL;
    return(copy);
}

/***************************************************************************
*
* Function:  PtrBlDstroy
* Purpose:   This routine removes a segmented pointer table from the memory.
* Arguments: ptrbl - the table
*            elDiscard - the stuff pointed by the leaf pointers is removed
*                        if and only if elDiscard is TRUE
* Returns:
* See also:  PtrBlCreate, PtrBlCopy, PtrBlDump, PtrBlPut, PtrBlGet,
*            PtrBlApply, PtrBlBackApply, PtrBlAddrApply, PtrBlBkdrApply,
*            SpaDiscard, SpaDispose
*
****************************************************************************
*/
FE2(void, PtrBlDstroy,
    PTRBL *, ptrbl,
    BOOL, elDiscard)
{
    if (ptrbl)
	{
	    if (ptrbl->root)
		{
		    if (elDiscard)
			SpaDiscard(ptrbl->root, ptrbl->ptrCount,
				   ptrbl->bits, ptrbl->tools->_discard);
		    else
			SpaDispose(ptrbl->root, ptrbl->ptrCount,
				   ptrbl->bits);
		}
	    Deallocate((char *) ptrbl);
	}
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  PtrBlDump
* Purpose:   This routine displays a segmented pointer table.
* Arguments: ptrbl - the table
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  PtrBlCreate, PtrBlCopy, PtrBlDestroy, PtrBlPut, PtrBlGet,
*            PtrBlApply, PtrBlBackApply, PtrBlAddrApply, PtrBlBkdrApply,
*            SpaDump
*
****************************************************************************
*/
FE2(void, PtrBlDump,
    PTRBL *, ptrbl,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fputs("PTRBL\n", _err);
    if (ptrbl->root)
	SpaDump(ptrbl->root, ptrbl->ptrCount,
		ptrbl->bits, ptrbl->tools->_dump, depth + 1);
    DumpIndent(depth);
    (void) fputs("End of PTRBL\n", _err);
    (void) fflush(_err);
}
#endif

/***************************************************************************
*
* Function:  PtrBlPut
* Purpose:   This routine resets a leaf pointer of a segmented pointer
*            table.
* Arguments: ptrbl - the table
*            ptr - the new value of the leaf pointer
*            number - the order number of the leaf pointer in the table
* Returns:
* See also:  PtrBlCreate, PtrBlCopy, PtrBlDestroy, PtrBlDump, PtrBlGet,
*            PtrBlApply, PtrBlBackApply, PtrBlAddrApply, PtrBlBkdrApply
*
****************************************************************************
*/
FE3(void, PtrBlPut,
    PTRBL *, ptrbl,
    char *, ptr,
    unsigned long, number)
{
    register VOID **spa;
    register unsigned bits = ptrbl->bits;

    if (number >= ptrbl->ptrCount)
	Troba3("Illegal PtrblPut call: %lu >= %lu",
	       number, ptrbl->ptrCount);
    spa = ptrbl->root;
    while (bits)
	{
	    spa = (VOID **) spa[(unsigned) ((number >> bits) & SEGMASK)];
	    bits -= SEGBITS;
	}
    spa[(unsigned) (number & SEGMASK)] = Pvoid(ptr);
}

/***************************************************************************
*
* Function:  PtrBlGet
* Purpose:   This routine returns the value of a leaf pointer of a
*            segmented pointer table.
* Arguments: ptrbl - the table
*            number - the order number of the leaf pointer in the table
* Returns:   the value of the leaf pointer
* See also:  PtrBlCreate, PtrBlCopy, PtrBlDestroy, PtrBlDump, PtrBlPut,
*            PtrBlApply, PtrBlBackApply, PtrBlAddrApply, PtrBlBkdrApply
*
****************************************************************************
*/
FE2(VOID *, PtrBlGet,
    PTRBL *, ptrbl,
    unsigned long, number)
{
    register VOID **spa;
    register unsigned bits = ptrbl->bits;

    if (number >= ptrbl->ptrCount)
	Troba3("Illegal PtrblGet call: %lu >= %lu",
	       number, ptrbl->ptrCount);
    spa = ptrbl->root;
    while (bits)
	{
	    spa = (VOID **) spa[(unsigned) ((number >> bits) & SEGMASK)];
	    bits -= SEGBITS;
	}
    return(spa[(unsigned) (number & SEGMASK)]);
}

/***************************************************************************
*
* Function:  PtrBlApply
* Purpose:   This routine traverses a segmented pointer table and calls
*            a boolean function at each leaf pointer. The leaf pointers
*            are visited in a growing order with respect to the order
*            numbers of the leaf pointers. The traversal is broken
*            if f returns TRUE.
* Arguments: ptrbl - the table
*            f - the function
* Returns:   the address of the breaking leaf pointer if such exists,
*            and NULL otherwise
* See also:  PtrBlCreate, PtrBlCopy, PtrBlDestroy, PtrBlDump, PtrBlPut,
*            PtrBlGet, PtrBlBackApply, PtrBlAddrApply, PtrBlBkdrApply,
*            SpaApply
*
****************************************************************************
*/
FE2(VOID **, PtrBlApply,
    PTRBL *, ptrbl,
    FOOL, f)
{
    return((ptrbl->root &&
	    (cause = SpaApply(ptrbl->root, ptrbl->ptrCount, ptrbl->bits, f)))
	   ? cause : NULL);
}

/***************************************************************************
*
* Function:  PtrBlBackApply
* Purpose:   This is like PtrBlApply, except that the order of the traversal
*            is reverse.
* Arguments: ptrbl - the table
*            f - the function
* Returns:   the address of the breaking leaf pointer if such exists,
*            and NULL otherwise
* See also:  PtrBlCreate, PtrBlCopy, PtrBlDestroy, PtrBlDump, PtrBlPut,
*            PtrBlGet, PtrBlApply, PtrBlAddrApply, PtrBlBkdrApply,
*            SpaBackApply
*
****************************************************************************
*/
FE2(VOID **, PtrBlBackApply,
    PTRBL *, ptrbl,
    FOOL, f)
{
    return((ptrbl->root &&
	    (cause =
	     SpaBackApply(ptrbl->root, ptrbl->ptrCount, ptrbl->bits, f)))
	   ? cause : NULL);
}

/***************************************************************************
*
* Function:  PtrBlAddrApply
* Purpose:   This is like PtrBlApply, except that the argument of the
*            given function is one addressing level higher.
* Arguments: ptrbl - the table
*            f - the fucntion
* Returns:   the address of the breaking leaf pointer if such exists,
*            and NULL otherwise
* See also:  PtrBlCreate, PtrBlCopy, PtrBlDestroy, PtrBlDump, PtrBlPut,
*            PtrBlGet, PtrBlApply, PtrBlBackApply, PtrBlBkdrApply,
*            SpaAddrApply
*
****************************************************************************
*/
FE2(VOID **, PtrBlAddrApply,
    PTRBL *, ptrbl,
    FOOL, f)
{
    return((ptrbl->root &&
	    (cause =
	     SpaAddrApply(ptrbl->root, ptrbl->ptrCount, ptrbl->bits, f)))
	   ? cause : NULL);
}

/***************************************************************************
*
* Function:  PtrBlBkdrApply
* Purpose:   This is like PtrBlBackApply, except that the argument of the
*            given function is one addressing level higher.
* Arguments: ptrbl - the table
*            f - the function
* Returns:   the address of the breaking leaf pointer if such exists,
*            and NULL otherwise
* See also:  PtrBlCreate, PtrBlCopy, PtrBlDestroy, PtrBlDump, PtrBlPut,
*            PtrBlGet, PtrBlApply, PtrBlBackApply, PtrBlAddrApply,
*            SpaBkdrApply
*
****************************************************************************
*/
FE2(VOID **, PtrBlBkdrApply,
    PTRBL *, ptrbl,
    FOOL, f)
{
    return((ptrbl->root &&
	    (cause =
	     SpaBkdrApply(ptrbl->root, ptrbl->ptrCount, ptrbl->bits, f)))
	   ? cause : NULL);
}
