/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  redirect.c
*
* Part of: PROD's program probe
*
* Authors: Keijo Heljanko
*
* Current version:
*     $Id: redirect.c,v 3.1 1997/04/17 15:44:07 kepa Exp $
*
* Description:
*     This module handles redirection of PROBE output to a file
*     or to a filter prgram.
*
* Version history:
*     $Log: redirect.c,v $
 * Revision 3.1  1997/04/17  15:44:07  kepa
 * *** empty log message ***
 *
***************************************************************************/

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "mbufs.h"
#include "mark.h"
#include "expr.h"
#include "gphutil.h"
#include "graph.h"
#include "formula.h"
#include "comline.h"
#include "probe.h"
#include "redirect.h"

static FILE *oldout = NULL;
static FILE *f = NULL;
#if (defined(UNIX) || defined(_WIN32))
static char *bfFlNm = NULL;
#endif

/***************************************************************************
*
* Function:  InitRedirect
* Purpose:
* Arguments: r - a pointer to a redirect struct
* Returns:   BOOL - Was the redirection successful
* See also:
*
****************************************************************************
*/
FE1(BOOL, InitRedirect,
    REDIRECT *, r)
{
    oldout = _out;
    if (r->type != REDIRECT_NONE)
	{
#ifdef DEBUG_REDIRECT
	    (void) fprintf(_err, "\nRedirection: %d \"%s\"\n",
			   r->type, r->name);
	    if (r->args)
		{
		    (void) fprintf(_err, "\nArgs: \"%s\"\n",
				   r->args);
		}
#endif
	    if (!(r->name))
		{
		    r->type = REDIRECT_NONE;
		    return(FALSE);
		}
	    switch (r->type)
		{
		case REDIRECT_FILE:
		    if (!(f = MyFopen(r->name, "w", (BOOL *) NULL)))
			{
			    r->type = REDIRECT_NONE;
			    return(FALSE);
			}
		    _out = f;
		    break;
		case REDIRECT_FILE_APPEND:
		    if (!(f = MyFopen(r->name, "a", (BOOL *) NULL)))
			{
			    r->type = REDIRECT_NONE;
			    return(FALSE);
			}
		    _out = f;
		    break;
		case REDIRECT_FILTER:
#if (defined(UNIX) || defined(_WIN32))
		    bfFlNm = FileName(prefDefault, "buf");
		    if (!(f = MyFopen(bfFlNm, "w", (BOOL *) NULL)))
			{
			    r->type = REDIRECT_NONE;
			    return(FALSE);
			}
		    _out = f;
#else
		    (void)
			fprintf
			(_err, 
			 "\nRedirecting to a filter not supported on this system\n");
		    (void) fflush(_err); 
#endif
		    break;
		case REDIRECT_NONE:
		default:
		    break;
		}
	}
    return(TRUE);
}

/***************************************************************************
*
* Function:  CloseRedirect
* Purpose:   Closes the redirection file and in case of a filter
*            feeds all the output through a forked filter process.
*            Also restores the original output destination.
* Arguments: r - a pointer to a redirect struct
* Returns:   BOOL - was the redirection closing successful
* See also:
*
****************************************************************************
*/
FE1(BOOL, CloseRedirect,
    REDIRECT *, r)
{
    _out = oldout;
    switch (r->type)
	{
	case REDIRECT_FILE:
	case REDIRECT_FILE_APPEND:
	    if (f)
		{
		    (void) fflush(f);
		    ClFclose(f);
		}
	    break;
	case REDIRECT_FILTER:
#if (defined(UNIX) || defined(_WIN32))
	    if (f)
		{
		    (void) fflush(f);
		    if (((unsigned long) -1L) ==
			MyFskTell(f, (unsigned long) 0L, (int) SEEK_SET))
			{
			    ClFclose(f);
			    if (bfFlNm)
				{
				    (void) unlink(bfFlNm);
				    ClDeallocate(bfFlNm);
				}
			    r->type = REDIRECT_NONE;
			    return(FALSE);
			}
		    ClFclose(f);
		}
	    RunFilterSystem(r->name, r->args);
	    if (bfFlNm)
		{
		    (void) unlink(bfFlNm);
		    ClDeallocate(bfFlNm);
		}
#endif
	    break;
	case REDIRECT_NONE:
	default:
	    break;
	}
    return(TRUE);
}

/***************************************************************************
*
* Function:  CleanupRedirect
* Purpose:   Deallocates all memory and other resources used in redirection.
*            Cleans the redirect struct.
* Arguments: 
* Returns:
* See also:
*
****************************************************************************
*/
FE1(BOOL, CleanupRedirect,
    REDIRECT *, r)
{
    r->type = REDIRECT_NONE;
    ClDeallocate(r->name);
    ClDeallocate(r->args);
    return(TRUE);
}

#if (defined(UNIX) || defined(_WIN32))
/***************************************************************************
*
* Function:  RunFilterSystem
* Purpose:   Feeds the contents of a temporary file through a filter
*            and back to _out. The filter arguments used are in string
*            cmdline.
*
* Arguments: temp - a file descriptor to a temporary file,
*                   must point at the beginning
*            filtername - Name of the filter program to be run, it must
*                         exist in the user search path.
*            cmdline - A string containing the commmandline used as
*                      arguments to the filter. NULL if none.
*
****************************************************************************
*/
FE2(void, RunFilterSystem,
    char *, filtername,
    char *, cmdline)
{
    char *toBeEx;

    if (!filtername)
	filtername = "";
    if (!cmdline)
	cmdline = "";
    toBeEx = (char *) Allocate(strlen(filtername) + 3 + strlen(bfFlNm) + 1
			       + strlen(cmdline) + 1);
    (void) strcpy(toBeEx, filtername);
    (void) strcat(toBeEx, " < ");
    (void) strcat(toBeEx, bfFlNm);
    (void) strcat(toBeEx, " ");
    (void) strcat(toBeEx, cmdline);
#ifdef DEBUG_REDIRECT
    (void) fprintf(_err, "System: \"%s\"\n", toBeEx);
    (void) fflush(_err);
#endif
    /* We make a string which is of the form:
       toBeEx = "filtername < buffer-file cmdline-args"
       it is then executed by "/bin/sh -c toBeEx" on Unix
       and by its moral equivalent in WIN32. */
    (void) MySystem(toBeEx, FALSE);
    Deallocate(toBeEx);
}
#endif
