/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  stackpro.c
*
* Part of: PROD's program probe
*
* Authors: Keijo Heljanko
*
* Current version:
*     $Id$
*
* Description:
*     This module creates a stack of items and stack operations on it.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "dassert.h"
#include "stackpro.h"


/***************************************************************************
*
* Function:  CreateStack
* Purpose:   Creates a stack.
* Arguments:  
* Returns:   A pointer to a created stack, NULL on failure.
* See also:
*
****************************************************************************
*/
FE0(STACK *, CreateStack)
{
    STACK *stack = NULL;
    stack = (STACK *)Allocate(sizeof(STACK));

    DEBUG_ASSERT(stack != NULL);
      
    /* Initialize all fields of stack to defaults. */

    stack->curr_numitems = 0;
    stack->curr_item = NULL;

    stack->blocks = NULL;
    stack->last = NULL;

    stack->numblocks = 0;
    stack->blocks_in_memory = 0;

    /* Paging the stack to disk will be added later. */

    stack->filename = NULL;

    AllocBlock(stack);
    
    return stack;
}

/***************************************************************************
*
* Function:  AllocBlock
* Purpose:   Allocates memory for one stack block.
* Arguments: stack - a pointer to a stack
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, AllocBlock,
    STACK *, stack)
{
	
    STACK_BLOCK *b;
    STACK_ITEM *p;
    
    /* Add here flushing an old block to disk if treshold exceeded. */

    /* Allocate a new stack block. */

    b = (STACK_BLOCK *)Allocate(sizeof(STACK_BLOCK));
    DEBUG_ASSERT(b != NULL);

    p = (STACK_ITEM *) HallocArray((unsigned long) STACK_BLOCK_SIZE(stack),
				   sizeof(STACK_ITEM));
    DEBUG_ASSERT(p != NULL);

    b->block = p;                /* Contents of the block. */
    b->prev = stack->blocks;     /* Maintain a linked list of blocks. */

    stack->curr_numitems = 0;    /* Now we have a fresh stack block. */
    stack->curr_item = p;
    stack->blocks = b;

    if(stack->last == NULL) {    /* Store here the last block in memory
				    for fast access. */
	stack->last = b;
    }

    stack->numblocks++;
    stack->blocks_in_memory++;
}


/***************************************************************************
*
* Function:  DeleteStack
* Purpose:   Removes a stack from memory.
* Arguments: stack - a pointer to a stack.
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, DeleteStack,
    STACK *, stack)
{
    DEBUG_ASSERT(stack != NULL);

    /* Delete all stack blocks. */

    ClearStack(stack);

    /* ClearStack leaves one block allocated
       on purpose. Deallocate it also so we
       won't leak memory. */
    
    Dehallocate((char *)stack->blocks->block);
    Deallocate((char *)stack->blocks);
    Deallocate((char *)stack);

    return;
}

/***************************************************************************
*
* Function:  ClearStack
* Purpose:   Removes items and all blocks except for the last one from
*            memory.
* Arguments: stack - a pointer to a stack.
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, ClearStack,
    STACK *, stack)
{

    STACK_BLOCK *p = NULL;
    STACK_BLOCK *r;
    STACK_ITEM *b = NULL;
    
    DEBUG_ASSERT(stack != NULL);

    for(p = stack->blocks; p != NULL; p = r) {
	r = p->prev;
	b = p->block;

	/* Keep the last block in the linked list allocated for us. */
	if(r == NULL) {
	    break;
	}

	if(b != NULL) {
	    Dehallocate((char *)b);
	}

	DEBUG_ASSERT(p != NULL);

	Deallocate((char *)p);
	
    }

    /* Initialize all fields of stack to defaults. */

    DEBUG_ASSERT(p != NULL);
    DEBUG_ASSERT(b != NULL);

    stack->curr_numitems = 0;
    stack->curr_item = b;

    stack->blocks = p;
    stack->last = p;

    stack->numblocks = 1;
    stack->blocks_in_memory = 1;

    return;
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  DumpStack
* Purpose:   Display the contents of a stack.
* Arguments: stack - a pointer to a stack
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, DumpStack,
    STACK *, stack,
    unsigned, depth)
{
    DEBUG_ASSERT(stack != NULL);

    /* This is only a partial dump of the data in stack. */
    
    DumpIndent(depth);
    (void) fputs("STACK:\n", _err);
    (void) fflush(_err);

    depth++;
    
    DumpIndent(depth);
    (void) fprintf(_err, "size %lu\n", StackSize(stack));
    DumpIndent(depth);
    (void) fprintf(_err, "blocks %lu\n", stack->numblocks);
    DumpIndent(depth);
    (void) fprintf(_err, "blocks in memory %lu\n", stack->blocks_in_memory);

    (void) fflush(_err);

}
#endif


/***************************************************************************
*
* Function:  RealPushStack
* Purpose:   Push an item to stack.
* Arguments: stack - a pointer to a stack
*            item - the item to be pushed on the stack
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, RealPushStack,
    STACK *, stack,
    STACK_ITEM, item)
{
    DEBUG_ASSERT(stack != NULL);
    DEBUG_ASSERT(stack->curr_numitems == STACK_BLOCK_SIZE(stack));

    if(stack->curr_numitems == STACK_BLOCK_SIZE(stack) || stack->numblocks == 0) {

	AllocBlock(stack);
    }


    DEBUG_ASSERT(stack->numblocks > 0);
    DEBUG_ASSERT(stack->blocks_in_memory > 0);
    DEBUG_ASSERT(stack->curr_numitems == 0);
    DEBUG_ASSERT(stack->curr_item != NULL);

    stack->curr_numitems++;
    *(stack->curr_item) = item;
    stack->curr_item++;

    return;
}
     
/***************************************************************************
*
* Function:  RealPopStack
* Purpose:   Pop an item from the top of the stack.
* Arguments: stack - a pointer to a stack
* Returns:   The topmost item on stack.
* See also:
*
****************************************************************************
*/
FE1(STACK_ITEM, RealPopStack,
    STACK *, stack)
{
    STACK_ITEM item;
    STACK_BLOCK *b;
    STACK_ITEM *p;
    
    DEBUG_ASSERT(stack != NULL);

    DEBUG_ASSERT(stack->curr_numitems == 0);

    if(stack->numblocks == 0) {
	ASSERT(FALSE);
    }
    
    stack->numblocks--;

    /* Add getting stuff from disk here. */
    
    stack->blocks_in_memory--;
    
    b = stack->blocks;
    if (b == NULL) {
	ASSERT(FALSE);
    }

    p = b->block;
    DEBUG_ASSERT(p != NULL);
    
    stack->blocks = stack->blocks->prev;
    DEBUG_ASSERT(stack->blocks != NULL);

    if(stack->last == b) {
	ASSERT(FALSE);
    }
    
    Dehallocate((char *)p);
    Deallocate((char *)b);

    stack->curr_numitems = STACK_BLOCK_SIZE(stack);
    stack->curr_numitems--;

    stack->curr_item = stack->blocks->block + STACK_BLOCK_SIZE(stack);
    stack->curr_item--;
    
    item = *(stack->curr_item);

    return item;
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  RealPeekStack
* Purpose:   Peeks the topmost item on the stack.
* Arguments: stack - a pointer to a stack
* Returns:   The topmost item on the stack.
* See also:
*
****************************************************************************
*/
FE1(STACK_ITEM, RealPeekStack,
    STACK *, stack)
{
    STACK_ITEM item;
    
    DEBUG_ASSERT(stack != NULL);

    item = PopStack(stack);
    PushStack(stack, item);

    return item;
}

/***************************************************************************
*
* Function:  StackSize
* Purpose:   Gives the amount of items on stack.
* Arguments: stack - a pointer to a stack
* Returns:   The number of items.
* See also:
*
****************************************************************************
*/
FE1(unsigned long, StackSize,
    STACK *, stack)
{
    DEBUG_ASSERT(stack != NULL);

    if(stack->numblocks == 0) {
	ASSERT(FALSE);
    } else {
	return (((stack->numblocks - 1) * STACK_BLOCK_SIZE(stack)) +
		stack->curr_numitems);
    }

    return 0;
}
#endif
